#!/usr/bin/env bash
#
# Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
#

___help_topic "===== Environment ==============="


function +requirements
    set --local options 'h/help' 'l/limited' 'n/noisy' 'q/quiet'
    argparse $options -- $argv
    or ___error "See '+requirements --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+requirements [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Compile the current 'requirements.in' file into a usable 'requirements.txt' file."; echo

          echo "Options:"
          echo "  -l | --limited"
          echo "    Limited to changes, Don't use the -U 'option'"
          echo "  -n | --noisy"
          echo "    Compile with verbose"
          echo "  -q | --quiet"
          echo "    Compile with quiet"
        set_color normal
        echo
        return 0
    end

    pushd $HINT_TESTING/environment

    set -l use_U "-U"
    if set -q _flag_limited
        set use_U ""
    end
    set -l use_verbose
    if set -q _flag_noisy
        set use_verbose "-v"
    end
    if set -q _flag_quiet
        set use_verbose "-q"
    end
    set -l extra --extra-index-url https://${ARTIFACTORY_USERNAME}:${ARTIFACTORY_PASSWORD}@usw1.packages.broadcom.com/artifactory/api/pypi/cb-pypi-prod-virtual/simple

    ___action "Compiling './environment/requirements.in' ..."
    pip-compile $extra $use_U $use_verbose requirements.in --trusted-host usw1.packages.broadcom.com -o requirements.txt
    or begin; ___error "Unable to compile requirements.in! "; popd; return 1; end

    if test $status -eq 0
        popd
    else
        popd
        ___error "Failure to compile!"
    end
end
___comm "+requirements"
___info "Compile the 'requirements.in' file in the environments folder into a usable 'requirements.txt'"

function +swagger
    set --local options 'h/help'
    argparse $options -- $argv
    or ___error "See '+swagger --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+swagger [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Create the 'swagger.yaml' file in the repository root based on code comments."; echo
        set_color normal
        echo
        return 0
    end

    pushd $HINT_ROOT

    ___action "Updating 'swagger.yaml' ..."
    swagger generate spec -o ./swagger.yaml --scan-models
    or begin; ___error "Unable to create swagger! "; popd; return 1; end

    if test $status -eq 0
        popd
    else
        popd
        ___error "Failure to compile!"
    end
end
___comm "+swagger"
___info "Create the 'swagger.yaml' file in the repository root based on code comments"


function ___virtualenv_create -d "Creates/updates the python virtualenv"
    set --local options 'n/noisy' 'u/update'
    argparse $options -- $argv
    or begin; ___error "Incorrect arguments for 'environs.fish': $argv"; exit 1; end

    set -l pip_verbose "-q"
    if set -q _flag_noisy
        set pip_verbose "-v"
    end
    set -l use_upgrade
    set -l message "ensure"
    if set -q _flag_update
        set use_upgrade "--upgrade"
        set message "update"
    end

    if not test -d ~/.virtualenvs/HINT
        if [ "$(which python3; echo $status)" != "0" ]
            ___error "\nPython 3.x is required for testing\n"
            exit 0
        end

        echo "CREATING 'HINT' virtualenv ..."
        python3 -m venv ~/.virtualenvs/HINT
    end

    source ~/.virtualenvs/HINT/bin/activate.fish

    ___action2 "... $message pip"
    ___debug "pip install $pip_verbose $use_upgrade pip"
    pip install $pip_verbose $use_upgrade pip
    ___action2 "... $message pip-tools"
    ___debug "pip install $pip_verbose $use_upgrade pip-tools"
    pip install $pip_verbose $use_upgrade pip-tools

    # make sure requirements are added; clear caches
    rm -rf ~/.cache/pip
    rm -rf ~/Library/Caches/pip

    # handle changes to the python requirements, make sure we have an exported one for this shell
    set -l hash "not-present"
    if test -f $HINT_TESTING/environment/requirements.txt
        set hash (md5sum $HINT_TESTING/environment/requirements.txt | awk '{print $1}')
    end
    set -l oldhash ""
    if test -f $HINT_TESTING/environment/.requirements_hash
        set oldhash (md5sum $HINT_TESTING/environment/.requirements_hash | awk '{print $1}')
    end

    ___debug "Old requirements hash: '$oldhash' -- new hash: '$hash'"
    if [ "$oldhash" != "$hash" ]
        ___action2 "... update requirements.txt "
        pip install $pip_verbose --upgrade -I --trusted-host usw1.packages.broadcom.com -q -r $HINT_TESTING/environment/requirements.txt
        or begin; ___error "Unable to install requirements.txt! "; return 1; end
        echo $hash > $HINT_ROOT/test/functional/environment/.requirements_hash
        ___action3 "    HINTING_FUNCTIONAL_REQUIREMENTS_HASH = $hash"
    else
        ___action2 "... requirements.txt ok"
    end

    ___debug "PYTHONPATH: $PYTHONPATH"
    set -l check "*$HINT_TESTING/suites/lib*"
    switch "$PYTHONPATH"
        case "$check"
            ___action2 "... PYTHONPATH ok (suites/lib)"
        case "*"
            set -gx PYTHONPATH ".:$PYTHONPATH:$HINT_TESTING/suites/lib"
            ___action2 "... updating PYTHONPATH: '$PYTHONPATH'"
    end

      ___debug "PATH: $PATH"
      set -l check "*$HINT_TESTING/utilities*"
      set -e found
      for arg in $PATH
          switch $arg
              case $check
                  ___action2 "... PATH ok"
                  ___debug "PATH: $PATH"
                  set found $arg
          end
      end
      if not set -q found
          set -x PATH $HINT_TESTING/utilities $PATH
          ___action2 "... updating PATH: '$PATH'"
      end
end


function ___virtualenv_remove
    if test -d ~/.virtualenvs/HINT
        ___action "REMOVING 'hinting' virtualenv ..."
        vf deactivate
        rm -rf ~/.virtualenvs/HINT
        set -eU HINTING_FUNCTIONAL_REQUIREMENTS_HASH
    end
end
