#!/usr/bin/env fish
#
# Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
#

___help_topic "===== Local Testing Functions ==============="

function +app
    set --local options 'h/help'
    argparse $options -- $argv
    or ___error "See '+app --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+start-hinting [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Start the hinting service in a local kubernetes cluster."
          echo
          echo "  -h | --help"
          echo "    Print this help"; echo
        set_color normal
        return 0
    end

    pushd $HINT_ROOT
    make local-app-start
    popd
end
___comm "+start-hinting [OPTIONS]"
___info "Start the hinting service locally"

function +infra
    set --local options 'h/help' 'l/locks'
    argparse $options -- $argv
    or ___error "See '+start-infra --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+start-infra [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Start the cluster infrastructure using helm."
          echo

          echo "Options:"
          echo "  -d | --delete"
          echo "    Delete local namespaces"; echo
          echo "  -h | --help"
          echo "    Print this help"; echo
          echo "  -l | --locks"
          echo "    Clear any Chart.lock files before deployment"; echo
        set_color normal
        return 0
    end

    pushd $HINT_ROOT

    set -gx ENV local
    make delete-namespaces

    if set -q _flag_locks
        set -l locks (find $HINT_ROOT/helm -name Chart.lock)
        for file in $locks
            ___notice "Removing $file ..."
            rm -f $file
        end
    end

    make local-infra-start-no-migrate
    or begin; ___error "Unable to start local infrastructure"; popd; return 1; end

    popd
end
___comm "+infra [OPTIONS]"
___info "Start the local infrastructure"


function +pipeline-image
    set --local options 'h/help'
    argparse $options -- $argv
    or ___error "See '+pipeline-image --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+pipeline-image [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Build and execute a docker container for functional testing similar to the pipeline"
          echo

          echo "Options:"
          echo "  -h | --help"
          echo "    Print this help"; echo
        set_color normal
        return 0
    end

    set -l check (docker ps | grep hinting-pytest)
    ## if running, stop existing
    if [ "$check" != "" ]
        set -l id (echo $check | awk '{ print $1 }')
        docker stop $id
    end

    pushd $HINT_TESTING

    # create container image
    ___debug "docker -D -l debug build -t hinting-pytest ."
    docker -D -l debug build -t hinting-pytest .
    or ___error "Unable to build image!"; or return 1

    # mount file system and run image
    ___debug "docker -D -l debug run -v $HINT_ROOT:/test --name=hinting_functionals hinting-pytest &"
    docker -D -l debug run -v $HINT_ROOT:/test  hinting-pytest &

    ___notice "Wait for 'database system is ready to accept connections' message before executing '+pipeline-test"
    popd
end
___comm "+pipeline-image [OPTIONS]"
___info "Build and execute a docker container for functional testing similar to the pipeline"

function +pipeline-kill
    set --local options 'h/help'
    argparse $options -- $argv
    or ___error "See '+pipeline-kill --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+pipeline-kill [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Stop any running pipeline container."
          echo

          echo "Options:"
          echo "  -h | --help"
          echo "    Print this help"; echo
        set_color normal
        return 0
    end

    set -l check (docker ps | grep hinting-pytest)
    if [ "$check" = "" ]
        ___warning "The 'hinting-pytest' container is not running, nothing to do"
        return 0
    end
    set -l id (echo $check | awk '{ print $1 }')
    docker stop $id
end
___comm "+pipeline-kill [OPTIONS]"
___info "Stop any running pipeline container"

function +pipeline-test
    set --local options  'd/debug' 'h/help' 's/shell'
    argparse $options -- $argv
    or ___error "See '+pipeline-test --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+pipeline-test [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Fire off commands to the running hinting-pytest container"
          echo

          echo "Options:"
          echo "  -d | --debug"
          echo "    Turn on script tracing"; echo
          echo "  -h | --help"
          echo "    Print this help"; echo
          echo "  -s | --shell"
          echo "    Go into the shell itself (manual work)"; echo
        set_color normal
        return 0
    end

    set -l check (docker ps | grep hinting-pytest)
    if [ "$check" = "" ]
        ___error "The 'hinting-pytest' container is not running, use '+pipeline-image' first!"
        return 1
    end
    set -l id (echo $check | awk '{ print $1 }')

    if set -q _flag_shell
        ___action "Once in, perform the following:"
        ___action2 "   ./test/functional/run_functionals.sh"
        docker exec -it $id bash
    else
        if set -q _flag_debug
            docker exec -t $id bash ./test/functional/run_functionals.sh -d
        else
            docker exec -t $id bash ./test/functional/run_functionals.sh
        end
    end
end
___comm "+pipeline-image [OPTIONS]"
___info "Build and execute a docker container for functional testing similar to the pipeline"

function +postgres
    set --local options 'h/help'
    argparse $options -- $argv
    or ___error "See '+posgress --help' for usage"; or return 1

    if set -q _flag_help
        set_color $ORIENT_HELP_CODE -o
          echo
          echo "+postgress [OPTIONS]"
        set_color normal; set_color $ORIENT_HELP_CODE
          echo "================================================================================"
        set_color normal; set_color $ORIENT_HELP_CODE2
          echo "Access the postgress pod within the local-infra."
          echo

          echo "Options:"
          echo "  -h | --help"
          echo "    Print this help"; echo
        set_color normal
        return 0
    end

    set -l postgres (kubectl get pods -n local-infra | grep 'ems-device-hints-infra-postgres-deploy-tf-postgres' | awk '{ print $1 }' )
    ___debug "Postgres: $postgres"
    if [ "$postgres" = "" ]
        ___error "infra service is not running locally!"
        return 1
    end

    ___warning "enter postgres with:  psql -h localhost -p 5432 -U \$POSTGRES_USER -d \$POSTGRES_DB"
    kubectl exec -it $postgres -n local-infra -- /bin/bash
end
___comm "+postgres [OPTIONS]"
___info "Access the local infrastructure's postgres"



function __get_si_pid
    set -l pid (ps | grep 'make local-infra-start' | grep -v 'grep' | awk '{ print $1 }')
    echo $pid
end
function __get_sh_pid
    set -l pid (ps | grep 'skaffold run --port-forward=user' | grep -v 'grep' | awk '{ print $1 }')
    echo $pid
end

