# Test Suite Standards
The functional test suites are designed to use `pytest`, which is configured into the environment via the `orient.fish` or 
`setup.sh` scripts. 

<style>
  .tb { border-collapse: collapse}
  .tb td { padding: 2px 10px; border: solid 1px grey}
  .tb th { padding: 2px 10px; border: solid 2px grey; background-color: goldenrod; color:black; text-align: left; }
  tr > th { font-weight: bold }
  tr > td:last-of-type { font-style: italic; }
  tr > td:first-of-type { font-weight: bold }
  xparm {color: goldenrod; font-style: italic; vertical-align: sub; }
  xnote {border-bottom: grey solid; }
  xcode {color: goldenrod; font-style: italic; }
</style>

<hr style="height:2px;border-width:0;color:grey;background-color:goldenrod">

## 1. Location
All test suits should reside within this folder (`suites`), and should be organized by purpose:
<table class="tb">
    <tr><th>Suite</th><th>Purpose</th></tr>
    <tr><td>basic</td><td>Tests that perform minimal acceptance testing of all endpoints (smoke tests)</td></tr>
    <tr><td>extensive</td><td>Tests that perform negative tests (bad data) as well as complex workflows (acceptable and negative)</td></tr>
    <tr><td>internal</td><td>Tests that validate the operation of the functional test models (unit tests for the functional test harness)</td></tr>
    <tr><td>performance</td><td>Long-execution time and spot tests not normally run</td></tr>
</table>

* To prevent any tests in the "performance" suite from executing as a part of the pipeline functional run, they should be specified with
  the "utility" marker:
  ```python
    @pytest.mark.utility(reason="functional test utility, not a test")
    def test_01_utility_sample(self, hc_connect):
        pass
  ```

## 2. Custom Pytest Options
The following custom options are available when executing tests from within the `suites` folder (with it's corresponding
`conftest.py` and `pytest.ini` files)"

### 2.1. Backend Connection
<table class="tb">
    <tr><th>Option</th><th>Purpose</th></tr>
    <tr><td>--ip <i>&lt;address&gt;</i></td><td>Specify the ip address of the hinting service (defaults to 127.0.0.1)</td></tr>
    <tr><td>--port <i>&lt;num&gt;</i></td><td>Specify the port of the hinting service (defaults to 5004)</td></tr>
</table>

### 2.2. Postgres Overrides
The following options will attempt to be filled in from the environment variables of the hinting service container, or
from an `envfile` at the root of the repository.
<table class="tb">
    <tr><th>Option</th><th>Purpose</th></tr>
    <tr><td>--pghost <i>&lt;address&gt;</i></td><td>Specify the address of the postgres service</td></tr>
    <tr><td>--pgport <i>&lt;num&gt;</i></td><td>Specify the port of the postgres service</td></tr>
    <tr><td>--pguser <i>&lt;string&gt;</i></td><td>Specify the postgres user for database access</td></tr>
    <tr><td>--pgpass <i>&lt;string&gt;</i></td><td>Specify the postgres user password</td></tr>
    <tr><td>--dbname <i>&lt;string&gt;</i></td><td>Specify the name of the postgres database</td></tr>
</table>

> _**NOTE:** Those tests that use the defined `PostgresDB` object for database validation should test the ".usable" field --
> if False (due to bad connection information or inability to connect), any method calls for validation should be bypassed._

### 2.3. Miscellaneous Options
<table class="tb">
    <tr><th>Option</th><th>Purpose</th></tr>
    <tr><td>--dbg</td><td>Run tests with DEBUG logging instead of INFO logging</td></tr>
    <tr><td>--envfile <i>&lt;path&gt;</i></td><td>Override the supposed envfile path, for the purpose of determining postgres parameters</td></tr>
</table>

## 3. Common Variables
The following variables are available for any test:

<table class="tb">
    <tr><th>Variable</th><th>Value</th></tr>
    <tr><td>self.LQ_VERSION</td><td>A constant that should be set to the current live query version</td></tr>
    <tr><td>self.ORGKEY</td><td>An OrgKey for the "local" environment, with a unique Id</td></tr>
    <tr><td>self.TOMORROW</td><td>Contains a timestamp equal to one day from now (assured to be after any generated hinting config values)</td></tr>
    <tr><td>self.YESTERDAY</td><td>Contains a timestamp equal to one day ago (assured to predate any generated hinting config values)</td></tr>
</table>

* If `BASEMESSAGE.LIVE_QUERY_VERSION` is updated in [**pkg/hints/basemessage.go**](../../../pkg/hints/basemessage.go),
  then you will need to update `cls.LQ_VERSION` to the same value in [**lib/test_common.py**](./lib/test_common.py).

## 4. Test Fixtures
The following fixtures are available as parameters for any test and become available to the test:

<table class="tb">
    <tr><th>Fixture</th><th>Contains</th></tr>
    <tr><td>ft_config</td><td>(<xcode>FunctionalTestConfig</xcode>) A data class containing the current automation settings</td></tr>
    <tr><td>hc_connect</td><td>(<xcode>HintControlConnect</xcode>) A class with connection methods for each hinting API</td></tr>
    <tr><td>pg_access</td><td>(<xcode>PostgresDB</xcode>) A class with connection to the postgres server along with special access methods</td></tr>
</table>

* Used as a part of the test definition:
  ```python
    def test_01_fixture_sample(self, ft_config, hc_connect):
        print(f"Connecting to {ft_config.SERVICE_IP} on port {ft_config.SERVICE_IP}")
        hc_connect.health_check_get()
        print(f"Health Status: {hc_connect.last_status}")
  ```
* When `hc_connect` API calls are used, the response code of the last call can be obtained from `hc_connect.last_status` and
  the actual response of the last call from `hc_connect.last_content`

---

# Notes
1) List of valid environments (for org_keys) can be found at [**psc/platform/orgkey-go**](https://gitlab.bit9.local/psc/platform/orgkey-go/-/blob/master/configuration.go)
in the definition of `environments`, in `newConfiguration`.

