#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging
import time

import pytest
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.device import Device
from lib.models.device_group import DeviceGroup
from lib.models.enumerations import DeviceType, DeviceChangeType, DeviceGroupChangeType, DeploymentType, OrgChangeType
from lib.models.environment import Env
from lib.models.organization import Org
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


class TestDefintionUpdateApi(TestCommon):
    """
    This class tests the posrtgres database update endpoints.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01a_update_device(self, hc_connect, pg_access):
        """
        Verify that we can update a device with minimal required information (at least one change)

        :class: functional
        """
        data = Device(changes=[DeviceChangeType.DEV_REREGISTER])
        dev_id = self.uid()
        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.group_change_time] is None
            assert check[pg_access.PgDev.sensor_config_update_time] is None
            assert check[pg_access.PgDev.sensor_certificate_update_time] is None
            assert check[pg_access.PgDev.hash_delete_request_time] is None
            assert check[pg_access.PgDev.hash_upload_request_time] is None
            assert check[pg_access.PgDev.logs_request_time] is None
            assert check[pg_access.PgDev.sensor_kit_update_time] is None
            assert check[pg_access.PgDev.perfmon_trigger_time] is None
            assert check[pg_access.PgDev.sensor_action_time] is None
            assert check[pg_access.PgDev.re_register_time] is not None
            assert check[pg_access.PgDev.live_query_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01b_update_device_full(self, hc_connect, pg_access, rd_access):
        """
        Verify that we can update a device that has been fully defined.

        :class: functional
        """
        data = Device.full()
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.group_change_time] is not None
            assert check[pg_access.PgDev.sensor_config_update_time] is not None
            assert check[pg_access.PgDev.sensor_certificate_update_time] is not None
            assert check[pg_access.PgDev.hash_delete_request_time] is not None
            assert check[pg_access.PgDev.hash_upload_request_time] is not None
            assert check[pg_access.PgDev.logs_request_time] is not None
            assert check[pg_access.PgDev.sensor_kit_update_time] is not None
            assert check[pg_access.PgDev.perfmon_trigger_time] is not None
            assert check[pg_access.PgDev.sensor_action_time] is not None
            assert check[pg_access.PgDev.re_register_time] is not None
            assert check[pg_access.PgDev.live_query_change_time] is not None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

        if rd_access.usable:
            resp = rd_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert resp['Id'] == dev_id
            assert resp['OrgKey'] == self.ORGKEY
            assert resp['GroupChangeTime'] is not None
            assert resp['SensorConfigUpdateTime'] is not None
            assert resp['SensorCertificateUpdateTime'] is not None
            assert resp['HashDeleteRequestTime'] is not None
            assert resp['HashUploadRequestTime'] is not None
            assert resp['LogsRequestTime'] is not None
            assert resp['SensorKitUpdateTime'] is not None
            assert resp['PerfmonTriggerTime'] is not None
            assert resp['SensorActionTime'] is not None
            assert resp['ReRegisterTime'] is not None
            assert resp['LiveQueryChangeTime'] is not None
        else:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01c_update_device_sequential(self, hc_connect, rd_access):
        """
        Verify that we can update a device multiple times with varied data.

        :class: functional
        """
        data = Device.full()
        the_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=the_id)
        self.verify_response(hc_connect, 200)

        if not rd_access.usable:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

        fields = ['GroupChangeTime', 'SensorConfigUpdateTime', 'SensorCertificateUpdateTime', 'HashDeleteRequestTime',
                  'HashUploadRequestTime', 'LogsRequestTime', 'SensorKitUpdateTime', 'PerfmonTriggerTime',
                  'SensorActionTime', 'ReRegisterTime', 'LiveQueryChangeTime']
        prev = {}
        for x in fields:
            prev[x] = None

        with self.Timer(tag="10 Device Updates"):
            for _ in range(10):
                data.live_query_time = self.stamp()  # update live query, since we are specifying it as a change
                _logger.debug(f"USE: {data.to_dict()}")
                hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=the_id)
                self.verify_response(hc_connect, 200)

                if rd_access.usable:
                    resp = rd_access.get_device(the_id, self.ORGKEY)
                    _logger.info(f">>> Returned: {resp}")
                    assert resp['Id'] == the_id
                    assert resp['OrgKey'] == self.ORGKEY

                    for x in fields:
                        assert resp[x] != prev[x], f"failed with {x}"
                        prev[x] = resp[x]
                time.sleep(0.1)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02a_update_devicegroup(self, hc_connect, pg_access):
        """
        Verify that we can update a device group with minimal required information.

        :class: functional
        """
        data = DeviceGroup(changes=[DeviceGroupChangeType.DG_POLICY_CHANGE])
        grp_id = self.uid()
        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] is None
            assert check[pg_access.PgGrp.last_policy_change_time] is not None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02b_update_devicegroup_full(self, hc_connect, pg_access, rd_access):
        """
        Verify that we can update a device group that has been fully defined.

        :class: functional
        """
        data = DeviceGroup.full()
        grp_id = self.uid()

        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] is not None
            assert check[pg_access.PgGrp.last_policy_change_time] is not None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

        if rd_access.usable:
            resp = rd_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert resp['Id'] == grp_id
            assert resp['OrgKey'] == self.ORGKEY
            assert resp['LiveQueryChangeTime'] is not None
            assert resp['LastPolicyChangeTime'] is not None
        else:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02c_update_devicegroup_sequential(self, hc_connect, rd_access):
        """
        Verify that we can update a device group multiple times with varied data.

        :class: functional
        """
        data = DeviceGroup.full()
        the_id = self.uid()

        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=the_id)
        self.verify_response(hc_connect, 200)

        if not rd_access.usable:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

        with self.Timer(tag="10 DeviceGroup Updates"):
            last_lqt = None
            last_pct = None
            for _ in range(10):
                data.live_query_time = self.stamp()  # update live query, since we are specifying it
                _logger.debug(f"USE: {data.to_dict()}")
                hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=the_id)
                self.verify_response(hc_connect, 200)

                if rd_access.usable:
                    resp = rd_access.get_group(the_id, self.ORGKEY)
                    _logger.info(f">>> Returned: {resp}")
                    assert resp['Id'] == the_id
                    assert resp['OrgKey'] == self.ORGKEY
                    assert resp['LiveQueryChangeTime'] != last_lqt
                    assert resp['LastPolicyChangeTime'] != last_pct
                    last_lqt = resp['LiveQueryChangeTime']
                    last_pct = resp['LastPolicyChangeTime']
                time.sleep(0.1)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_update_organization(self, hc_connect, pg_access):
        """
        Verify that we can update an organization with minimal required information (i.e. at least one change)

        :class: functional
        """
        data = Org(is_kill_switch_on=True)
        orgkey = self.new_orgkey("enki")

        hc_connect.v1_org_update(data, org_key=orgkey)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(orgkey)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.org_key] == orgkey
            assert check[pg_access.PgOrg.is_kill_switch_on]
            assert check[pg_access.PgOrg.change_timestamps] is None
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is None
            assert check[pg_access.PgOrg.canary_payload_timestamps] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03b_update_organization_full(self, hc_connect, pg_access, rd_access):
        """
        Verify that we can update an organization that has been fully defined.

        :class: functional
        """
        data = Org.full()
        orgkey = self.new_orgkey("enki")

        hc_connect.v1_org_update(data, org_key=orgkey)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(orgkey)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.org_key] == orgkey
            assert check[pg_access.PgOrg.is_kill_switch_on]
            assert check[pg_access.PgOrg.change_timestamps] is not None
            assert len(check[pg_access.PgOrg.change_timestamps]) == len(OrgChangeType.all())
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is not None
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == len(DeploymentType.all())
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.ENDPOINT.name]) == len(
                OrgChangeType.all())
            assert check[pg_access.PgOrg.canary_payload_timestamps] is not None
            assert len(check[pg_access.PgOrg.canary_payload_timestamps]) == len(DeviceType.all())
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

        if rd_access.usable:
            resp = rd_access.get_org(orgkey)
            _logger.info(f">>> Returned: {resp}")
            assert resp['OrgKey'] == orgkey
            assert resp['IsKillSwitchOn']['Bool']
            for x in DeviceType.all(as_string=True):
                assert resp['CanaryPayloadTimestamps'][x] is not None
            for x in OrgChangeType.all(as_value=True):
                assert resp['ChangeTimestamps'][x] is not None
            for x in DeploymentType.all(as_string=True):
                for y in OrgChangeType.all(as_value=True):
                    assert resp['DeploymentTypeChangeTimestamps'][x][y] is not None
        else:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03c_update_organization_sequential(self, hc_connect, rd_access):
        """
        Verify that we can update an organization multiple times with varied data.

        :class: functional
        """
        data = Org.full()
        the_id = self.uid()
        the_key = self.new_orgkey("enki", the_id)

        hc_connect.v1_org_update(data, org_key=the_key)
        self.verify_response(hc_connect, 200)

        if not rd_access.usable:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

        prev_cpt = {}
        for x in DeviceType.all(as_string=True):
            prev_cpt[x] = None
        prev_ct = {}
        for x in OrgChangeType.all(as_value=True):
            prev_ct[x] = None
        prev_dtct = {}
        for x in DeploymentType.all(as_string=True):
            for y in OrgChangeType.all(as_value=True):
                prev_dtct[f"{x}|{y}"] = None

        with self.Timer(tag="10 Organization Updates"):
            for _ in range(10):
                data.live_query_time = self.stamp()  # update live query, since we are specifying it
                _logger.debug(f"USE: {data.to_dict()}")
                hc_connect.v1_org_update(data, org_key=the_key)
                self.verify_response(hc_connect, 200)

                if rd_access.usable:
                    resp = rd_access.get_org(the_key)
                    _logger.info(f">>> Returned: {resp}")
                    assert resp['OrgKey'] == the_key
                    assert resp['IsKillSwitchOn']['Bool']
                    for x in DeviceType.all(as_string=True):
                        assert resp['CanaryPayloadTimestamps'][x] != prev_cpt[x], f"failed on {x}"
                        prev_cpt[x] = resp['CanaryPayloadTimestamps'][x]
                    for x in OrgChangeType.all(as_value=True):
                        assert resp['ChangeTimestamps'][x] != prev_ct[x], f"failed on {x}"
                        prev_ct[x] = resp['ChangeTimestamps'][x]
                    for x in DeploymentType.all(as_string=True):
                        for y in OrgChangeType.all(as_value=True):
                            assert resp['DeploymentTypeChangeTimestamps'][x][y] != prev_dtct[f"{x}|{y}"], f"failed on {x}|{y}"
                            prev_dtct[f"{x}|{y}"] = resp['DeploymentTypeChangeTimestamps'][x][y]
                time.sleep(0.1)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_update_environment(self, hc_connect, pg_access):
        """
        Verify that we can update an environment with minimal required information.

        :class: functional
        """
        data = Env(policy_service_time_changed=True)
        name = f"QA{self.uid()}"

        hc_connect.v1_env_update(data, env_name=name)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert check[pg_access.PgEnv.name] == name
            assert check[pg_access.PgEnv.cert_service_time] is None
            assert check[pg_access.PgEnv.policy_service_update_time] is not None
            assert check[pg_access.PgEnv.canary_payload_timestamps] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04b_update_environment_full(self, hc_connect, pg_access, rd_access):
        """
        Verify that we can update an environment that has been fully defined.

        :class: functional
        """
        data = Env.full()
        name = f"QA{self.uid()}"

        hc_connect.v1_env_update(data, env_name=name)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert check[pg_access.PgEnv.name] == name
            assert check[pg_access.PgEnv.cert_service_time] is not None
            assert check[pg_access.PgEnv.policy_service_update_time] is not None
            assert check[pg_access.PgEnv.canary_payload_timestamps] is not None
            assert len(check[pg_access.PgEnv.canary_payload_timestamps]) == len(DeviceType.all())
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

        if rd_access.usable:
            resp = rd_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert resp['Name'] == name
            assert resp['CertServiceTime'] is not None
            assert resp['PolicyServiceUpdateTime'] is not None
            for x in DeviceType.all(as_string=True):
                assert resp['CanaryPayloadTimestamps'][x] is not None
        else:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04c_update_environment_sequential(self, hc_connect, rd_access):
        """
        Verify that we can update an organization multiple times with varied data.

        :class: functional
        """
        name = f"QA{self.uid()}"
        data = Env.full()

        hc_connect.v1_env_update(data, env_name=name)
        self.verify_response(hc_connect, 200)

        prev_cpt = {}
        for x in DeviceType.all(as_string=True):
            prev_cpt[x] = None
        prev_cst = None
        prev_psut = None

        if not rd_access.usable:
            _logger.warning("NOTE: No access to redis, cannot ensure content!")

        with self.Timer(tag="10 Environment Updates"):
            for _ in range(10):
                _logger.debug(f"USE: {data.to_dict()}")
                hc_connect.v1_env_update(data, env_name=name)
                self.verify_response(hc_connect, 200)

                if rd_access.usable:
                    resp = rd_access.get_env(name)
                    _logger.info(f">>> Returned: {resp}")
                    assert resp['Name'] == name
                    assert resp['CertServiceTime'] != prev_cst
                    prev_cst = resp['CertServiceTime']
                    assert resp['PolicyServiceUpdateTime'] != prev_psut
                    prev_psut = resp['PolicyServiceUpdateTime']
                    for x in DeviceType.all(as_string=True):
                        assert resp['CanaryPayloadTimestamps'][x] != prev_cpt[x]
                        prev_cpt[x] = resp['CanaryPayloadTimestamps'][x]
                time.sleep(0.1)
