#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging

import pytest
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.device import Device
from lib.models.device_group import DeviceGroup
from lib.models.enumerations import HintType, DeviceChangeType, DeviceGroupChangeType, OrgChangeType, \
    DeviceType, DeploymentType
from lib.models.environment import Env
from lib.models.hinting import HintRequest
from lib.models.organization import Org
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


class TestHintingApi(TestCommon):
    """
    This class tests the hint generator endpoints in cooperation with the configuration.
    """

    DEV = Device(changes=[DeviceChangeType.DEV_SENSOR_KIT_UPDATE])
    GRP = DeviceGroup(changes=[DeviceGroupChangeType.DG_LIVE_QUERY_CHANGE])
    ORG = Org(is_kill_switch_on=True)
    ORG1 = Org(changes=OrgChangeType.all())
    ORG2 = Org(canary_payload_device_types=DeviceType.all())
    ORG3 = Org(deployment_type_changes=DeploymentType.all_dict_org())
    ENV = Env(cert_service_time_changed=True)

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01a_post_v1_hints(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent.
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01b_post_v1_hints_up_to_date(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent.
        This test simplifies things such that device, group, org and env timestamps are all older
        than the device's latest times (i.e. device is up to date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=info.done_time, last_update_time=info.done_time,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert not response.last_update_time 
        assert not response.last_live_query_time 
        assert response.hints == set()

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01c_post_v1_hints_with_audit(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent along with an audit
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        HintType.ONETIME_HASH_DELETE_LIST,
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER,
                                        HintType.REP_REQUEST},
                           audit_last_live_query_time=info.done_time, audit_last_update_time=info.done_time)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01d_post_v1_hints_with_audit_missing(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent along with an audit (items missing).
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        HintType.ONETIME_HASH_DELETE_LIST,
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER,
                                        HintType.APC_UPLOAD,  # << will be missing in logs
                                        },
                           audit_last_live_query_time=info.done_time, audit_last_update_time=info.done_time)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01e_post_v1_hints_with_audit_extra(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent along with an audit (items missing).
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        # HintType.ONETIME_HASH_DELETE_LIST,  # Will be extra in logs
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER,
                                        },
                           audit_last_live_query_time=info.done_time, audit_last_update_time=info.done_time)
        # HintType.ONETIME_HASH_DELETE_LIST removed

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01f_post_v1_hints_with_audit_live_query_time(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent along with an audit (live query differs).
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        HintType.ONETIME_HASH_DELETE_LIST,
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER,
                                        },
                           audit_last_live_query_time=self.YESTERDAY, audit_last_update_time=info.done_time)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER, 
            HintType.REP_REQUEST
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01g_post_v1_hints_with_audit_update_time(self, hc_connect):
        """
        Verify that POST /v1/hints works as expected when basic data is sent along with an audit (update differs).
        This test simplifies things such that device, group, org and env timestamps are all newer
        than the device's latest times (i.e. device was out of date).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.YESTERDAY,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        HintType.ONETIME_HASH_DELETE_LIST,
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER,
                                        },
                           audit_last_live_query_time=info.done_time, audit_last_update_time=self.YESTERDAY)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    # NOTE: Older "proof" tests attempted to use hinting to determine if device or org changes to the database
    #       (specifivally those columns that are json blobs) merge or replace.  This testing is now more
    #       directly done with postgres access and these older tests gave been removed.

    # TODO: test_02a_post_v1_hints_missing_device_type
    # TODO: test_02b_post_v1_hints_missing_deployment_type
