#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging
import os
import pkgutil
import tempfile
from logging.config import fileConfig
from typing import Optional

import pytest
from _pytest.config import Config
from _pytest.nodes import Item
from _pytest.runner import CallInfo
from filelock import FileLock

from lib.configuration import FunctionalTestConfig
from lib.connect.hint_connect import HintControlConnect
from lib.postgres import PostgresDB
from lib.redis import RedisDB
from lib.utilities import Artifacts, Setup

################################################################################
# Testing Definitions -- common definitions for all testing
#

################################################################################

_logger = logging.getLogger(__name__)

# Define basic logging configuration
try:
    logini = pkgutil.get_data(__name__, "logging.ini")
    with tempfile.NamedTemporaryFile() as tf:
        tf.write(logini)
        tf.flush()
        logging.config.fileConfig(tf.name, disable_existing_loggers=False)
except Exception as err:
    _logger.warning(f"Could not load logging info: {err}")


class ProcessWatcher(object):
    """
    A workaround class that uses a file to maintain a running count of xdist processes as they begin and end
    (as determined by our session logic).  Knowing when the first process starts (which lock out the others until it
    is done), or the last process is ending (so the metrics and cleanup is only done once) is handled with this class.
    """
    _lock_filename = "/tmp/PROCESS_WATCHER_FILE_LOCK"
    _tracker_filename = "/tmp/PROCESS_WATCHER_DATA"

    def __init__(self):
        pass

    class _ProcessWatcherResult(object):
        def __init__(self, value: int):
            self._value = value

        @property
        def value(self) -> int:
            return self._value

    class _StartedResult(_ProcessWatcherResult):
        # noinspection PyUnusedFunction
        @property
        def is_first_in(self) -> bool:
            """Return True if process watch after start is 1 (first through the gate)."""
            return self._value == 1

    class _EndedResult(_ProcessWatcherResult):
        @property
        def is_last_out(self) -> bool:
            """Return True if process watch after end is 0 (last through the gate)."""
            return self._value == 0

    def clear(self):
        with FileLock(self._lock_filename):
            open(self._tracker_filename, 'w').close()

    def _create_file(self):
        open(self._tracker_filename, 'a').close()

    def _read_value(self) -> int:
        self._create_file()
        with open(self._tracker_filename, 'r+') as tracker_file:
            return int(tracker_file.read().strip() or 0)

    def _write_value(self, value: int):
        with open(self._tracker_filename, 'w') as tracker_file:
            tracker_file.write(str(value))

    def process_start(self) -> _StartedResult:
        with FileLock(self._lock_filename):
            result = self._StartedResult(self._read_value() + 1)
            self._write_value(result.value)
            return result

    def process_end(self) -> _EndedResult:
        with FileLock(self._lock_filename):
            result = self._EndedResult(self._read_value() - 1)
            if result.value < 0:
                _logger.warning('ProcessTracker ran into value less than 0.')
            self._write_value(result.value)
            return result


def _init_once_only_per_test_run(pt_config: Config) -> None:
    """
    Preliminary environment setup that should only be performed once, whether single-process or when using xdist.

    :param pt_config: pytest configuration
    """
    # clear test run aritifacts
    artifacts = Artifacts()
    artifacts.remove_all()

    Setup().normalize_logging(_logger, debug=pt_config.option.dbg)

    # Initialialize error output storage

    artifacts.create(artifacts.RUN_FAILURES, replace=True)

    ft_config = FunctionalTestConfig()
    Setup.process_automation_options(pt_config, ft_config)

    # initialize process tracker
    ProcessWatcher().clear()


def _is_running_xdist(config: Config) -> bool:
    """
    Check set of arguments passed into pytest for the -n or --numprocesses options (signifying that xdist is
    being used).

    :param config: pytest configuration
    :return: True if -n option used
    """
    look_for = ['-n', '--numprocesses']
    check = [str(x).lower() in look_for for x in config.invocation_params.args]
    return True in check


# noinspection PyUnusedLocal
def pytest_xdist_setupnodes(config: Config, specs) -> None:
    """
    Xdist hook called before any remote nodes are set up.

    :param config: pytest configuration
    :param specs: i/o specs (ignored)
    """
    if _is_running_xdist(config):  # do pre-initialization, won't be done in pytest_cmdline_main
        _init_once_only_per_test_run(config)


def pytest_cmdline_main(config: Config) -> None:
    """
    Pytest hook called when processing the command line, before other hooks.

    :param config: pytest configuration
    """
    if not _is_running_xdist(config):  # no xdist, so initialization must be called here
        _init_once_only_per_test_run(config)


# ----- pytest hooks ------------------------------------------------------------

def pytest_addoption(parser) -> None:
    """
    Function to define acceptable command line options.

    :param parser: parser object containing global options
    """
    try:
        # service settings
        parser.addoption("--ip", action="store", default=None,
                         help="ip for service")
        parser.addoption("--port", action="store", default=None, type=int,
                         help="port for service")

        # postgres overrides
        parser.addoption("--pghost", action="store", default=None,
                         help="Alternate host for postgres")
        parser.addoption("--pgport", action="store", default=None, type=int,
                         help="Alternate port for postgres")
        parser.addoption("--pguser", action="store", default=None,
                         help="Alternate user for postgres")
        parser.addoption("--pgpass", action="store", default=None,
                         help="Alternate password for postgres")
        parser.addoption("--dbname", action="store", default=None,
                         help="Alternate database name for postgres")

        # redis overrides
        parser.addoption("--rhost", action="store", default=None,
                         help="Alternate host for redis")
        # parser.addoption("--pgport", action="store", default=None, type=int,
        #                  help="Alternate port for postgres")
        parser.addoption("--rpass", action="store", default=None,
                         help="Alternate password for redis")

        # Special controls
        parser.addoption("--dbg", action="store_true", default=False,
                         help="Change local logging to DEBUG")
        parser.addoption("--envfile", action="store", default=None,
                         help="Path of local envfile, for data scraping purposes")
    except Exception as err2:
        _logger.warning(f"Failure to add options: {err2}")


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_protocol(item: Item, nextitem: Optional[Item]) -> Optional[object]:
    """
    FUTURE: Reserved for specialized handling when tests are run.

    :param item: current test object
    :param nextitem: next test object
    :return: test object
    """
    yield  # do all other pytest_sessionstart methods first, then use ours

    if nextitem is None:
        _logger.info("\n" + "=" * 80 + "\n" + f"  Testing Completed!")


@pytest.hookimpl(hookwrapper=True)
def pytest_sessionstart(session: pytest.Session) -> None:
    """
    Initialize the test environment for the test session.

    :param session: the current test session
    """
    yield  # do all other pytest_sessionstart methods first, then use ours

    _logger.info(
        "\n" + "=" * 80 + "\n" + f"  Starting pytest with {session.config.invocation_params.args}\n" + "=" * 80)

    procname = f"{session.config.workerinput['workerid']}" if hasattr(session.config, "workerinput") else "controller"

    with FileLock("/tmp/pytest_sessionstart.lock"):
        try:
            ft_config = FunctionalTestConfig()
            Setup().process_automation_options(session.config, ft_config)

            if ProcessWatcher().process_start().is_first_in:
                _logger.debug(f"<< PID for {procname}")  # add to log for easy match
                Setup().finalize_working_environment(session.config, ft_config)
                _logger.info(Setup().environment_banner())
            else:
                pass  # FUTURE: reserved for future needs
        except Exception as err2:
            _logger.error(f'Automation Configuration failed!\n  {err2}', exc_info=True)
            pytest.exit(f"{err2}", returncode=1)

        # track worker process count
        worker_count = os.getenv("PYTEST_XDIST_WORKER_COUNT")
        if worker_count is not None:
            FunctionalTestConfig().PROCESSES = worker_count


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_setup(item: pytest.Item) -> None:
    """
    FUTURE: Called to run the setup for test item (the call phase).

    The default implementation calls ``item.runtest()``.
    """
    outcome = yield
    result = outcome.get_result()

    # noinspection PyUnresolvedReferences
    message = f"START {item.cls.__name__} >> {item.name}"
    _logger.info(f":::::[ {message} ]:::::")


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item: pytest.Item) -> None:
    """
    FUTURE: Called to run the test for test item (the call phase).

    The default implementation calls ``item.runtest()``.
    """
    outcome = yield
    result = outcome.get_result()

    markers = [mk.name for mk in item.own_markers]


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item: Item, call: CallInfo) -> None:
    """
    Add test failures to a simplified file.

    {note}
    Helpful hint from https://www.codementor.io/@adammertz/writing-a-simple-pytest-hook-zc5wvoj5t
    {end}
    :param item: pytest item
    :param call: pytest call
    """
    # All code prior to yield statement would be run prior
    # to any other of the same fixtures defined

    outcome = yield  # Run all other pytest_runtest_makereport non wrapped hooks
    result = outcome.get_result()

    artifacts = Artifacts()
    artifacts.create(artifacts.TEST_TIMING, f"{item.nodeid}||{result.when}||{result.duration:.3f}",
                     separate=False)

    if result.when == "call":
        if result.failed:
            # critical tests prevent reruns
            markers = [mk.name for mk in item.own_markers]

            rerun = ""
            # noinspection PyUnresolvedReferences
            if hasattr(item, "execution_count") and item.execution_count > 1:
                rerun = f", rerun {item.execution_count - 1}"

            header = f":::::[ {item.nodeid}{rerun} ]:::::\n"
            artifacts.create(artifacts.RUN_FAILURES, f"{header}{result.longreprtext}")

    else:  # still want failure info on setup or teardown!
        if result.failed:
            artifacts.create(artifacts.RUN_FAILURES, f">>>{result.when}<<<\n{result.longreprtext}")


# noinspection PyUnusedLocal
@pytest.hookimpl(hookwrapper=True)
def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:
    """
    FUTURE: This pytest hook is run after the entire pytest test run, and only once even if using xdist.

    :param session: current pytest session
    :param exitstatus: current session exit status
    """
    outcome = yield

    procname = f"{session.config.workerinput['workerid']}" if hasattr(session.config, "workerinput") else "controller"

    # submit our individual report entries
    with FileLock("/tmp/pytest_sessionfinish.lock"):
        if ProcessWatcher().process_end().is_last_out:
            # FUTURE: any metrics gathering
            _logger.info(f"Exit status: {exitstatus}")
        else:
            pass  # FUTURE: retained for futre needs


# ----- Common fixtures

# noinspection PyUnusedFunction
@pytest.fixture(autouse=True, scope="session")
def ft_config() -> FunctionalTestConfig:
    """
    Fixture to return the current FunctionalTestConfig object for the process.

    :return: FunctionalTestConfig object
    """
    ft_config = FunctionalTestConfig()
    yield ft_config


# noinspection PyUnusedFunction
@pytest.fixture(autouse=True, scope="session")
def hc_connect() -> HintControlConnect:
    """
    Fixture to return an API connection object usable by all tests.

    :return: Connect object
    """
    yield FunctionalTestConfig().CONNECTION


# noinspection PyUnusedFunction
@pytest.fixture(autouse=True, scope="session")
def pg_access() -> PostgresDB:
    """
    Fixture to return an object allowing for access to the postgres database.

    :return: PostgresDB object
    """
    yield FunctionalTestConfig().POSTGRES


# noinspection PyUnusedFunction
@pytest.fixture(autouse=True, scope="session")
def rd_access() -> RedisDB:
    """
    Fixture to return an object allowing for access to the redis database.

    :return: RedisDB object
    """
    yield FunctionalTestConfig().REDIS
