#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging

import pytest
from psc_orgkey import encode_orgkey
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.device_group import DeviceGroup
from lib.models.enumerations import DeviceGroupChangeType
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestDeviceGroupsExtensive(TestCommon):
    """
    This class tests devicegroup endpoints beyond normal "happy path" tests.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01a_v1_devicegroup_update_uri_orgkey_emptystring(self, hc_connect):
        """
        Ensure group update with empty string orgkey is caught

        :class: functional, negative
        """
        data = DeviceGroup.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_devicegroup_update(data, org_key="", group_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01b_v1_devicegroup_update_uri_orgkey_invalid(self, hc_connect):
        """
        Ensure group update with an invalid orgkey (cannot decode) is caught

        :class: functional, negative
        """
        data = DeviceGroup.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_devicegroup_update(data, org_key="WONT_DECODE", group_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01c_v1_devicegroup_update_uri_orgkey_unknown(self, hc_connect):
        """
        Ensure group update with a valid orgkey (but env does not exist) is allowed

        :class: functional
        """
        data = DeviceGroup.full()
        key = encode_orgkey("nosuch_env", self.uid())

        hc_connect.v1_devicegroup_update(data, org_key=key, group_id=self.uid())
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02a_v1_devicegroup_update_uri_id_emptystring(self, hc_connect):
        """
        Ensure group update with empty string id is caught

        :class: functional, negative
        """
        data = DeviceGroup.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id="")
        self.verify_response(hc_connect, 404, "404 page not found")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02b_v1_devicegroup_update_uri_id_non_integer(self, hc_connect):
        """
        Ensure group update with non-integer id is caught

        :class: functional, negative
        """
        data = DeviceGroup.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id="twenty-three")
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_v1_devicegroup_update_body_empty_string(self, hc_connect):
        """
        Ensure group update with empty string body data is caught

        :class: functional, negative
        """
        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_devicegroup_update(None, org_key=self.ORGKEY, group_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03b_v1_devicegroup_update_body_empty_structure(self, hc_connect, pg_access):
        """
        Ensure group update with empty body data is allowed

        :class: functional, negative
        """
        grp_id = self.uid()
        hc_connect.v1_devicegroup_update({}, org_key=self.ORGKEY, group_id=grp_id)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] is None
            assert check[pg_access.PgGrp.last_policy_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_v1_devicegroup_update_body_invalid_changes(self, hc_connect, pg_access):
        """
        Ensure group update with valid group change type not used for hinting is flagged as an error.

        :class: functional
        """
        data = DeviceGroup.full().to_dict()
        data['changes'] = ["DG_DISABLE_SENSOR_AUTOUPDATE"]
        grp_id = self.uid()

        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04b_v1_devicegroup_update_body_duplicate_changes(self, hc_connect, pg_access):
        """
        Ensure group update with duplicate change entries has no ill effects

        :class: functional, negative
        """
        grp_id = self.uid()
        data = DeviceGroup(changes=[DeviceGroupChangeType.DG_LIVE_QUERY_CHANGE, DeviceGroupChangeType.DG_LIVE_QUERY_CHANGE])
        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] is not None
            assert check[pg_access.PgGrp.last_policy_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05a_v1_devicegroup_update_body_live_query_override(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value overrides "now" if specified when doing DG_LIVE_QUERY_CHANGE

        :class: functional, negative
        """
        grp_id = self.uid()
        data = DeviceGroup(changes=[DeviceGroupChangeType.DG_LIVE_QUERY_CHANGE], live_query_time=self.pinned())
        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] == self.pinned()
            assert check[pg_access.PgGrp.last_policy_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05b_v1_devicegroup_update_body_live_query_override_not_live_query(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value does not overrides "now" if specified when not doing DG_LIVE_QUERY_CHANGE

        :class: functional, negative
        """
        grp_id = self.uid()
        data = DeviceGroup(changes=[DeviceGroupChangeType.DG_POLICY_CHANGE], live_query_time=self.pinned())
        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert check[pg_access.PgGrp.live_query_change_time] is None
            assert check[pg_access.PgGrp.last_policy_change_time] != self.pinned()
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05c_v1_devicegroup_update_body_default_live_query(self, hc_connect, pg_access):
        """
        Ensure "now" is used if live_query_time not specifiued when doing DG_LIVE_QUERY_CHANGE

        :class: functional, negative
        """
        grp_id = self.uid()
        data = DeviceGroup(changes=[DeviceGroupChangeType.DG_LIVE_QUERY_CHANGE])
        hc_connect.v1_devicegroup_update(data, org_key=self.ORGKEY, group_id=grp_id)

        if pg_access.usable:
            resp = pg_access.get_group(grp_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find group"
            check = resp[0]
            assert check[pg_access.PgGrp.id] == grp_id
            assert check[pg_access.PgGrp.org_key] == self.ORGKEY
            assert self.is_now(check[pg_access.PgGrp.live_query_change_time])
            assert check[pg_access.PgGrp.last_policy_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")
