#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging

import pytest
from psc_orgkey import encode_orgkey
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.device import Device
from lib.models.enumerations import DeviceChangeType
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestDevicesExtensive(TestCommon):
    """
    This class tests device endpoints beyond normal "happy path" tests.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01a_v1_device_update_uri_orgkey_emptystring(self, hc_connect):
        """
        Ensure device update with empty string orgkey is caught

        :class: functional, negative
        """
        data = Device.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_device_update(data, org_key="", device_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01b_v1_device_update_uri_orgkey_invalid(self, hc_connect):
        """
        Ensure device update with an invalid orgkey (cannot decode) is caught

        :class: functional, negative
        """
        data = Device.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_device_update(data, org_key="WONT_DECODE", device_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01c_v1_device_update_uri_orgkey_unknown(self, hc_connect, pg_access):
        """
        Ensure device update with a valid orgkey (but env & org_id represented do not exist) is allowed

        :class: functional
        """
        data = Device.full()
        dev = self.uid()
        key = encode_orgkey("nosuch_env", self.uid())

        hc_connect.v1_device_update(data, org_key=key, device_id=dev)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_device(dev, key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev
            assert check[pg_access.PgDev.org_key] == key
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02a_v1_device_update_uri_id_emptystring(self, hc_connect):
        """
        Ensure device update with empty string id is caught

        :class: functional, negative
        """
        data = Device.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id="")
        self.verify_response(hc_connect, 404, "404 page not found")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02b_v1_device_update_uri_id_non_integer(self, hc_connect):
        """
        Ensure device update with non-integer id is caught

        :class: functional, negative
        """
        data = Device.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id="twenty-three")
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_v1_device_update_body_empty_string(self, hc_connect):
        """
        Ensure device update with empty string data body is caught

        :class: functional, negative
        """
        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_device_update(None, org_key=self.ORGKEY, device_id=self.uid())
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03b_v1_device_update_body_empty_structure(self, hc_connect, pg_access):
        """
        Ensure device update with empty body is allowed.

        :class: functional
        """
        dev = self.uid()
        hc_connect.v1_device_update({}, org_key=self.ORGKEY, device_id=dev)

        if pg_access.usable:
            resp = pg_access.get_device(dev, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.group_change_time] is None
            assert check[pg_access.PgDev.sensor_config_update_time] is None
            assert check[pg_access.PgDev.sensor_certificate_update_time] is None
            assert check[pg_access.PgDev.hash_delete_request_time] is None
            assert check[pg_access.PgDev.hash_upload_request_time] is None
            assert check[pg_access.PgDev.logs_request_time] is None
            assert check[pg_access.PgDev.sensor_kit_update_time] is None
            assert check[pg_access.PgDev.perfmon_trigger_time] is None
            assert check[pg_access.PgDev.sensor_action_time] is None
            assert check[pg_access.PgDev.re_register_time] is None
            assert check[pg_access.PgDev.live_query_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    # NOTE: Tests after this will only be concerned with pg_access fields of interest, not all of them

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04_v1_device_update_body_invalid_changes(self, hc_connect):
        """
        Ensure device update with valid device change type not used for hinting is ignored.

        :class: functional
        """
        data = Device.full().to_dict()
        data['changes'] = ["DEV_DASHBOARD_DEREGISTER"]
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05_v1_device_update_body_duplicate_changes(self, hc_connect, pg_access):
        """
        Ensure device update with duplicate device change type has no ill effects.

        :class: functional
        """
        data = Device(changes=[DeviceChangeType.DEV_REREGISTER, DeviceChangeType.DEV_REREGISTER])
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.re_register_time] is not None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06a_v1_device_update_body_live_query_override(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value overrides "now" if specified when doing DEV_LIVE_QUERY_CHANGE

        :class: functional
        """
        data = Device(changes=[DeviceChangeType.DEV_LIVE_QUERY_CHANGE], live_query_time=self.pinned())
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.live_query_change_time] == self.pinned()
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06b_v1_device_update_body_live_query_override_not_live_query(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value does not override "now" if specified when not doing DEV_LIVE_QUERY_CHANGE

        :class: functional
        """
        data = Device(changes=[DeviceChangeType.DEV_REREGISTER], live_query_time=self.pinned())
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.re_register_time] != self.pinned()
            assert check[pg_access.PgDev.live_query_change_time] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06c_v1_device_update_body_live_query_default(self, hc_connect, pg_access):
        """
        Ensure no livequery value uses "now" if specified when doing DEV_LIVE_QUERY_CHANGE

        :class: functional
        """
        data = Device(changes=[DeviceChangeType.DEV_LIVE_QUERY_CHANGE])
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert self.is_now(check[pg_access.PgDev.live_query_change_time])
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    # ----- Regression tests

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_10_dser_46240_bad_change_prevents_valid_change(self, hc_connect, pg_access):
        """
        DSER-46240:
        At present all models do a validation check of any change strings, returning a 400 error and adding a log
        message if a supplied entry is unknown.  While well intentioned, this prevents the use of the other valid
        change types that were supplied.  This is especially bad for organizations which has 4 potential points that
        can prevent ANY organization data if any one point has "bad" data.

        NOTE: Old bug, likely moot.  Test added to verify no ill effects

        :class: functional
        """
        data = Device.full().to_dict()
        data['changes'] = ["DEV_LIVE_QUERY_CHANGE", "DEV_NO_SUCH_CHANGE"]
        dev_id = self.uid()

        hc_connect.v1_device_update(data, org_key=self.ORGKEY, device_id=dev_id)

        if pg_access.usable:
            resp = pg_access.get_device(dev_id, self.ORGKEY)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find device"
            check = resp[0]
            assert check[pg_access.PgDev.id] == dev_id
            assert check[pg_access.PgDev.org_key] == self.ORGKEY
            assert check[pg_access.PgDev.live_query_change_time] is not None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")
