#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging

import pytest
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.enumerations import DeviceType
from lib.models.environment import Env
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestEnvironmentsExtensive(TestCommon):
    """
    This class tests environments endpoints beyond normal "happy path" tests.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01_v1_env_update_uri_name_emptystring(self, hc_connect):
        """
        Ensure env update with empty string name is caught

        :class: functional, negative
        """
        data = Env.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_env_update(data, env_name="")
        self.verify_response(hc_connect, 404, "404 page not found")

    def test_02a_v1_env_update_ensure_additive_union(self, hc_connect, pg_access):
        """
        Ensure env update with two different canary_payload_device_types definitions results
        in the UNION of the two, not a replacement.

        :class: functional
        """
        name = f"enki{self.uid()}"
        data1 = Env(canary_payload_device_types=[DeviceType.WINDOWS])
        data2 = Env(canary_payload_device_types=[DeviceType.IOS])

        hc_connect.v1_env_update(data1, env_name=name)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_env_update(data2, env_name=name)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert len(check[pg_access.PgEnv.canary_payload_timestamps]) == 2
            assert DeviceType.WINDOWS.name in check[pg_access.PgEnv.canary_payload_timestamps] and DeviceType.IOS.name in check[pg_access.PgEnv.canary_payload_timestamps]
            assert check[pg_access.PgEnv.canary_payload_timestamps][DeviceType.WINDOWS.name] != check[3][DeviceType.IOS.name]

    def test_02b_v1_env_update_ensure_subtractive_update(self, hc_connect, pg_access):
        """
        Ensure env that is fully populated and then updated with a single canary_payload_device_types definition results
        in only in the UPDATE of new data, not a replacement.

        :class: functional
        """
        name = f"enki{self.uid()}"
        data1 = Env.full()
        data2 = Env(canary_payload_device_types=[DeviceType.IOS])

        hc_connect.v1_env_update(data1, env_name=name)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_env_update(data2, env_name=name)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert len(check[pg_access.PgEnv.canary_payload_timestamps]) == len(DeviceType.all())
            assert check[pg_access.PgEnv.canary_payload_timestamps][DeviceType.WINDOWS.name] != check[pg_access.PgEnv.canary_payload_timestamps][DeviceType.IOS.name]
            assert check[pg_access.PgEnv.canary_payload_timestamps][DeviceType.WINDOWS.name] == check[pg_access.PgEnv.canary_payload_timestamps][DeviceType.MAC.name]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_v1_env_update_uri_body_missing(self, hc_connect):
        """
        Ensure env update with missing body is caught

        :class: functional, negative
        """
        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_env_update(None, env_name=f"enki{self.uid()}")
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_v1_env_update_uri_body_empty(self, hc_connect, pg_access):
        """
        Ensure env update with empty body is allowed

        :class: functional, negative
        """
        name = f"enki{self.uid()}"
        hc_connect.v1_env_update({}, env_name=name)

        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert check[pg_access.PgEnv.name] == name
            assert check[pg_access.PgEnv.cert_service_time] is None
            assert check[pg_access.PgEnv.policy_service_update_time] is None
            assert check[pg_access.PgEnv.canary_payload_timestamps] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_v1_env_update_body_invalid_canary(self, hc_connect):
        """
        Ensure env update with invalid device type used for canary_payload_device_types is caught

        :class: functional
        """
        data = Env.full().to_dict()
        data['canary_payload_device_types'] = ["SAMSUNG"]

        hc_connect.v1_env_update(data, env_name=f"enki{self.uid()}")
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_v1_env_update_body_duplicate_canary(self, hc_connect, pg_access):
        """
        Ensure env update with duplicate device type used for canary_payload_device_types has no ill effects

        :class: functional
        """
        name = f"enki{self.uid()}"
        data = Env(canary_payload_device_types=[DeviceType.WINDOWS, DeviceType.WINDOWS])

        hc_connect.v1_env_update(data, env_name=name)
        if pg_access.usable:
            resp = pg_access.get_env(name)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find env"
            check = resp[0]
            assert check[pg_access.PgEnv.canary_payload_timestamps] is not None
            assert len(check[pg_access.PgEnv.canary_payload_timestamps]) == 1
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")
