#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import datetime
import logging
import time

import pytest
from psc_orgkey import encode_orgkey, decode_orgkey
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.enumerations import HintType, DeploymentType, DeviceType
from lib.models.environment import Env
from lib.models.hinting import HintRequest
from lib.models.organization import Org
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


class TestHintingExtensive(TestCommon):
    """
    This class tests device endpoints beyond normal "happy path" tests.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    # NOTE: In all tests, we cannot be assured of the exact time returned with the hinting call.  The
    #      `prepare_for_hinting()` returns a structure containing a "start_time" field with the time before we
    #      create the database entries for testing as well as a "done_time" field that with the time right after doing
    #      all database operations.  The best we can validate is that the returned timestamps are within those
    #      two values.

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01_get_hint_request_up_to_date(self, hc_connect):
        """
        Get Hints when both request timestamps are AFTER all hinting timestamps.  This simulates the case where
        config timestamps have not changed since the device last checked in, and hence there should be no hints.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.TOMORROW,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert not response.last_update_time
        assert not response.last_live_query_time
        assert response.hints == set()  # no hints

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02a_get_hint_live_query_outdated(self, hc_connect):
        """
        Get Hints when only the last_live_query_time is out of date with the hinting timestamps.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY, last_update_time=self.TOMORROW,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert not response.last_update_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.LIVE_QUERY,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02c_get_hint_update_outdated(self, hc_connect):
        """
        Get Hints when only the last_update_time is outdated.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert not response.last_live_query_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.REP_REQUEST,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
        }

    # NOTE: test with both last_live_query_time and last_update_time out of date is in the basic suite

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02d_get_hint_update_sequential(self, hc_connect):
        """
        With a workflow of hint change updates, verify the returned hints.

        NOTE: need at least 1 second between sections to account for per-second resolution in
              "internal/hints/hint_generator.go:UpdateHints()"

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        # baseline
        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }
        time.sleep(1.1)
        data.update_times(response.last_update_time, response.last_live_query_time)

        # no database updates, no hints
        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.hints == set()
        time.sleep(1.1)
        data.update_times(response.last_update_time, response.last_live_query_time)

        # Update environment, others unchanged
        hc_connect.v1_env_update(info.env, info.name)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.hints == {
            HintType.HASH_LIST_REFRESH,
            HintType.NEW_POLICY,
        }
        time.sleep(1.1)
        data.update_times(response.last_update_time, response.last_live_query_time)

        # Update organization, others unchanged
        hc_connect.v1_org_update(info.org, info.orgkey)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.HASH_LIST_REFRESH,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
        }
        time.sleep(1.1)
        data.update_times(response.last_update_time, response.last_live_query_time)

        # Update group, others unchanged
        hc_connect.v1_devicegroup_update(info.group, info.orgkey, info.group_id)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.hints == {
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
        }
        time.sleep(1.1)
        data.update_times(response.last_update_time, response.last_live_query_time)

        # Update device, others unchanged
        hc_connect.v1_device_update(info.dev, info.orgkey, info.dev_id)

        response = hc_connect.v1_hints_post(data)  # baseline
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.hints == {
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_UPLOAD,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.REP_REQUEST,
        }

    # NOTE: test with both last_live_query_time and last_update_time out of date is in the basic suite

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02e_get_hint_update_device_type_missing(self, hc_connect):
        """
        Get Hints when device_type is missing.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert not response.last_live_query_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02f_get_hint_update_deployment_type_missing(self, hc_connect):
        """
        Get Hints when deployment_type is missing.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert not response.last_live_query_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02g_get_hint_update_device_type_unknown(self, hc_connect):
        """
        Get Hints when device_type is unknown.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['device_type'] = "AMIGA"

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert not response.last_live_query_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_02h_get_hint_update_deployment_type_unknown(self, hc_connect):
        """
        Get Hints when deployment_type is missing.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.TOMORROW, last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['deployment_type'] = "CLOUD"

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert not response.last_live_query_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03_get_hint_timestamps_zeroed(self, hc_connect):
        """
        Get Hints when last_live_query_time and last_update_time are both epoch time (0).

        NOTE: this is the same situation as being WAY outdated, but just a special case.

        :class: functional
        """
        epoch = self.epoch()
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=epoch, last_update_time=epoch,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }

    # NOTE: No way to easily "tune" the hinting with different times for env, org, group or device

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_get_hint_device_not_exist(self, hc_connect):
        """
        Attempt to get a hint using a device id that does not exist.  Any device-specific hints should not be returned.
        NOTE: Same scenario as if you specified a valid device and valid org that the device was not in -- device not found

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.HASH_LIST_REFRESH,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04b_get_hint_devicegroup_not_exist(self, hc_connect):
        """
        Attempt to get a hint using a device group id that does not exist. Any group-specific hints should not be returned.
        NOTE: Same scenario as if you specified a valid device group and valid org that the device group was not in -- device not found

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)

        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04c_get_hint_organization_not_exist(self, hc_connect):
        """
        Attempt to get a hint using a valid organization orgkey that does not exist.  Hints reliant on org,
        device and group will not be sent (since group and device have orgkey as a part of their index).

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        fakekey = encode_orgkey(info.name, self.uid())  # non-existing org in valid env

        data = HintRequest(org_key=fakekey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert response.last_live_query_time is None
        assert response.hints == {
            HintType.HASH_LIST_REFRESH,
            HintType.NEW_POLICY,
            HintType.ZIP_CONTAINER,
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04d_get_hint_environment_not_exist(self, hc_connect):
        """
        Attempt to get a hint using an orgkey for an environment that was deleted (proves environment delete works).
        This is also the same as specifying an env that never existed.

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect, no_env=True)

        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04e_get_hint_everything_missing(self, hc_connect):
        """
        Attempt to get a hint using information for orgkey, group and id that do not exist and where the
        env for the corresponding orgkey also does not exist.
        (simulated new device checking in that has not had stuff added to the hint configuration yet).

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect, no_env=True)
        fakekey = encode_orgkey(info.name, self.uid())  # non-existing org in valid env

        data = HintRequest(org_key=fakekey, device_group_id=self.uid(), device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert response.last_update_time is None
        assert response.last_live_query_time is None
        assert response.hints == set()

    # FIXME: Work in progress
    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04f_get_hint_new_device(self, hc_connect):
        """
        From DSER-46434:
          - Hinting Microservice May 404 on Receipt of Device Information for Newly Created Devices

        Attempt to get a hint using information supposedly from a device that is newly checking in, but nothing
        has been sent to the hint service about the group or device (it is assumed that env and org have been created
        in order to get sensor installs).

        Data taken from ticket:
            "update_body":"{\"org_key\":\"83LUMHX93\",\"device_group_id\":195745,\"device_id\":429756,\"version\":18,
            \"last_update_time\":0,\"audit_hints\":[\"HASH_LIST_REFRESH\",\"NEW_POLICY\",\"NEW_CONFIG_DATA\",\"ZIP_CONTAINER\"],
            \"audit_last_update_time\":1680671376510}"

        :class: functional
        """
        orgkey = "83LUMHX93"
        info = decode_orgkey(orgkey)
        env = Env.full()
        hc_connect.v1_env_update(env, info.env)
        org = Org.full()
        hc_connect.v1_org_update(org, orgkey)

        data = HintRequest(org_key=orgkey, device_group_id=195745, device_id=429756,
                           version=18, last_update_time=self.epoch(),
                           audit_hints={HintType.HASH_LIST_REFRESH, HintType.NEW_POLICY, HintType.NEW_CONFIG_DATA,
                                        HintType.ZIP_CONTAINER},
                           audit_last_update_time=self.stamp())

        hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 200)

    # NOTE: MessageCode specific tests (test_05x) removed; DSER-45237

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06a_get_hint_kill_switch_false(self, hc_connect):
        """
        Verify that when the organization  is REQ_UNINSTALL_CODE you won't see HintType.KILL_SWITCH_ON

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)

        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)
        response = hc_connect.v1_hints_post(data)
        assert HintType.KILL_SWITCH_ON in response.hints

        info.org.is_kill_switch_on = False
        hc_connect.v1_org_update(info.org, data.org_key)
        response = hc_connect.v1_hints_post(data)
        assert HintType.KILL_SWITCH_ON not in response.hints

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_07a_get_hint_remote_rules_required_true(self, hc_connect):
        """
        Get Hints when remote_rules_required is True and message code is REQ_HELLO

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=True)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    # NOTE: MessageCode specific tests (test_07b) removed; DSER-45237

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_08_get_hint_unknown_message_code(self, hc_connect):
        """
        Get Hints when unknown message code is specified (log message will be raised, but otherwise harmless)
        NOTE: Simulates monolith/other services using a newly defined code -- we should squawk but otherwise work

        NOTE: MessageCode no longer used; this test should verify that supplied message code is ignored (DSER-45237)

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['message_code'] = "REQ_NEW_MESSAGE_CODE"

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"{response}")
        assert hc_connect.last_status == 200
        assert info.start_time < response.last_update_time < info.done_time
        assert info.start_time < response.last_live_query_time < info.done_time
        assert response.hints == {
            HintType.BLADE_ACTION,
            HintType.CONFER_UPLOAD,
            HintType.HASH_LIST_REFRESH,
            HintType.HASH_LIST_UPLOAD,
            HintType.KILL_SWITCH_ON,
            HintType.LIVE_QUERY,
            HintType.NEW_CONFIG_DATA,
            HintType.NEW_POLICY,
            HintType.ONETIME_HASH_DELETE_LIST,
            HintType.REREGISTER,
            HintType.SENSOR_ACTION,
            HintType.SENSOR_CERTIFICATE_UPLOAD,
            HintType.SENSOR_TRIGGER_PERFMON,
            HintType.SOFTWARE_UPGRADE,
            HintType.UNINSTALL_CODE,
            HintType.ZIP_CONTAINER,
            HintType.REP_REQUEST
        }

    # ----- Plucked from E2E tests --------------------------------------------------

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_10a_get_hint_org_key_casing(self, hc_connect):
        """
        Verify that supplied casing of org_key has no effect.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        assert info.orgkey == info.orgkey.upper()

        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)
        response1 = hc_connect.v1_hints_post(data)

        data = HintRequest(org_key=info.orgkey.lower(), device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False)
        response2 = hc_connect.v1_hints_post(data)
        assert response1.hints == response2.hints

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_11a_get_hint_device_missing(self, hc_connect):
        """
        Attempt to get a hint but device id submission is missing

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        del data['device_id']

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_11b_get_hint_device_bogus(self, hc_connect):
        """
        Attempt to get a hint but device id submission is not int

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['device_id'] = "BOGUS"

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_12a_get_hint_group_missing(self, hc_connect):
        """
        Attempt to get a hint but group id submission is missing

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        del data['device_group_id']

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_12b_get_hint_group_bogus(self, hc_connect):
        """
        Attempt to get a hint but group id submission is not int

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['device_group_id'] = "BOGUS"

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_13a_get_hint_orgkey_missing(self, hc_connect):
        """
        Attempt to get a hint but orgkey submission is missing

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        del data['org_key']

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_13b_get_hint_orgkey_empty_string(self, hc_connect):
        """
        Attempt to get a hint but orgkey submission is empty string

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['org_key'] = ""

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_14a_get_hint_live_query_string(self, hc_connect):
        """
        Attempt to get a hint but submitted livequery time is date as string, not millis

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['last_live_query_time'] = "2022-10-14T22:11:20+0000"

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_14b_get_hint_last_update_string(self, hc_connect):
        """
        Attempt to get a hint but submitted last update time is date as string, not millis

        :class: functional, negative
        """
        info = self.prepare_for_hinting(hc_connect)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=self.uid(),
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=self.YESTERDAY,
                           last_update_time=self.YESTERDAY,
                           remote_rules_required=False).to_dict()
        data['last_update_time'] = "2022-10-14T22:11:20+0000"

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_hints_post(data)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_16a_post_v1_hints_with_audit_timestamp_within_tolerance(self, hc_connect):
        """
        Verify that hinting is not returned when last time is out of date, but within a 1 second tolerance.

        :class: functional
        """
        info = self.prepare_for_hinting(hc_connect)
        early = info.done_time - datetime.timedelta(milliseconds=500)
        data = HintRequest(org_key=info.orgkey, device_group_id=info.group_id, device_id=info.dev_id,
                           device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                           version=self.LQ_VERSION, last_live_query_time=early, last_update_time=early,
                           remote_rules_required=False,
                           audit_hints={HintType.BLADE_ACTION,
                                        HintType.CONFER_UPLOAD,
                                        HintType.HASH_LIST_REFRESH,
                                        HintType.HASH_LIST_UPLOAD,
                                        HintType.KILL_SWITCH_ON,
                                        HintType.LIVE_QUERY,
                                        HintType.NEW_CONFIG_DATA,
                                        HintType.NEW_POLICY,
                                        HintType.ONETIME_HASH_DELETE_LIST,
                                        HintType.REREGISTER,
                                        HintType.SENSOR_ACTION,
                                        HintType.SENSOR_CERTIFICATE_UPLOAD,
                                        HintType.SENSOR_TRIGGER_PERFMON,
                                        HintType.SOFTWARE_UPGRADE,
                                        HintType.UNINSTALL_CODE,
                                        HintType.ZIP_CONTAINER},
                           audit_last_live_query_time=info.done_time, audit_last_update_time=info.done_time)

        response = hc_connect.v1_hints_post(data)
        _logger.info(f"Last Update:  {response.last_update_time}, Last Livequery: {response.last_live_query_time}\n" +
                     f"Hints: {response.hints}")
        self.verify_response(hc_connect, 200)
        assert response.last_live_query_time is None
        assert response.last_update_time is None
        assert response.hints == set()
