#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging

import pytest
from psc_orgkey import encode_orgkey
from pytest_sguid_plugin.sguid_plugin import SGUID

from lib.models.enumerations import DeviceType, OrgChangeType, DeploymentType
from lib.models.organization import Org
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestOrganizationsExtensive(TestCommon):
    """
    This class tests organization endpoints beyond normal "happy path" tests.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01a_v1_org_update_uri_orgkey_emptystring(self, hc_connect):
        """
        Ensure org update with empty string orgkey is caught

        :class: functional, negative
        """
        data = Org.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_org_update(data, org_key="")
        self.verify_response(hc_connect, 404, "404 page not found")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01b_v1_org_update_uri_orgkey_invalid(self, hc_connect):
        """
        Ensure org update with an invalid orgkey (cannot decode) is caught

        :class: functional, negative
        """
        data = Org.full()

        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_org_update(data, org_key="WONT_DECODE")
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_01c_v1_org_update_uri_orgkey_unknown(self, hc_connect):
        """
        Ensure org update with a valid orgkey (but env & org_id represented do not exist) is allowed

        :class: functional
        """
        data = Org.full()
        key = encode_orgkey("nosuch_env", self.uid())

        hc_connect.v1_org_update(data, org_key=key)
        self.verify_response(hc_connect, 200)

    def test_02a_v1_org_update_ensure_additive_union_canary(self, hc_connect, pg_access):
        """
        Ensure env update with two different canary_payload_device_types definitions results
        in the UNION of the two, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org(canary_payload_device_types=[DeviceType.WINDOWS])
        data2 = Org(canary_payload_device_types=[DeviceType.IOS])

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.canary_payload_timestamps]) == 2
            assert DeviceType.WINDOWS.name in check[pg_access.PgOrg.canary_payload_timestamps] and DeviceType.IOS.name in check[pg_access.PgOrg.canary_payload_timestamps]
            assert check[pg_access.PgOrg.canary_payload_timestamps][DeviceType.WINDOWS.name] != check[4][DeviceType.IOS.name]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    def test_02b_v1_org_update_ensure_subtractive_update_canary(self, hc_connect, pg_access):
        """
        Ensure org that is fully populated and then updated with a single canary_payload_device_types definition results
        in only in the UPDATE of the new data, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org.full()
        data2 = Org(canary_payload_device_types=[DeviceType.IOS])

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.canary_payload_timestamps]) == len(DeviceType.all())
            assert check[pg_access.PgOrg.canary_payload_timestamps][DeviceType.WINDOWS.name] != check[pg_access.PgOrg.canary_payload_timestamps][DeviceType.IOS.name]
            assert check[pg_access.PgOrg.canary_payload_timestamps][DeviceType.WINDOWS.name] == check[pg_access.PgOrg.canary_payload_timestamps][DeviceType.MAC.name]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    def test_02c_v1_org_update_ensure_additive_union_changes(self, hc_connect, pg_access):
        """
        Ensure env update with two different changes definitions results
        in the UNION of the two, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org(changes=[OrgChangeType.ORG_HASH_LIST_CHANGE])
        data2 = Org(changes=[OrgChangeType.ORG_KILL_SWITCH_ON])

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.change_timestamps]) == 2
            assert str(OrgChangeType.ORG_HASH_LIST_CHANGE.value) in check[pg_access.PgOrg.change_timestamps] and str(
                OrgChangeType.ORG_KILL_SWITCH_ON.value) in check[pg_access.PgOrg.change_timestamps]
            assert check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)] != check[pg_access.PgOrg.change_timestamps][
                str(OrgChangeType.ORG_KILL_SWITCH_ON.value)]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    def test_02d_v1_org_update_ensure_subtractive_update_changes(self, hc_connect, pg_access):
        """
        Ensure org that is fully populated and then updated with a single changes definition results
        in only in the UPDATE of the new data, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org.full()
        data2 = Org(changes=[OrgChangeType.ORG_NET_CFG_CHANGE])

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.change_timestamps]) == len(OrgChangeType.all())
            assert check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_NET_CFG_CHANGE.value)] != check[pg_access.PgOrg.change_timestamps][
                str(OrgChangeType.ORG_KILL_SWITCH_ON.value)]
            assert check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)] == check[pg_access.PgOrg.change_timestamps][
                str(OrgChangeType.ORG_KILL_SWITCH_ON.value)]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    def test_02e_v1_org_update_ensure_additive_union_deployment(self, hc_connect, pg_access):
        """
        Ensure env update with two different deployment_type_changes definitions results
        in the UNION of the two, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org(deployment_type_changes={DeploymentType.AWS: OrgChangeType.all()})
        data2 = Org(deployment_type_changes={DeploymentType.ENDPOINT: OrgChangeType.all()})

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == 2
            assert DeploymentType.AWS.name in check[pg_access.PgOrg.deployment_type_change_timestamps] and DeploymentType.ENDPOINT.name in check[pg_access.PgOrg.deployment_type_change_timestamps]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.AWS.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)] != \
                   check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.ENDPOINT.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    def test_02f_v1_org_update_ensure_subtractive_update_deployment(self, hc_connect, pg_access):
        """
        Ensure org that is fully populated and then updated with a single deployment_type_changes definition results
        in only in the UPDATE of the new data, not a replacement.

        :class: functional
        """
        key = self.new_orgkey()
        data1 = Org.full()
        data2 = Org(deployment_type_changes={DeploymentType.ENDPOINT: OrgChangeType.all()})

        hc_connect.v1_org_update(data1, org_key=key)
        self.verify_response(hc_connect, 200)
        hc_connect.v1_org_update(data2, org_key=key)
        self.verify_response(hc_connect, 200)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == len(DeploymentType.all())
            assert check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.ENDPOINT.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)] != \
                   check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.AWS.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.AZURE.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)] == \
                   check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.AWS.name][str(OrgChangeType.ORG_HASH_LIST_CHANGE.value)]
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03a_v1_org_update_body_missing(self, hc_connect):
        """
        Ensure org update with missing body is caught

        :class: functional, negative
        """
        key = self.new_orgkey()
        with pytest.raises(hc_connect.HTTPError):
            hc_connect.v1_org_update(None, org_key=key)
        self.verify_response(hc_connect, 400)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_03b_v1_org_update_body_empty(self, hc_connect, pg_access):
        """
        Ensure org update with empty body is allowed.

        :class: functional
        """
        key = self.new_orgkey()
        hc_connect.v1_org_update({}, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.is_kill_switch_on] is None
            assert check[pg_access.PgOrg.change_timestamps] is None
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is None
            assert check[pg_access.PgOrg.canary_payload_timestamps] is None
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04a_v1_org_update_body_invalid_canary(self, hc_connect):
        """
        Ensure org update with invalid device type used for canary_payload_device_types is caught

        :class: functional, negative
        """
        key = self.new_orgkey()
        data = Org.full().to_dict()
        data['canary_payload_device_types'] = ["SAMSUNG"]

        hc_connect.v1_org_update(data, org_key=key)
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_04b_v1_org_update_body_duplicate_canary(self, hc_connect, pg_access):
        """
        Ensure org update with duplicate device type used for canary_payload_device_types has no ill effects

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(canary_payload_device_types=[DeviceType.WINDOWS, DeviceType.WINDOWS])

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.canary_payload_timestamps] is not None
            assert len(check[pg_access.PgOrg.canary_payload_timestamps]) == 1
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05a_v1_org_update_body_invalid_changes(self, hc_connect):
        """
        Ensure org update with invalid org change type used for changes is caught

        :class: functional, negative
        """
        key = self.new_orgkey()
        data = Org.full().to_dict()
        data['changes'] = ["ORG_UNKNOWN_CHANGE"]

        hc_connect.v1_org_update(data, org_key=key)
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_05b_v1_org_update_body_duplicate_changes(self, hc_connect, pg_access):
        """
        Ensure org update with duplicate org change type used for changes has no ill effects

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(changes=[OrgChangeType.ORG_HASH_LIST_CHANGE, OrgChangeType.ORG_HASH_LIST_CHANGE])

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.change_timestamps] is not None
            assert len(check[pg_access.PgOrg.change_timestamps]) == 1
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06a_v1_org_update_body_invalid_deployment(self, hc_connect):
        """
        Ensure org update with invalid deployment type used for changes is caught

        :class: functional, negative
        """
        key = self.new_orgkey()
        data = Org.full().to_dict()
        del data['deployment_type_changes']['VDI']  # make room
        data['deployment_type_changes']['STRANGE'] = ["ORG_BLADE_ACTION_CHANGE"]

        hc_connect.v1_org_update(data, org_key=key)
        self.verify_response(hc_connect, 200)

    # NOTE: being a dict, unable to have duplicate deployment entries

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06b_v1_org_update_body_invalid_deployment_change(self, hc_connect):
        """
        Ensure org update with deployment type having an invalid org changes type is caught

        :class: functional, negative
        """
        key = self.new_orgkey()
        data = Org.full().to_dict()
        data['deployment_type_changes']['ENDPOINT'] = ["ORG_NET_CFG_CHANGE"]
        data['deployment_type_changes']['ENDPOINT'].append("SOME_UNKNOWN_CHANGE")

        hc_connect.v1_org_update(data, org_key=key)
        self.verify_response(hc_connect, 200)

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_06d_v1_org_update_body_duplicate_deployment_change(self, hc_connect, pg_access):
        """
        Ensure org update with duplicate deployment type used for changes is has no ill efects, with sequent entry
        taking precidence.

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(deployment_type_changes={
            DeploymentType.VDI: [OrgChangeType.ORG_HASH_LIST_CHANGE, OrgChangeType.ORG_HASH_LIST_CHANGE]})

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is not None
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == 1
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name]) == 1
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_07a_v1_org_update_deployment_livequery_override(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value overrides "now" in deplotment if specified when doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(deployment_type_changes={DeploymentType.VDI: [OrgChangeType.ORG_LIVE_QUERY_CHANGE]},
                   live_query_time=self.pinned())

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is not None
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == 1
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name]) == 1
            assert check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name][str(OrgChangeType.ORG_LIVE_QUERY_CHANGE.value)].startswith(
                "1999-09")
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_07b_v1_org_update_deployment_livequery_override_not_livequery(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value does not override "now" in deplyment if specified when not doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(deployment_type_changes={DeploymentType.VDI: [OrgChangeType.ORG_KILL_SWITCH_ON]},
                   live_query_time=self.pinned())

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is not None
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == 1
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name]) == 1
            assert self.is_now(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name][str(OrgChangeType.ORG_KILL_SWITCH_ON.value)])
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_07c_v1_org_update_deployment_livequery_default(self, hc_connect, pg_access):
        """
        Ensure no livequery override uses "now" in deplyment when doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(deployment_type_changes={DeploymentType.VDI: [OrgChangeType.ORG_LIVE_QUERY_CHANGE]})

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.deployment_type_change_timestamps] is not None
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps]) == 1
            assert len(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name]) == 1
            assert self.is_now(check[pg_access.PgOrg.deployment_type_change_timestamps][DeploymentType.VDI.name][
                str(OrgChangeType.ORG_LIVE_QUERY_CHANGE.value)])
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_08a_v1_org_update_changes_livequery_override(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value overrides "now" in changes if specified when doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(changes=[OrgChangeType.ORG_LIVE_QUERY_CHANGE], live_query_time=self.pinned())

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.change_timestamps] is not None
            assert len(check[pg_access.PgOrg.change_timestamps]) == 1
            assert check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_LIVE_QUERY_CHANGE.value)].startswith("1999-09")
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_08b_v1_org_update_changes_livequery_override_not_livequery(self, hc_connect, pg_access):
        """
        Ensure supplied livequery value does not override "now" in changes if specified when not doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(changes=[OrgChangeType.ORG_BLADE_ACTION_CHANGE], live_query_time=self.pinned())

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.change_timestamps] is not None
            assert len(check[pg_access.PgOrg.change_timestamps]) == 1
            assert self.is_now(check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_BLADE_ACTION_CHANGE.value)])
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")

    @pytest.mark.sguid(SGUID.psc_platform_ems_device_hints)
    def test_08c_v1_org_update_changes_livequery_default(self, hc_connect, pg_access):
        """
        Ensure no livequery value uses "now" in changes if specified when doing ORG_LIVE_QUERY_CHANGE

        :class: functional
        """
        key = self.new_orgkey()
        data = Org(changes=[OrgChangeType.ORG_LIVE_QUERY_CHANGE])

        hc_connect.v1_org_update(data, org_key=key)

        if pg_access.usable:
            resp = pg_access.get_org(key)
            _logger.info(f">>> Returned: {resp}")
            assert len(resp) == 1, f"Did not find org"
            check = resp[0]
            assert check[pg_access.PgOrg.change_timestamps] is not None
            assert len(check[pg_access.PgOrg.change_timestamps]) == 1
            assert self.is_now(check[pg_access.PgOrg.change_timestamps][str(OrgChangeType.ORG_LIVE_QUERY_CHANGE.value)])
        else:
            _logger.warning("NOTE: No access to postgres, cannot ensure content!")
