#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import logging
from typing import Dict, List

from lib.connect.hint_connect import HintControlConnect
from lib.models.device import Device
from lib.models.enumerations import DeviceChangeType
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestConnect(TestCommon):
    """
    This class validates the connection internal utilities
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    def test_01a_convert_data_was_model(self):
        """Validate that submitted models are turned into dicts.

        :class: internal
        """
        dev = Device(changes=[DeviceChangeType.DEV_REREGISTER], live_query_time=self.pinned())
        check = HintControlConnect._convert_data(dev)
        assert isinstance(check, Dict)
        assert 'changes' in check
        assert check['changes'] == ['DEV_REREGISTER']
        assert 'live_query_time' in check
        assert check['live_query_time'] == 937180800000

    def test_01b_convert_data_was_dict(self):
        """Validate that submitted dicts are returned as is.

        :class: internal
        """
        data = {'foo': 'bar'}
        check = HintControlConnect._convert_data(data)
        assert isinstance(check, Dict)
        assert 'foo' in check
        assert check['foo'] == 'bar'

    def test_01c_convert_data_was_other(self):
        """Validate that other submitted types are returned as is (a warning will be raised!)

        :class: internal
        """
        data = "foobar"
        check = HintControlConnect._convert_data(data)
        assert isinstance(check, str)
        assert check == "foobar"

    def test_02a_get_field_was_model(self):
        """Validate that fields are properly obtained from models.

        :class: internal
        """
        dev = Device(changes=[DeviceChangeType.DEV_REREGISTER])
        check = HintControlConnect._get_field(dev, "changes")
        assert isinstance(check, List)

    def test_02b_get_field_was_other_object(self):
        """Validate that fields are properly obtained from other objects.

        :class: internal
        """
        data = TestCommon
        check = HintControlConnect._get_field(data, "INVOKED")
        assert isinstance(check, str)

    def test_02c_get_field_was_dict(self):
        """Validate that fields are properly obtained from dicts.

        :class: internal
        """
        data = {'foo': [1, 2, 3]}
        check = HintControlConnect._get_field(data, "foo")
        assert isinstance(check, List)

    def test_02d_get_field_was_other_type(self):
        """Validate that None is returned for other types.

        :class: internal
        """
        data = "{'foo': [1, 2, 3]}"
        check = HintControlConnect._get_field(data, "foo")
        assert check is None
