#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import logging
from typing import Dict

import pytest
from psc_orgkey import encode_orgkey, decode_orgkey

from lib.models.base_model import BaseModel
from lib.models.device import Device
from lib.models.device_group import DeviceGroup
from lib.models.enumerations import DeploymentType, DeviceType, HintType, OrgChangeType, DeviceChangeType, \
    DeviceGroupChangeType
from lib.models.environment import Env
from lib.models.hinting import HintRequest, HintResult
from lib.models.organization import Org
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestModels(TestCommon):
    """
    This class validates the test harness model structures.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    def test_00a_date_to_hinting_time(self):
        """Validate supplied datetime converts properly to hinting time.

        :class: internal
        """
        data = self.pinned()
        check = BaseModel._date_to_hinting_time(data)
        assert check == 937180800000

    def test_00b_date_to_hinting_time_none(self):
        """Validate supplied None type is handled properly.

        :class: internal
        """
        check = BaseModel._date_to_hinting_time(None)
        assert check is None

    def test_00c_hinting_time_to_date(self):
        """Validate hinting time in millis converts properly to datetime.

        :class: internal
        """
        check = BaseModel._hinting_time_to_date(937180800000)
        assert check.year == 1999
        assert check.month == 9
        assert check.day == 13
        assert check.hour == 0
        assert check.minute == 0
        assert check.second == 0
        assert check.microsecond == 0

    def test_00d_hinting_time_to_date_none(self):
        """Validate hinting time returned as None returns as None.

        :class: internal
        """
        check = BaseModel._hinting_time_to_date(None)
        assert check is None

    def test_00e_convert_times_in_dict_none(self):
        """Validate dict conversion if timestamps returns None if supplied with None

        :class: internal
        """
        check = BaseModel()._convert_times_in_dict(None)
        assert check is None

    def test_00f_convert_times_in_dict_empty(self):
        """Validate dict conversion if dict is empty.

        :class: internal
        """
        data = {}
        check = BaseModel()._convert_times_in_dict(data)
        assert isinstance(check, Dict)
        assert len(check) == 0

    def test_00g_convert_times_in_dict_non_enum_key(self):
        """Validate dict conversion when keyed by non-enums.

        :class: internal
        """
        data = {"foo": self.pinned()}
        check = BaseModel()._convert_times_in_dict(data)
        assert isinstance(check, Dict)
        assert "foo" in check
        assert check["foo"] == 937180800000

    def test_00h_convert_times_in_dict_enum_key(self):
        """Validate dict conversion when keyed by enums.

        :class: internal
        """
        data = {DeviceType.WINDOWS: self.pinned()}
        check = BaseModel()._convert_times_in_dict(data)
        assert isinstance(check, Dict)
        assert "WINDOWS" in check
        assert check["WINDOWS"] == 937180800000

    def test_00i_convert_times_in_dict_multi_level(self):
        """Validate dict conversion when there are multiple levels.

        :class: internal
        """
        data = {DeviceType.WINDOWS: {OrgChangeType.ORG_KILL_SWITCH_ON: self.pinned()}}
        check = BaseModel()._convert_times_in_dict(data)
        assert isinstance(check, Dict)
        assert "WINDOWS" in check
        check2 = check["WINDOWS"]
        assert "ORG_KILL_SWITCH_ON" in check2
        assert check2["ORG_KILL_SWITCH_ON"] == 937180800000

    def test_00j_safe_enum_none(self):
        """Validate _safe_enum handles None.

        :class: internal
        """
        check = BaseModel._safe_enum(None)
        assert check is None

    def test_00k_safe_enum(self):
        """Validate _safe_enum handles None.

        :class: internal
        """
        check = BaseModel._safe_enum(DeviceType.WINDOWS)
        assert check == "WINDOWS"

    def test_00l_safe_non_enum(self):
        """Validate _safe_enum handles None.

        :class: internal
        """
        check = BaseModel._safe_enum("foobar")
        assert check == "foobar"

    def test_01a_org_key_encode(self):
        """Validate proper OrgKey when created with encode.

        :class: internal
        """
        encoded = encode_orgkey(env='dev01', org_id=1)
        assert encoded == 'J7G6DTLN'

    @pytest.mark.utility(reason="functional test utility, not a test")
    def test_01a2_org_key_encode(self):
        """Validate proper OrgKey when created with encode.

        :class: internal
        """
        for i in range(100):
            encoded = encode_orgkey(env=self.unique().lower(), org_id=i + 1)
            _logger.info(encoded)

    def test_01b_org_key_decode(self):
        """Validate proper OrgKey when updated with new key string.

        FUTURE: Update test when real encoding used
        :class: internal
        """
        decoded = decode_orgkey(orgkey='J7G6DTLN')
        assert 'dev01' == decoded.env
        assert 1 == decoded.org_id

    def test_11a_device(self):
        """Validate Device() initialization.

        :class: internal
        """
        target = Device([DeviceChangeType.DEV_REREGISTER])
        assert len(target.__dict__) == 2  # check for changes in structure fields
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['changes'] == ["DEV_REREGISTER"]
        assert check['live_query_time'] is None

    def test_11b_device_full(self):
        """Validate Device() initialization when fully flled in.

        :class: internal
        """
        target = Device(DeviceChangeType.all(), live_query_time=self.pinned())
        assert len(target.__dict__) == 2  # check for changes in structure fields
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert len(check['changes']) == len(DeviceChangeType.all())
        assert check['live_query_time'] == 937180800000

    # noinspection PyTypeChecker,PyArgumentList
    def test_11c_device_bad(self):
        """Verify that bad entries are trapped on Device init.

        :class: internal, negative
        """
        with pytest.raises(AssertionError):  # 'changes' not list
            Device(changes=DeviceChangeType.DEV_REREGISTER)
        with pytest.raises(AssertionError):  # 'changes' not list of DeviceChangeType
            Device(changes=["DEV_REREGISTER"])
        with pytest.raises(AssertionError):  # 'changes' len > max
            Device(changes=[DeviceChangeType.all()] + [DeviceChangeType.DEV_REREGISTER])
        with pytest.raises(AssertionError):  # 'live_query_time' not datetime
            Device(live_query_time=123456789)

    def test_11d_device_valid_empty(self):
        """Verify that bad entries are trapped on Device init.

        :class: internal, negative
        """
        Device()
        Device(changes=[])

    def test_11d_device_vary(self):
        """Validate Device().vary does not produce bad entries.

        :class: internal
        """
        target = Device.full()
        with self.Timer(tag=f"100 device variants"):
            for _ in range(100):
                target.vary()
                try:
                    target.is_valid()
                except AssertionError as err:
                    _logger.warning(f"is_valid failed with: {target.to_dict()}:\n  {err}")
                    raise

    def test_12a_device_group(self):
        """Validate DeviceGroup() initialization.

        :class: internal
        """
        target = DeviceGroup(changes=[DeviceGroupChangeType.DG_POLICY_CHANGE])
        assert len(target.__dict__) == 2  # check for changes in structure size
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['changes'] == ["DG_POLICY_CHANGE"]

    def test_12b_device_group_full(self):
        """Validate DeviceGroup() initialization.

        :class: internal
        """
        target = DeviceGroup(changes=DeviceGroupChangeType.all(), live_query_time=self.pinned())
        assert len(target.__dict__) == 2  # check for changes in structure size
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert len(check['changes']) == len(DeviceGroupChangeType.all())
        assert check['live_query_time'] == 937180800000

    # noinspection PyTypeChecker, PyArgumentList
    def test_12c_device_group_bad(self):
        """Verify that bad id entries are trapped on DeviceGroup init.

        :class: internal, negative
        """
        with pytest.raises(AssertionError):  # 'changes' not list
            DeviceGroup(changes=DeviceGroupChangeType.DG_POLICY_CHANGE)
        with pytest.raises(AssertionError):  # 'changes' not list of DeviceChangeType
            DeviceGroup(changes=["DEV_REREGISTER"])
        with pytest.raises(AssertionError):  # 'changes' len > max
            DeviceGroup(changes=[DeviceGroupChangeType.all()] + [DeviceGroupChangeType.DG_POLICY_CHANGE])
        with pytest.raises(AssertionError):  # 'live_query_time' not datetime
            DeviceGroup(live_query_time=123456789)

    def test_12d_device_group_allowed_empty(self):
        """Verify that allowed empty changes are accepted..

        :class: internal, negative
        """
        DeviceGroup()
        DeviceGroup(changes=[])

    def test_12d_device_group_vary(self):
        """Validate DeviceGroup().vary does not produce bad entries.

        :class: internal
        """
        target = DeviceGroup.full()
        with self.Timer(tag=f"100 device variants"):
            for _ in range(100):
                target.vary()
                try:
                    target.is_valid()
                except AssertionError as err:
                    _logger.warning(f"is_valid failed with: {target.to_dict()}:\n  {err}")
                    raise

    def test_13a_organization(self):
        """Validate Organization() initialization.

        :class: internal
        """
        target = Org(is_kill_switch_on=True,
                     canary_payload_device_types=DeviceType.all(),
                     changes=OrgChangeType.all(),
                     deployment_type_changes=DeploymentType.all_dict_org(),
                     live_query_time=self.pinned())
        assert len(target.__dict__) == 5  # check for changes in structure size
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['canary_payload_device_types'] == [
            'WINDOWS', 'ANDROID', 'MAC', 'IOS', 'LINUX', 'OTHER'
        ]
        assert len(check['changes']) == len(OrgChangeType)  # len, since this is so large
        assert len(check['deployment_type_changes']) == len(DeploymentType)  # len, since this is really large
        assert len(check['deployment_type_changes']["ENDPOINT"]) == len(OrgChangeType)
        assert check['is_kill_switch_on']
        assert check['live_query_time'] == 937180800000

    # noinspection PyTypeChecker,PyArgumentList
    def test_13b_organization_bad(self):
        """Verify that bad id entries are trapped on Organization init.

        :class: internal, negative
        """
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' not list
            Org(canary_payload_device_types=DeviceType.WINDOWS)
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' len > max
            Org(canary_payload_device_types=DeviceType.all() + [DeviceType.WINDOWS])
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' entry not device type
            Org(canary_payload_device_types=["WINDOWS"])
        with pytest.raises(AssertionError):  # 'changes' not list
            Org(changes=OrgChangeType.ORG_HASH_LIST_CHANGE)
        with pytest.raises(AssertionError):  # 'changes' len > max
            Org(changes=OrgChangeType.all() + [OrgChangeType.ORG_HASH_LIST_CHANGE])
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' entry not device type
            Org(changes=["ORG_HASH_LIST_CHANGE"])
        with pytest.raises(AssertionError):  # 'deployment_type_changes' not dict
            Org(deployment_type_changes=DeploymentType.ENDPOINT)
        # NOTE: Not possible to have more entries than Deployment types since it is a dict with no duplicates
        with pytest.raises(AssertionError):  # 'deployment_type_changes' entry not device type
            Org(deployment_type_changes={"ENDPOINT": OrgChangeType.all()})
        with pytest.raises(AssertionError):  # 'deployment_type_changes' entry value not list
            Org(deployment_type_changes={DeploymentType.ENDPOINT: "ORG_HASH_LIST_CHANGE"})
        with pytest.raises(AssertionError):  # 'deployment_type_changes' entry value len > max
            Org(deployment_type_changes={
                DeploymentType.ENDPOINT: OrgChangeType.all() + [OrgChangeType.ORG_KILL_SWITCH_ON]})
        with pytest.raises(AssertionError):  # 'deployment_type_changes' entry value list item not org change type
            Org(deployment_type_changes={DeploymentType.ENDPOINT: ["ORG_KILL_SWITCH_ON"]})
        with pytest.raises(AssertionError):  # 'is_kill_switch_on' not bool
            Org(is_kill_switch_on="True")
        with pytest.raises(AssertionError):  # 'live_query_time' not datetime
            Device(live_query_time=123456789)

    def test_13c_organization_empty_ok(self):
        """
        Verify that empty entries ok for Organization init.
            -- See DSER-49110

        :class: internal, negative
        """
        Org()
        Org(canary_payload_device_types=[])
        Org(changes=[])
        Org(deployment_type_changes={})
        Org(deployment_type_changes={DeploymentType.ENDPOINT: []})

    def test_13d_organization_vary(self):
        """Validate Org().vary does not produce bad entries.

        :class: internal
        """
        target = Org.full()
        with self.Timer(tag=f"1000 organization variants"):
            for _ in range(1000):
                target.vary()
                try:
                    target.is_valid()
                except AssertionError as err:
                    _logger.warning(f"is_valid failed with: {target.to_dict()}:\n  {err}")
                    raise

    def test_14a_environment_full(self):
        """Validate Env() initialization.

        :class: internal
        """

        target = Env(canary_payload_device_types=DeviceType.all(), cert_service_time_changed=True,
                     policy_service_time_changed=True)
        assert len(target.__dict__) == 3  # check for changes in structure size
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['canary_payload_device_types'] == [
            'WINDOWS', 'ANDROID', 'MAC', 'IOS', 'LINUX', 'OTHER'
        ]
        assert check['cert_service_time_changed']
        assert check['policy_service_time_changed']

    def test_14b_environment_partial_1(self):
        """Validate Env() initialization with partial data.

        :class: internal
        """

        target = Env(canary_payload_device_types=[DeviceType.IOS])
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['canary_payload_device_types'] == [
            'IOS',
        ]
        assert not check['cert_service_time_changed']
        assert not check['policy_service_time_changed']

    def test_14c_environment_partial_2(self):
        """Validate Env() initialization with partial data.

        :class: internal
        """

        target = Env(cert_service_time_changed=True)
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check.get('canary_payload_device_types') == []
        assert check['cert_service_time_changed']
        assert not check['policy_service_time_changed']

    def test_14d_environment_partial_3(self):
        """Validate Env() initialization with partial data.

        :class: internal
        """

        target = Env(policy_service_time_changed=True)
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check.get('canary_payload_device_types') == []
        assert not check['cert_service_time_changed']
        assert check['policy_service_time_changed']

    # noinspection PyTypeChecker,PyArgumentList
    def test_14e_environment_bad(self):
        """Verify that bad entries are trapped on Env init.

        :class: internal, negative
        """
        with pytest.raises(AssertionError):  # 'cert_service_time_changed' not bool
            Env(cert_service_time_changed="True")
        with pytest.raises(AssertionError):  # 'policy_service_time_changed' not bool
            Env(policy_service_time_changed="True")
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' not list
            Env(canary_payload_device_types=DeviceType.WINDOWS)
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' not list of DeviceType
            Env(canary_payload_device_types=["IOS"])
        with pytest.raises(AssertionError):  # 'canary_payload_device_types' len > max
            Env(canary_payload_device_types=DeviceType.all() + [DeviceType.IOS])

    # noinspection PyTypeChecker,PyArgumentList
    def test_14f_environment_allowed_empty(self):
        """Verify that allowed empty cases produce no error.

        :class: internal, negative
        """
        Env()
        Env(canary_payload_device_types=[])

    def test_14g_envionment_vary(self):
        """Validate Env().vary does not produce bad entries.

        :class: internal
        """
        target = Env.full()
        with self.Timer(tag=f"100 environment variants"):
            for _ in range(1000):
                target.vary()
                try:
                    target.is_valid()
                except AssertionError as err:
                    _logger.warning(f"is_valid failed with: {target.to_dict()}:\n  {err}")
                    raise

    def test_20a_hint_request(self):
        """Validate HintRequest() initialization.

        :class: internal
        """
        orgkey = encode_orgkey("dev01", 42)
        target = HintRequest(org_key=orgkey, device_group_id=1889, device_id=2300,
                             device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                             version=1, last_update_time=self.pinned(0), last_live_query_time=self.pinned(1),
                             remote_rules_required=False)
        assert len(target.__dict__) == 12  # check for changes in structure
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['device_group_id'] == 1889
        assert check['device_id'] == 2300
        assert check['device_type'] == DeviceType.WINDOWS.name
        assert check['deployment_type'] == DeploymentType.ENDPOINT.name
        assert check['last_live_query_time'] == 937180860000
        assert check['last_update_time'] == 937180800000
        assert check['org_key'] == orgkey
        assert not check['remote_rules_required']
        assert check['version'] == 1

    # noinspection PyArgumentList,PyTypeChecker
    def test_20b_hint_request_bad(self):
        """Verify that bad entries are trapped on HintRequest init.

        :class: internal, negative
        """
        with pytest.raises(TypeError):  # missing device_group_id
            HintRequest(device_id=1, last_live_query_time=self.stamp(), last_update_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)
        with pytest.raises(AssertionError):  # invalid device_group_id
            HintRequest(device_group_id="x", device_id=1, last_live_query_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=1)
        with pytest.raises(TypeError):  # missing device_id
            HintRequest(device_group_id=1, last_live_query_time=self.stamp(), last_update_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)
        with pytest.raises(AssertionError):  # invalid device_id
            HintRequest(device_group_id=1, device_id="x", last_live_query_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=1)

        # NOTE: last_live_query_time is optional

        with pytest.raises(AssertionError):  # invalid last_live_query_time
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=0, last_update_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)
        with pytest.raises(TypeError):  # missing last_update_time
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)
        with pytest.raises(AssertionError):  # invalid last_update_time
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(), last_update_time=0,
                        device_type=DeviceType.WINDOWS, deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)

        # NOTE: device_type is optional

        with pytest.raises(AssertionError):  # invalid device_type
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(), last_update_time=self.stamp(),
                        device_type="WINDOWS", deployment_type=DeploymentType.ENDPOINT,
                        org_key="test", remote_rules_required=True,
                        version=1)

        # NOTE: deployment_type is optional

        with pytest.raises(AssertionError):  # invalid deployment_type
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(), last_update_time=self.stamp(),
                        device_type=DeviceType.WINDOWS, deployment_type="ENDPOINT",
                        org_key="test", remote_rules_required=True,
                        version=1)

        # NOTE: message_code removed (DSER-45237)

        with pytest.raises(TypeError):  # missing org_key
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        remote_rules_required=True,
                        version=1)
        with pytest.raises(AssertionError):  # invalid org_key
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key=42, remote_rules_required=True,
                        version=1)

        # NOTE: remote_rules_required is optional

        with pytest.raises(AssertionError):  # invalid remote_rules_required
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required="True",
                        version=1)
        with pytest.raises(TypeError):  # missing version
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True)
        with pytest.raises(AssertionError):  # invalid version
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version="nope")

    def test_20c_hint_request_full_audit(self):
        """Validate HintRequest() initialization with full audit info.

        :class: internal
        """
        orgkey = encode_orgkey("dev01", 42)
        target = HintRequest(org_key=orgkey, device_group_id=1889, device_id=2300,
                             version=1, last_update_time=self.pinned(0), last_live_query_time=self.pinned(1),
                             remote_rules_required=False,
                             audit_hints={HintType.NEW_CONFIG_DATA, HintType.SOFTWARE_UPGRADE},
                             audit_last_update_time=self.pinned(0), audit_last_live_query_time=self.pinned(1)
                             )
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['device_group_id'] == 1889
        assert check['device_id'] == 2300
        assert check['last_live_query_time'] == 937180860000
        assert check['last_update_time'] == 937180800000
        assert check['org_key'] == orgkey
        assert not check['remote_rules_required']
        assert check['version'] == 1
        assert check['audit_hints'] == sorted(["NEW_CONFIG_DATA", "SOFTWARE_UPGRADE"])
        assert check['audit_last_live_query_time'] == 937180860000
        assert check['audit_last_update_time'] == 937180800000

    def test_20d_hint_request_minimal_audit(self):
        """Validate HintRequest() initialization with full audit info.

        :class: internal
        """
        orgkey = encode_orgkey("dev01", 42)
        target = HintRequest(org_key=orgkey, device_group_id=1889, device_id=2300,
                             version=1, last_update_time=self.pinned(0), last_live_query_time=self.pinned(1),
                             remote_rules_required=False,
                             audit_hints=set(),
                             )
        check = target.to_dict()
        _logger.info(f"Data = {check}")

        assert check['device_group_id'] == 1889
        assert check['device_id'] == 2300
        assert check['last_live_query_time'] == 937180860000
        assert check['last_update_time'] == 937180800000
        assert check['org_key'] == orgkey
        assert not check['remote_rules_required']
        assert check['version'] == 1
        assert check['audit_hints'] == []
        assert check.get('audit_last_live_query_time') is None
        assert check.get('audit_last_update_time') is None

    def test_20e_hint_request_bad_audit(self):
        """Verify that bad audit entries are trapped on HintRequest init.

        :class: internal, negative
        """

        with pytest.raises(AssertionError):
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=2,
                        audit_hints="HASH_LIST_REFRESH")  # audit_hints Not list
        with pytest.raises(AssertionError):
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=2,
                        audit_hints=[23])  # Not string entry in audit_hints list
        with pytest.raises(AssertionError):
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=2,
                        audit_hints=["HASH_LIST_REFRESH"], audit_last_live_query_time=0)  # audit_last_live_query_time Not datetime
        with pytest.raises(AssertionError):
            HintRequest(device_group_id=1, device_id=1, last_live_query_time=self.stamp(),
                        last_update_time=self.stamp(),
                        org_key="test", remote_rules_required=True,
                        version=2,
                        audit_hints=["HASH_LIST_REFRESH"], audit_last_update_time=0)  # audit_last_update_time Not datetime

    def test_21a_hint_result_no_hints(self):
        """
        Verify that HintResult objects with no hints and required data are accepted and translate properly.

        :class: internal
        """
        target = {'hints': [], 'last_live_query_time': 937180800000,
                  'last_update_time': 937180860000}
        converted = HintResult.from_dict(target)

        assert converted.hints == set()
        assert converted.last_live_query_time == self.pinned(0)
        assert converted.last_update_time == self.pinned(1)

    def test_21b_hint_result(self):
        """
        Verify that HintResult objects with some hints and required data are accepted and translate properly.

        :class: internal
        """
        target = {'hints': ["NEW_CONFIG_DATA", "HASH_LIST_REFRESH"], 'last_live_query_time': 937180800000,
                  'last_update_time': 937180860000}
        converted = HintResult.from_dict(target)

        assert converted.hints == {HintType.NEW_CONFIG_DATA, HintType.HASH_LIST_REFRESH}
        assert converted.last_live_query_time == self.pinned(0)
        assert converted.last_update_time == self.pinned(1)

    def test_21c_hint_result_bad(self):
        """Verify that bad entries are trapped on HintResult conversion.

        :class: internal, negative
        """
        with pytest.raises(KeyError):  # No hints param
            HintResult.from_dict({'last_live_query_time': 937180800000, 'last_update_time': 937180860000})
        with pytest.raises(KeyError):  # Hints param not list of strings
            HintResult.from_dict(
                {'hints': [2], 'last_live_query_time': 937180800000, 'last_update_time': 937180860000})
        with pytest.raises(KeyError):  # Hints param not list of valid strings
            HintResult.from_dict(
                {'hints': ["2"], 'last_live_query_time': 937180800000, 'last_update_time': 937180860000})
        with pytest.raises(TypeError):  # last_update_time not time string
            HintResult.from_dict({'hints': ["NEW_CONFIG_DATA", "HASH_LIST_REFRESH"],
                                  'last_live_query_time': '1999-09-13T00:00:00.000000Z',
                                  'last_update_time': 937180860000})
        with pytest.raises(TypeError):  # last_live_query_time not time string
            HintResult.from_dict(
                {'hints': ["NEW_CONFIG_DATA", "HASH_LIST_REFRESH"], 'last_live_query_time': 937180860000,
                 'last_update_time': '1999-09-13T00:00:00.000000Z'})
        with pytest.raises(TypeError):  # last_update_time not incorrect time format
            HintResult.from_dict(
                {'hints': ["NEW_CONFIG_DATA", "HASH_LIST_REFRESH"],
                 'last_live_query_time': '1999-09-13T00:00:00.000000Z',
                 'last_update_time': '1999-09-13'})
