#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import logging
from datetime import datetime, timezone

import pytest
import time

from lib.connect.hint_connect import HintControlConnect
from lib.test_common import TestCommon

_logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class TestUtilities(TestCommon):
    """
    This class validates the test harness common utilities.
    """

    # -------------------------------------------------------------------------------- #
    # ----- Begin Tests -------------------------------------------------------------- #
    # -------------------------------------------------------------------------------- #

    def test_01_unique(self):
        """Validate unique always returns a unique string.

        :class: internal
        """
        uniq1 = self.unique("foo")
        uniq2 = self.unique("foo")
        assert uniq1 != uniq2
        assert uniq1.startswith("foo")
        assert uniq2.startswith("foo")

    def test_02_uid(self):
        """Validate uid always returns a unique int.

        :class: internal
        """
        uid1 = self.uid()
        uid2 = self.uid()
        assert uid1 != uid2

    def test_03_stamp(self):
        """Validate stamp always returns a datetime limited to milliseconds.

        :class: internal
        """
        now = datetime.now(timezone.utc)
        check = self.stamp(22)

        the_diff = check - now
        assert 1321 >= the_diff.seconds >= 1319  # 22 minutes, depending one where the current time rounds
        assert check.microsecond == int(check.microsecond / 1000) * 1000

    def test_04_epoch(self):
        """Validate epoch always returns the epoch date.

        :class: internal
        """
        check = self.epoch()
        assert check.year == 1970
        assert check.month == 1
        assert check.day == 1
        assert check.hour == 0
        assert check.minute == 0
        assert check.second == 0
        assert check.microsecond == 0

    def test_05_pinned(self):
        """Validate pinned always returns a specific date.

        :class: internal
        """
        check = self.pinned(22)
        assert check.year == 1999
        assert check.month == 9
        assert check.day == 13
        assert check.hour == 0
        assert check.minute == 22
        assert check.second == 0
        assert check.microsecond == 0

    def test_06a_verify_response_invalid(self):
        """Validate verify_response complains if you don't provide an actual HintControlConnect object.

        :class: internal, negative
        """
        with pytest.raises(TypeError):
            self.verify_response(TestCommon, 200)

    def test_06b_verify_response_unused(self):
        """Validate nothing happens if there is no last response code (new, unused connection).

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        self.verify_response(hcc, 200)

    def test_06c_verify_response_code(self):
        """Validate response code matct (no message).

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        self.verify_response(hcc, 200)

    def test_06d_verify_response_code(self):
        """Validate response code mismatch.

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 400
        with pytest.raises(self.PYTEST_FAIL):
            self.verify_response(hcc, 200)

    def test_06e_verify_response_message_no_last_content(self):
        """Validate we fail if expecing message but none was returned.

        :class: internal, negative
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        with pytest.raises(self.PYTEST_FAIL):
            self.verify_response(hcc, 200, "message somewhere")

    def test_06f_verify_response_message(self):
        """Validate response message matches "message" item.

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        hcc.last_content = {"message": "Check for message somewhere in here!"}
        self.verify_response(hcc, 200, "message somewhere")

    def test_06g_verify_response_message_no_last_message(self):
        """Validate we catch lack of "message" (probably returned data)

        :class: internal, negative
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        hcc.last_content = {"data": "No message, this was good data!"}
        with pytest.raises(self.PYTEST_FAIL):
            self.verify_response(hcc, 200, "message somewhere")

    def test_06h_verify_response_message_incorrect(self):
        """Validate we catch message that does not match.

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        hcc.last_content = {"message": "No message, this will not match"}
        with pytest.raises(self.PYTEST_FAIL):
            self.verify_response(hcc, 200, "message somewhere")

    def test_06i_verify_response_last_content_incorrect_format(self):
        """Validate we catch a non-dict format (should never happen).

        :class: internal
        """
        hcc = HintControlConnect("localhost", 80)
        hcc.last_status = 200
        hcc.last_content = "No message, message somewhere this will not match"
        with pytest.raises(self.PYTEST_FAIL):
            self.verify_response(hcc, 200, "message somewhere")

    def test_07_is_now(self):
        """Validate is_now() timestamp strings are noted as the same within a given threshold.

        :class: internal
        """
        then = self.stamp()
        time.sleep(4)  # wait 4 seconds, within default threshold of 5
        assert self.is_now(then)

        time.sleep(4)
        assert not self.is_now(then)

