#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging
from typing import Optional, Dict, Any, Union

from lib.models.base_model import BaseModel
from lib.models.device import Device
from lib.models.device_group import DeviceGroup
from lib.models.environment import Env
from lib.models.hinting import HintRequest, HintResult
from lib.models.organization import Org
from .base_connect import BaseConnect

_logger = logging.getLogger(__name__)

# routing basics
ROOT = "/hinting"
V1 = f"{ROOT}/v1"


class HintControlConnect(BaseConnect):
    """
    Class to handle rest calls that configure or modify the hinting service.
    """

    @staticmethod
    def _convert_data(source: Any) -> Any:
        """Implement model or dict as data for API call."""
        if isinstance(source, BaseModel):
            return source.to_dict()  # actual data structure
        elif isinstance(source, Dict):
            return source
        else:
            _logger.debug(f"NOTE: Using {type(source)} for converted data!")
            return source  # raw dict or other data for negative testing

    @staticmethod
    def _get_field(source: Union[Dict, BaseModel], key: str) -> Optional[Any]:
        """Get data field from model or dict, or nothing if anythjing else."""
        if isinstance(source, BaseModel):
            return getattr(source, key)  # get from model
        elif isinstance(source, Dict):
            return source.get(key)  # get from dict
        else:
            # noinspection PyBroadException
            try:
                return getattr(source, key)  # get from object, if possible
            except Exception:
                return None  # no field to get

    # ----- Database Endpoints --------------------

    def v1_device_update(self, the_data: Device, org_key: str, device_id: int) -> Optional[Dict]:
        """
        Update a device entry in the database.

        :param the_data: Device() object with data, or raw dict for negative testing
        :param org_key: organization id
        :param device_id: device id
        :return: No return at present
        """
        return self._put(f"{V1}/orgs/{org_key}/devices/{device_id}", self._convert_data(the_data))

    def v1_devicegroup_update(self, the_data: DeviceGroup, org_key: str, group_id: int) -> Optional[Dict]:
        """
        Update a device group entry in the database.

        :param the_data: DeviceGroup() object with data, or raw dict for negative testing
        :param org_key: organization id
        :param group_id: group id
        :return: No return at present
        """
        return self._put(f"{V1}/orgs/{org_key}/devicegroups/{group_id}", self._convert_data(the_data))

    def v1_org_update(self, the_data: Org, org_key: str) -> Optional[Dict]:
        """
        Update an organization entry in the database.

        :param the_data: Org() object with data, or raw dict for negative testing
        :param org_key: organization id
        :return: No return at present
        """
        return self._put(f"{V1}/orgs/{org_key}", self._convert_data(the_data))

    def v1_env_update(self, the_data: Env, env_name: str) -> Optional[Dict]:
        """
        Update an environment entry in the database.

        :param the_data: Env() object with data, or raw dict for negative testing
        :param env_name: environment name
        :return: No return at present
        """
        return self._put(f"{V1}/envs/{env_name}", self._convert_data(the_data))

    # ----- Miscellaneous Endpoints --------------------

    def health_check_get(self, is_ready: bool = None) -> bool:
        """
        Check health of the service.

        :param is_ready: (optional) add isReady check
        :return: True if health ok
        """
        add_is_ready = "" if is_ready is None else f"?isReady={is_ready}"
        # noinspection PyBroadException
        try:
            self._get(f"{ROOT}/healthCheck{add_is_ready}")
            return self.last_status == 200
        except Exception as err:
            _logger.debug(f"Healthcheck saw {err}")
            return False

    # ----- Hinting Endpoints --------------------

    def v1_hints_post(self, the_data: Union[HintRequest, Dict]) -> Optional[HintResult]:
        """
        List the get hinting information for a given device, group and organization.

        :param the_data: HintRequest() object with data, or raw dict for negative testing
        :return: list of hints
        """
        result = self._post(f"{V1}/hints", self._convert_data(the_data))
        if result is None:
            return None
        else:
            return HintResult.from_dict(result)
