#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import logging
from datetime import datetime, timezone, timedelta
from enum import Enum
from typing import Optional, Dict, Any

__all__ = ["BaseModel"]

_logger = logging.getLogger(__name__)


################################################################################
# This module should hold all relevant structures sent to and returned by the
# service.
################################################################################


class BaseModel(object):
    @staticmethod
    def _date_to_hinting_time(source: datetime) -> Optional[int]:
        """Convert datetime to date format expected by the service, if provided.

        NOTE: hinting service uses milliseconds after epoch time as int64
        """
        if source is None:
            return None
        else:
            return int(source.timestamp() * 1000)

    @staticmethod
    def _hinting_time_to_date(source: int) -> Optional[datetime]:
        """Converts epoch milliseconds to a usable datetime, if provided.

        NOTE: hinting service uses milliseconds after epoch time as int64
        """
        if source is None:
            return None
        else:
            return datetime.fromtimestamp(source / 1000.0, tz=timezone.utc)

    def _convert_times_in_dict(self, source: Optional[Dict]) -> Optional[Dict]:
        """Converts dict structures containing timestamps safely."""
        if source is None:
            return None

        cvt = {}
        for k, v in source.items():
            # handle enums
            use_k = k if not isinstance(k, Enum) else k.name

            if isinstance(v, datetime):
                cvt[use_k] = self._date_to_hinting_time(v)
            elif isinstance(v, Dict):
                cvt[use_k] = self._convert_times_in_dict(v)
            else:
                cvt[use_k] = v
        return cvt

    @staticmethod
    def _safe_enum(source: Optional[Enum]) -> Optional[Any]:
        """Safely converts possible Enums for JSON use."""
        if source is None:
            return None
        elif isinstance(source, Enum):
            # return source.value
            return source.name
        else:
            if not isinstance(source, str):
                _logger.debug(f"NOTE: supplied `{source}` was not a str!")
            return source

    @staticmethod
    def stamp(delta: int = 0) -> datetime:
        """generate a quick timestamp based on current time, to the nearest millisecond

        :param delta: time variance in minutes (default 0)
        """
        use_datetime = datetime.now(timezone.utc) + timedelta(minutes=delta)
        try:
            ms = round(max(0, min(use_datetime.microsecond, 999999)), -3)
            return use_datetime.replace(microsecond=ms)
        except Exception as err:
            _logger.error(
                f"micro={use_datetime.microsecond} -> {err}")  # Something odd pops up spurriously, log for now to try and figure what
        return use_datetime

    def to_dict(self):
        raise NotImplementedError("Must be implemented in child classes.")

    @staticmethod
    def from_dict(source: Dict) -> 'BaseModel':
        raise NotImplementedError("Must be implemented in child classes.")
