#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

from datetime import datetime
from typing import Dict, List

from .base_model import BaseModel
from .enumerations import DeviceChangeType

__all__ = ["Device"]


class Device(BaseModel):
    """
    Representation of the API body structure expected on a device update.

        See: `APIDeviceBody` in internal/app/models/device.go
    """

    def __init__(self,
                 changes: List[DeviceChangeType] = None,
                 live_query_time: datetime = None,
                 ):
        """Intiatialize a Device structure."""
        self.changes = changes if changes is not None else []
        self.live_query_time = live_query_time

        # Validate
        self.is_valid()

    def is_valid(self) -> None:
        assert isinstance(self.changes, List)
        assert len(self.changes) <= len(DeviceChangeType)
        for item in self.changes:
            assert isinstance(item, DeviceChangeType)
        assert self.live_query_time is None or isinstance(self.live_query_time, datetime)

    @staticmethod
    def full(no_live_query: bool = False) -> 'Device':
        """
        Return a fully filled in device.

        :param no_live_query: If True, do not specify a live query time (any DEV_LIVE_QUERY_CHANGE entry
                              will use current time on update)
        """
        if not isinstance(no_live_query, bool):
            raise TypeError("`no_live_query` must be a bool!")
        lqt = None if no_live_query else Device.stamp()

        dev = Device(DeviceChangeType.all(), live_query_time=lqt)
        return dev

    def vary(self) -> None:
        """
        Vary the changes for a given device.
        """
        while True:
            self.changes = DeviceChangeType.random()  # Always defined
            if len(self.changes) > 0:
                break

    # ------------------------------

    def to_dict(self) -> Dict:
        """Marshall to a dict suitable for JSON conversion."""
        output = {
            "changes": [e.name for e in self.changes],
            "live_query_time": self._date_to_hinting_time(self.live_query_time),
        }
        return output

    def from_dict(self):
        # FUTURE: implement if we ever have a Get call
        raise NotImplementedError()
