#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

from datetime import datetime
from typing import Dict, List

from .base_model import BaseModel
from .enumerations import DeviceGroupChangeType

__all__ = ["DeviceGroup"]


class DeviceGroup(BaseModel):
    """
    Representation of the API body structure expected on a device group update.

        See: `APIDeviceGroupBody` in internal/app/models/device_group.go
    """

    def __init__(self,
                 changes: List[DeviceGroupChangeType] = None,
                 live_query_time: datetime = None,
                 ):
        """Intiatialize a DeviceGroup structure."""
        self.changes = changes if changes is not None else []
        self.live_query_time = live_query_time

        # Validate
        self.is_valid()

    def is_valid(self) -> None:
        assert isinstance(self.changes, List)
        assert len(self.changes) <= len(DeviceGroupChangeType)
        for item in self.changes:
            assert isinstance(item, DeviceGroupChangeType)
        assert self.live_query_time is None or isinstance(self.live_query_time, datetime)

    @staticmethod
    def full(no_live_query: bool = False) -> 'DeviceGroup':
        """
        Return a fully filled in device group.

        :param no_live_query: If True, do not specify a live query time (any DG_LIVE_QUERY_CHANGE entries
                              will use current time on update)
        """
        if not isinstance(no_live_query, bool):
            raise TypeError("`no_live_query` must be a bool!")
        lqt = None if no_live_query else DeviceGroup.stamp()

        grp = DeviceGroup(DeviceGroupChangeType.all(), live_query_time=lqt)
        return grp

    def vary(self) -> None:
        """
        Vary the changes for a given device group.
        """
        while True:
            self.changes = DeviceGroupChangeType.random()  # Always defined
            if len(self.changes) > 0:
                break

    # ------------------------------

    def to_dict(self) -> Dict:
        """Marshall to a dict suitable for JSON conversion."""
        output = {
            "changes": [e.name for e in self.changes],
            "live_query_time": self._date_to_hinting_time(self.live_query_time),
        }
        return output

    def from_dict(self):
        raise NotImplementedError()
