#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import random
from enum import Enum
from typing import List, Dict, Optional, Union

__all__ = ["DeploymentType", "DeviceType", "HintType", "OrgChangeType", "DeviceChangeType",
           "DeviceGroupChangeType"]


class DeviceChangeType(Enum):
    """
    Enumeration of valid device change types
        See: internal/app/models/enums/devicechangetype/devicechangetype.go

    NOTE: This is the subset from the backend definition that is actually used for hinting
    """
    DEV_GROUP_CHANGE = 301
    DEV_HASH_SAMPLE_REQUEST = 302
    DEV_SENSOR_LOGS_REQUEST = 303
    DEV_HASH_DELETE_REQUEST = 304
    DEV_SENSOR_CERTIFICATE_REQUEST = 306
    DEV_SENSOR_CONFIG_UPDATE = 307
    DEV_SENSOR_TRIGGER_PERFMON = 308
    DEV_SENSOR_KIT_UPDATE = 309
    DEV_SENSOR_ACTION = 310
    DEV_LIVE_QUERY_CHANGE = 317
    DEV_REREGISTER = 321
    DEV_REP_REQUEST = 331

    @staticmethod
    def all() -> List['DeviceChangeType']:
        """Return list of all DeviceChangeType."""
        return [e for e in DeviceChangeType]

    @staticmethod
    def random() -> List['DeviceChangeType']:
        """Return a list of randomly picked DeviceChangeType"""
        ret = []
        for item in DeviceChangeType.all():
            if random.choice([True, False]):
                ret.append(item)
        return ret


class DeviceGroupChangeType(Enum):
    """
    Enumeration of valid device group change types
        See: internal/app/models/enums/devicegroupchangetype/devicegroupchangetype.go

    NOTE: This is the subset from the backend definition that is actually used for hinting
    """
    DG_POLICY_CHANGE = 201
    DG_LIVE_QUERY_CHANGE = 203

    @staticmethod
    def all() -> List['DeviceGroupChangeType']:
        return [e for e in DeviceGroupChangeType]

    @staticmethod
    def random() -> List['DeviceGroupChangeType']:
        """Return a list of randomly picked DeviceGroupChangeType"""
        ret = []
        for item in DeviceGroupChangeType.all():
            if random.choice([True, False]):
                ret.append(item)
        return ret


class DeviceType(int, Enum):
    """
    Enumeration of valid device types
        See: internal/app/models/enums/devicetype/devicetype.go

    NOTE: "UNKNOWN" is not used
    """
    WINDOWS = 1
    ANDROID = 2
    MAC = 3
    IOS = 4
    LINUX = 5
    OTHER = 10

    @staticmethod
    def all(as_string: bool = False) -> Union[List['DeviceType'], List[str]]:
        if as_string:
            return [e.name for e in DeviceType]
        else:
            return [e for e in DeviceType]

    @staticmethod
    def random() -> List['DeviceType']:
        """Return a list of randomly picked DeviceType"""
        ret = []
        for item in DeviceType.all():
            if random.choice([True, False]):
                ret.append(item)
        return ret


class HintType(Enum):
    """
    Enumeration of valid hint types returned
        See: internal/hints/hinttype/hinttype.go
    """
    NEW_POLICY = 0

    # NOTE: There is no hint type 1

    NEW_CONFIG_DATA = 2
    HASH_LIST_REFRESH = 3
    HASH_LIST_UPLOAD = 4
    CONFER_UPLOAD = 5
    INVALIDATE_NON_WHITELIST_REP = 6
    STATUS_UPDATE = 7
    APC_UPLOAD = 8
    RESUBMIT = 9
    REQUEST_NEWS = 10
    VT_ANALYSIS_UPLOAD = 11
    SOFTWARE_UPGRADE = 12
    SOFTWARE_PATCH = 13
    YARA_AV_SIGNATURES = 14
    YARA_NETWORK_SIGNATURES = 15
    YARA_PE_HEADER_SIGNATURES = 16
    IP_BLOCK_LIST = 17
    IP_FORENSICS_LIST = 18
    HASH_FORENSICS_LIST = 19
    ONETIME_HASH_DELETE_LIST = 20
    PERMANENT_HASH_DELETE_LIST = 21
    SENSOR_CERTIFICATE_UPLOAD = 22
    SENSOR_TRIGGER_PERFMON = 23
    SENSOR_ACTION = 24
    KILL_SWITCH_ON = 25
    ZIP_CONTAINER = 26
    UNINSTALL_CODE = 27
    LIVE_QUERY = 28
    ASYNCREPUTATION = 29
    BLADE_ACTION = 30
    REREGISTER = 31
    REP_REQUEST = 32


class OrgChangeType(int, Enum):
    """
    Enumeration of valid org change types
        See: internal/app/models/enums/orgchangetype/orgchangetype.go
    """
    ORG_NET_CFG_CHANGE = 101
    ORG_HASH_LIST_CHANGE = 102
    ORG_AV_INTEGRATION_CHANGE = 108
    ORG_KILL_SWITCH_ON = 109
    ORG_UNINSTALL_CODE_CHANGE = 110
    ORG_LIVE_QUERY_CHANGE = 111
    ORG_SET_MELTDOWN_REGISTRY_KEY = 112
    ORG_THREATHUNTER_RULES = 113
    ORG_DISABLE_DEFENSE_RULES = 114
    ORG_BLADE_ACTION_CHANGE = 115

    @staticmethod
    def all(as_value: bool = False) -> Union[List['OrgChangeType'], List[str]]:
        if as_value:
            return [f"{e.value}" for e in OrgChangeType]
        else:
            return [e for e in OrgChangeType]

    @staticmethod
    def random() -> List['OrgChangeType']:
        """Return a list of randomly picked OrgChangeType"""
        ret = []
        for item in OrgChangeType.all():
            if random.choice([True, False]):
                ret.append(item)
        return ret


# NOTE: DeploymentType Must be defined after OrgChangeType due to `all_dict_org()` and `random_dict()`
class DeploymentType(Enum):
    """
    Enumeration of valid deployment types
        See: internal/app/models/enums/deploymenttype

    NOTE: "UNKNOWN" is not used
    """
    ENDPOINT = 1
    WORKLOAD = 2
    VDI = 3
    AWS = 4
    AZURE = 5
    GCP = 6

    @staticmethod
    def all(as_string: bool = False) -> Union[List['DeploymentType'], List[str]]:
        """Return list of all DeploymentType."""
        if as_string:
            return [e.name for e in DeploymentType]
        else:
            return [e for e in DeploymentType]

    @staticmethod
    def all_dict_org() -> Dict['DeploymentType', List[OrgChangeType]]:
        """Return map of all DeploymentType and OrgChangeType."""
        ret = {}
        for item in [e for e in DeploymentType]:
            ret[item] = OrgChangeType.all()
        return ret

    @staticmethod
    def random() -> List['DeploymentType']:
        """Return a list of randomly picked OrgChangeType"""
        ret = []
        for item in DeploymentType.all():
            if random.choice([True, False]):
                ret.append(item)
        return ret

    @staticmethod
    def random_dict() -> Optional[Dict['DeploymentType', List[OrgChangeType]]]:
        """Return a list of randomly picked DeploymentType and OrgChangeType."""
        ret = {}
        for item in DeploymentType.random():
            ret[item] = OrgChangeType.random()
        if len(ret) > 0:
            return ret
        else:
            return None
