#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import random
from typing import Dict, List

from .base_model import BaseModel
from .enumerations import DeviceType

__all__ = ["Env"]


class Env(BaseModel):
    """
    Representation of the API body structure expected on an environment update.

        See: `APIEnvBody` in internal/app/models/environment.go
    """

    def __init__(self,
                 canary_payload_device_types: List[DeviceType] = None,
                 cert_service_time_changed: bool = None,
                 policy_service_time_changed: bool = None,
                 ):
        """Intiatialize an Env structure."""
        self.canary_payload_device_types = canary_payload_device_types if canary_payload_device_types is not None else []
        self.cert_service_time_changed = cert_service_time_changed
        self.policy_service_time_changed = policy_service_time_changed

        # Validate
        self.is_valid()

    def is_valid(self) -> None:
        assert isinstance(self.canary_payload_device_types, List)
        assert len(self.canary_payload_device_types) <= len(DeviceType)
        for item in self.canary_payload_device_types:
            assert isinstance(item, DeviceType)
        assert self.cert_service_time_changed is None or isinstance(self.cert_service_time_changed, bool)
        assert self.policy_service_time_changed is None or isinstance(self.policy_service_time_changed, bool)

    @staticmethod
    def full() -> 'Env':
        """Return a fully filled in device group."""
        env = Env(canary_payload_device_types=DeviceType.all(), cert_service_time_changed=True,
                  policy_service_time_changed=True)
        return env

    def vary(self) -> None:
        """
        Vary the changes for a given environment.
        """
        while True:
            self.canary_payload_device_types = DeviceType.random() or []
            self.cert_service_time_changed = random.choice([True, False])
            self.policy_service_time_changed = random.choice([True, False])
            if self.cert_service_time_changed or self.policy_service_time_changed or self.canary_payload_device_types:
                break

    # ------------------------------

    def to_dict(self) -> Dict:
        """Marshall to a dict suitable for JSON conversion."""
        output = {
            "cert_service_time_changed": self.cert_service_time_changed,
            "policy_service_time_changed": self.policy_service_time_changed,
        }
        if self.canary_payload_device_types is not None:
            output["canary_payload_device_types"] = [e.name for e in self.canary_payload_device_types]
        return output

    def from_dict(self):
        raise NotImplementedError()
