#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

from datetime import datetime
from typing import Dict, Set, Optional

from .base_model import BaseModel
from .enumerations import HintType, DeploymentType, DeviceType

__all__ = ["HintRequest", "HintResult"]


class HintRequest(BaseModel):
    """
    See: internal/hints/hint_request.go
        type jsonHintRequest struct {
            OrgKey              string  `json:"org_key" binding:"required"`
            DeviceGroupId       int64   `json:"device_group_id" binding:"min=1"`
            DeviceId            int64   `json:"device_id" binding:"min=1"`
            DeviceType          string  `json:"device_type"`
            DeploymentType      string  `json:"deployment_type"`
            Version             *int    `json:"version"`
            LastUpdateTime      *int64  `json:"last_update_time"`
            LastLiveQueryTime   *int64  `json:"last_live_query_time"`
            RemoteRulesRequired bool    `json:"remote_rules_required"`
            AuditHintsFromCBD
        }

        type AuditHintsFromCBD struct {
            AuditHints                  []string `json:"audit_hints"`
            AuditLastUpdateTimestamp    *int64   `json:"audit_last_update_time"`
            AuditLastLiveQueryTimestamp *int64   `json:"audit_last_live_query_time"`
        }

        NOTE: in Go, 'AuditHintsFromCBD' is basically appended to the bottom of 'jsonHintRequest'
    """

    def __init__(self,
                 org_key: str,
                 device_group_id: int,
                 device_id: int,
                 version: int,
                 last_update_time: datetime,
                 device_type: DeviceType = None,
                 deployment_type: DeploymentType = None,
                 last_live_query_time: datetime = None,
                 remote_rules_required: bool = None,

                 # NOTE: If audit_hints is None, audit time entries are ignored
                 audit_hints: Set[HintType] = None,
                 audit_last_update_time: datetime = None,
                 audit_last_live_query_time: datetime = None
                 ):
        """Intiatialize a HintRequest structure.

        All parameters are required
        """
        self.org_key = org_key
        self.device_group_id = device_group_id
        self.device_id = device_id
        self.device_type = device_type
        self.deployment_type = deployment_type
        self.version = version
        self.last_update_time = last_update_time
        self.last_live_query_time = last_live_query_time
        self.remote_rules_required = remote_rules_required

        self.audit_hints = audit_hints
        self.audit_last_update_time = audit_last_update_time
        self.audit_last_live_query_time = audit_last_live_query_time

        # Validate
        self.is_valid()

    def is_valid(self) -> None:
        assert isinstance(self.device_group_id, int)
        assert isinstance(self.device_id, int)
        assert self.deployment_type is None or isinstance(self.deployment_type, DeploymentType)
        assert self.device_type is None or isinstance(self.device_type, DeviceType)
        assert isinstance(self.last_update_time, datetime)
        assert self.last_live_query_time is None or isinstance(self.last_live_query_time, datetime)
        assert isinstance(self.org_key, str)
        assert self.remote_rules_required is None or isinstance(self.remote_rules_required, bool)
        assert isinstance(self.version, int)

        if self.audit_hints is not None:
            assert isinstance(self.audit_hints, Set)
            for item in self.audit_hints:
                assert isinstance(item, HintType)
        if self.audit_last_update_time is not None:
            assert isinstance(self.audit_last_update_time, datetime)
        if self.audit_last_live_query_time is not None:
            assert isinstance(self.audit_last_live_query_time, datetime)

    # ------------------------------

    def to_dict(self) -> Dict:
        """Return the HintRequest object as a dict in the order stated in the Go struct."""
        output = {"device_group_id": self.device_group_id,
                  "device_id": self.device_id,
                  "last_update_time": self._date_to_hinting_time(self.last_update_time),
                  "org_key": self.org_key,
                  "version": self.version,
                  }
        if self.device_type is not None:
            output["device_type"] = self._safe_enum(self.device_type)
        if self.deployment_type is not None:
            output["deployment_type"] = self._safe_enum(self.deployment_type)
        if self.remote_rules_required is not None:
            output["remote_rules_required"] = self.remote_rules_required
        if self.last_live_query_time is not None:
            output["last_live_query_time"] = self._date_to_hinting_time(self.last_live_query_time)

        if self.audit_hints is not None:
            hints = []
            for item in self.audit_hints:
                hints.append(item.name)
            output['audit_hints'] = sorted(hints)
            if self.audit_last_update_time is not None:
                output['audit_last_update_time'] = self._date_to_hinting_time(self.audit_last_update_time)
            if self.audit_last_live_query_time is not None:
                output['audit_last_live_query_time'] = self._date_to_hinting_time(self.audit_last_live_query_time)
        return output

    def from_dict(self):
        raise NotImplementedError()

    def update_times(self, last_contact: Optional[datetime], live_query: Optional[datetime]) -> None:
        """
        Convenience function to update the last contact and live query times in a previously used HintRequest object,
        presumably from the most recent hinting call based on that object.

        Timestamps are only updated if not None

        :param last_contact: timestamp
        :param live_query: timestamp
        """
        if last_contact is not None:
            self.last_update_time = last_contact
        if live_query is not None:
            self.last_live_query_time = live_query


################################################################################

class HintResult(BaseModel):
    """
    See: internal/hints/hint_result.go
        type HintResult struct {
            Hints             *hinttype.HintSet `json:"hints"`
            LastLiveQueryTime *time.Time        `json:"last_live_query_time"`
            LastUpdateTime    *time.Time        `json:"last_update_time"`
        }
    """

    def __init__(self,
                 hints: Set[HintType],
                 last_live_query_time: datetime,
                 last_update_time: datetime,
                 ):
        """Intiatialize a HintResult structure.

        No validation -- this is purely initialization; use the 'from_dict()' method to initialize from source.
        """
        self.hints = hints
        self.last_live_query_time = last_live_query_time
        self.last_update_time = last_update_time

    def __str__(self) -> str:
        hints = [h.name for h in self.hints]
        return f"last_live_query_time: {self.last_live_query_time}, last_update_time: {self.last_update_time}, " + \
            f"{sorted(hints)}"

    # ------------------------------

    def to_dict(self):
        raise NotImplementedError()

    @staticmethod
    def from_dict(source: Dict) -> 'HintResult':
        llt = HintRequest._hinting_time_to_date(source.get("last_live_query_time"))
        lut = HintRequest._hinting_time_to_date(source.get("last_update_time"))
        hints = set()
        for item in source['hints']:  # assume always return empty list for no hints
            hints.add(HintType[item])
        newobj = HintResult(hints, llt, lut)
        return newobj
