#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################

import random
from datetime import datetime
from typing import Dict, List

from .base_model import BaseModel
from .enumerations import DeploymentType, DeviceType, OrgChangeType

__all__ = ["Org"]


class Org(BaseModel):
    """
    Representation of the API body structure expected on an organization update.

        See: `APIOrganizationBody` in internal/app/models/organization.go
    """

    def __init__(self,
                 canary_payload_device_types: List[DeviceType] = None,
                 changes: List[OrgChangeType] = None,
                 deployment_type_changes: Dict[DeploymentType, List[OrgChangeType]] = None,
                 is_kill_switch_on: bool = False,
                 live_query_time: datetime = None,
                 ):
        """Intiatialize an Organization structure."""
        self.canary_payload_device_types = canary_payload_device_types
        self.changes = changes
        self.deployment_type_changes = deployment_type_changes
        self.is_kill_switch_on = is_kill_switch_on
        self.live_query_time = live_query_time

        # Validate
        self.is_valid()

    def is_valid(self) -> None:
        if self.canary_payload_device_types is not None:
            assert isinstance(self.canary_payload_device_types, List)
            assert len(self.canary_payload_device_types) <= len(DeviceType)
            for item in self.canary_payload_device_types:
                assert isinstance(item, DeviceType)
        if self.changes is not None:
            assert isinstance(self.changes, List)
            assert len(self.changes) <= len(OrgChangeType)
            for item in self.changes:
                assert isinstance(item, OrgChangeType)
        if self.deployment_type_changes is not None:
            assert isinstance(self.deployment_type_changes, Dict)
            assert len(self.deployment_type_changes) <= len(DeploymentType)
            for key, value in self.deployment_type_changes.items():
                assert isinstance(key, DeploymentType)
                assert isinstance(value, List)
                assert len(value) <= len(OrgChangeType)
                for item in value:
                    assert isinstance(item, OrgChangeType)
        assert isinstance(self.is_kill_switch_on, bool)
        assert self.live_query_time is None or isinstance(self.live_query_time, datetime)

    @staticmethod
    def full(no_live_query: bool = False) -> 'Org':
        """
        Return a fully filled in organization.

        :param no_live_query: If True, do not specify a live query time (any ORG_LIVE_QUERY_CHANGE[111] entries
                              will use current time on update)
        """
        if not isinstance(no_live_query, bool):
            raise TypeError("`no_live_query` must be a bool!")
        lqt = None if no_live_query else Org.stamp()

        org = Org(canary_payload_device_types=DeviceType.all(), changes=OrgChangeType.all(),
                  deployment_type_changes=DeploymentType.all_dict_org(),
                  is_kill_switch_on=True, live_query_time=lqt)
        return org

    def vary(self) -> None:
        """
        Vary the changes for a given organization.
        """
        while True:
            self.canary_payload_device_types = DeviceType.random() or None
            self.changes = OrgChangeType.random() or None
            self.deployment_type_changes = DeploymentType.random_dict() or None
            self.is_kill_switch_on = random.choice([True, False])
            ok = False
            if self.is_kill_switch_on or self.canary_payload_device_types or self.changes or self.deployment_type_changes:
                ok = True
            if self.deployment_type_changes is not None:
                for k, v in self.deployment_type_changes.items():
                    if len(v) == 0:
                        ok = False
            if ok:
                break

    # ------------------------------

    def to_dict(self) -> Dict:
        """Marshall to a dict suitable for JSON conversion."""
        output = {
            "is_kill_switch_on": self.is_kill_switch_on,
        }
        if self.canary_payload_device_types is not None:
            output["canary_payload_device_types"] = [e.name for e in self.canary_payload_device_types]
        if self.changes is not None:
            output["changes"] = [e.name for e in self.changes]
        if self.deployment_type_changes is not None:
            dtc = {}
            for key in self.deployment_type_changes.keys():
                dtc[key.name] = [e.name for e in self.deployment_type_changes[key]]
            output["deployment_type_changes"] = dtc
        if self.live_query_time is not None:
            output["live_query_time"] = self._date_to_hinting_time(self.live_query_time)
        return output

    def from_dict(self):
        raise NotImplementedError()
