#  Copyright (c) 2023 VMware Carbon Black, all rights reserved

import logging
import os
from configparser import ConfigParser
from enum import Enum
from typing import List, Tuple

import psycopg2

from lib.configuration import FunctionalTestConfig

_logger = logging.getLogger(__name__)


class PostgresDB(object):
    """
    Core class to handle connections to the postgres used by the hinting service.
    """

    def __init__(self, pghost: str = None, pgport: int = None,
                 pguser: str = None, pgpass: str = None,
                 dbname: str = None):
        self.pghost = pghost
        self.pgport = pgport
        self.pguser = pguser
        self.pgpass = pgpass
        self.dbname = dbname

        self.conn = None
        self.cur = None
        self.usable = False

        # if fields are None, try to fill in with enviornment variables
        self._normalize_env()
        self._normalize_envfile()

        # Uncomment when trying to determine why postgess connection is not working
        # _logger.info(f"dbname={self.dbname}, user={self.pguser}, password={self.pgpass}, host={self.pghost}, port={self.pgport}")

        if self.pghost is None or self.pgport is None or self.pguser is None or self.pgpass is None or self.dbname is None:
            _logger.warning(
                "\n" + "*" * 80 + "\n" + f"Not enough information to connect to postgress database:\n  {self}\n" +
                "*" * 80 + "\n")
        else:
            try:
                self.conn = psycopg2.connect(dbname=self.dbname, user=self.pguser, password=self.pgpass,
                                             host=self.pghost, port=self.pgport)
                self.usable = True
                self.cur = self.conn.cursor()
                _logger.debug(f"Created connection -- {self}")
            except Exception as err:
                _logger.warning(
                    "\n" + "*" * 80 + "\n" + f"Unable to make connection to postgres database:\n  {self}\n" +
                    f"{err}\n" + "*" * 80 + "\n")
                self.usable = False

    def __str__(self) -> str:
        pwmask = "None" if self.pgpass is None else "*****"
        u_flag = "" if self.usable else "     [UNUSABLE]"
        return f"PostgresDB ({self.dbname} @ {self.pghost}:{self.pgport}) -> {self.pguser}/{pwmask}{u_flag}"

    def __del__(self):
        if self.cur is not None:
            self.cur.close()
        if self.conn is not None:
            self.conn.close()

    def _normalize_env(self):
        """
        For any parameters that are still None, try to fill in the blanks with known environment variables.
        """
        if self.pghost is None:
            if "PGHOST" in os.environ:
                self.pghost = os.environ["PGHOST"]
        if self.pgport is None:
            if "PGPORT" in os.environ:
                self.pgport = os.environ["PGPORT"]
            if self.pgport is None and "POSTGRES_PORT_5432_TCP_PORT" in os.environ:  # alternate
                self.pgport = os.environ["POSTGRES_PORT_5432_TCP_PORT"]
        if self.pguser is None:
            if "PGUSER" in os.environ:
                self.pguser = os.environ["PGUSER"]
        if self.pgpass is None:
            if "POSTGRES_PASSWORD" in os.environ:
                self.pgpass = os.environ["POSTGRES_PASSWORD"]
            if self.pgpass is None and "PGPASSWORD" in os.environ:  # alternate
                self.pgpass = os.environ["PGPASSWORD"]
        if self.dbname is None:
            if "PGDATABASE" in os.environ:
                self.dbname = os.environ["PGDATABASE"]

    def _normalize_envfile(self):
        """
        For any parameters that are still None, try to fill in the blanks with known envfile settings.

        NOTE: envfile assumed to be located at the top of the repo (hinting/envfile)
        """
        if self.pghost is not None and self.pgport is not None and self.pguser is not None and self.pgpass is not None and self.dbname is not None:
            return  # nothing to do, everything set
        ftc = FunctionalTestConfig()
        the_file = os.path.join(ftc.REPO_ROOT, "envfile")
        if ftc.ENVFILE is not None:
            the_file = ftc.ENVFILE
        if not os.path.exists(the_file):
            _logger.debug(f"Unable to normalize_envfile: does not exist at '{the_file}'")
            return

        config_object = ConfigParser()
        with open(the_file) as stream:
            config_object.read_string("[DEFAULT]\n" + stream.read())

        if self.pghost is None:
            if config_object.has_option("DEFAULT", "pghost"):
                self.pghost = config_object.get("DEFAULT", "pghost")
        if self.pgport is None:
            if config_object.has_option("DEFAULT", "pgport"):
                self.pgport = int(config_object.get("DEFAULT", "pgport"))
        if self.pguser is None:
            if config_object.has_option("DEFAULT", "pguser"):
                self.pguser = config_object.get("DEFAULT", "pguser")
        if self.pgpass is None:
            if config_object.has_option("DEFAULT", "pgpassword"):
                self.pgpass = config_object.get("DEFAULT", "pgpassword")
        if self.dbname is None:
            if config_object.has_option("DEFAULT", "pgdatabase"):
                self.dbname = config_object.get("DEFAULT", "pgdatabase")

    # ----- Data access methods --------------------------------------------------

    class PgDev(int, Enum):
        id = 0
        org_key = 1
        group_change_time = 2
        sensor_config_update_time = 3
        sensor_certificate_update_time = 4
        hash_delete_request_time = 5
        hash_upload_request_time = 6
        logs_request_time = 7
        sensor_kit_update_time = 8
        perfmon_trigger_time = 9
        sensor_action_time = 10
        re_register_time = 11
        live_query_change_time = 12

    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    def get_device(self, dev_id: int = None, org_key: str = None) -> List[Tuple]:
        """
        Get information from the devices table. Use '.PgDev' enum for clearer field access
        - PgDev.id
        - PgDev.org_key
        - PgDev.group_change_time
        - PgDev.sensor_config_update_time
        - PgDev.sensor_certificate_update_time
        - PgDev.hash_delete_request_time
        - PgDev.hash_upload_request_time
        - PgDev.logs_request_time
        - PgDev.sensor_kit_update_time
        - PgDev.perfmon_trigger_time
        - PgDev.sensor_action_time
        - PgDev.re_register_time
        - PgDev.live_query_change_time

        :param dev_id: optional device id (otherwise all)
        :param org_key: optional org_key (otherwise all)
        :return: List of row entries, as a Tuple (see above)
        """
        if dev_id is not None and not isinstance(dev_id, int):
            raise TypeError("If specified, `dev_id` must be an int!")
        if org_key is not None and not isinstance(org_key, str):
            raise TypeError("If specified, `org_key` must be a str!")

        call = "SELECT * from devices"
        wheres = []
        if dev_id is not None:
            wheres.append(f"id={dev_id}")
        if org_key is not None:
            wheres.append(f"org_key='{org_key}'")
        if len(wheres) > 0:
            call += " where " + " and ".join(wheres)
        _logger.debug(f"PG Call: {call}")

        self.cur.execute(call)
        return self.cur.fetchall()

    class PgGrp(int, Enum):
        id = 0
        org_key = 1
        live_query_change_time = 2
        last_policy_change_time = 3

    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    def get_group(self, grp_id: int = None, org_key: str = None) -> List[Tuple]:
        """
        Get information from the device_groups table. Use '.PgGrp' enum for clearer field access

        - PgGrp.id
        - PgGrp.org_key
        - PgGrp.live_query_change_time
        - PgGrp.last_policy_change_time

        :param grp_id: optional device group id (otherwise all)
        :param org_key: optional org_key
        :return: List of row entries, as a Tuple (see above)
        """
        if grp_id is not None and not isinstance(grp_id, int):
            raise TypeError("If specified, `grp_id` must be an int!")
        if org_key is not None and not isinstance(org_key, str):
            raise TypeError("If specified, `org_key` must be a str!")

        call = "SELECT * from device_groups"
        wheres = []
        if grp_id is not None:
            wheres.append(f"id={grp_id}")
        if org_key is not None:
            wheres.append(f"org_key='{org_key}'")
        if len(wheres) > 0:
            call += " where " + " and ".join(wheres)
        _logger.debug(f"PG Call: {call}")

        self.cur.execute(call)
        return self.cur.fetchall()

    class PgOrg(int, Enum):
        org_key = 0
        is_kill_switch_on = 1
        change_timestamps = 2
        deployment_type_change_timestamps = 3
        canary_payload_timestamps = 4

    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    def get_org(self, org_key: str = None) -> List[Tuple]:
        """
        Get information from the orgs table.  Use '.PgOrg' enum for clearer field access

        - PgOrg.org_key
        - PgOrg.is_kill_switch_on
        - PgOrg.change_timestamps
        - PgOrg.deployment_type_change_timestamps
        - PgOrg.canary_payload_timestamps

        :param org_key: optional org_key (otherwise all)
        :return: List of row entries, as a Tuple (see above)
        """
        if org_key is not None and not isinstance(org_key, str):
            raise TypeError("If specified, `org_key` must be a str!")

        call = "SELECT * from organizations"
        if org_key is not None:
            call += f" where org_key='{org_key}'"
        _logger.debug(f"PG Call: {call}")

        self.cur.execute(call)
        return self.cur.fetchall()

    class PgEnv(int, Enum):
        name = 0
        cert_service_time = 1
        policy_service_update_time = 2
        canary_payload_timestamps = 3

    # noinspection SqlDialectInspection,SqlNoDataSourceInspection
    def get_env(self, name: str = None) -> List[Tuple]:
        """
        Get information from the envs table.  Use '.PgEnv' enum for clearer field access

        - PgEnv.name
        - PgEnv.cert_service_time
        - PgEnv.policy_service_update_time
        - PgEnv.canary_payload_timestamps

        :param name: optional name (otherwise all)
        :return: List of row entries, as a Tuple (see above)
        """
        if name is not None and not isinstance(name, str):
            raise TypeError("If specified, `name` must be a str!")

        call = "SELECT * from environments"
        if name is not None:
            call += f" where name='{name}'"
        _logger.debug(f"PG Call: {call}")

        self.cur.execute(call)
        return self.cur.fetchall()
