#  Copyright (c) 2023 VMware Carbon Black, all rights reserved
import json
import logging
import os
from configparser import ConfigParser
from typing import Dict

import redis

from lib.configuration import FunctionalTestConfig

_logger = logging.getLogger(__name__)


class RedisDB(object):
    """
    Core class to handle connections to the redis used by the hinting service.
    """

    def __init__(self, r_endpoint: str = None, r_pass: str = None):
        self.r_endpoint = r_endpoint
        self.r_port = 6379  # might need variability
        self.r_pass = r_pass

        self.conn = None
        self.usable = False

        # if fields are None, try to fill in with enviornment variables
        self._normalize_env()
        self._normalize_envfile()

        if self.r_endpoint is None or self.r_port is None or self.r_pass is None:
            _logger.warning(
                "\n" + "*" * 80 + "\n" + f"Not enough information to connect to redis:\n  {self}\n" +
                "*" * 80 + "\n")
        else:
            try:
                self.conn = redis.Redis(host=self.r_endpoint, port=self.r_port, password=self.r_pass)
                self.usable = True
                _logger.debug(f"Created connection -- {self}")

                # test connection
                self.conn.get("SAMPLE")
            except Exception as err:
                _logger.warning(
                    "\n" + "*" * 80 + "\n" + f"Unable to make connection to redis:\n  {self}\n" +
                    f"{err}\n" + "*" * 80 + "\n")
                self.usable = False

    def __str__(self) -> str:
        pwmask = "None" if self.r_pass is None else "*****"
        u_flag = "" if self.usable else "     [UNUSABLE]"
        return f"RedisDB ({self.r_endpoint}:{self.r_port}) -> {pwmask}{u_flag}"

    def __del__(self):
        if self.conn is not None:
            self.conn.close()

    def _normalize_env(self):
        """
        For any parameters that are still None, try to fill in the blanks with known environment variables.
        """
        if self.r_endpoint is None:
            if "REDIS_CLUSTER_ENDPOINT" in os.environ:
                self.r_endpoint = os.environ["REDIS_CLUSTER_ENDPOINT"]
        # if self.r_port is None:
        #     if "PGPORT" in os.environ:
        #         self.r_port = os.environ["PGPORT"]
        if self.r_pass is None:
            if "REDIS_AUTH_TOKEN" in os.environ:
                self.r_pass = os.environ["REDIS_AUTH_TOKEN"]

    def _normalize_envfile(self):
        """
        For any parameters that are still None, try to fill in the blanks with known envfile settings.

        NOTE: envfile assumed to be located at the top of the repo (hinting/envfile)
        """
        if self.r_endpoint is not None and self.r_port is not None and self.r_pass is not None:
            return  # nothing to do, everything set
        ftc = FunctionalTestConfig()
        the_file = os.path.join(ftc.REPO_ROOT, "envfile")
        if ftc.ENVFILE is not None:
            the_file = ftc.ENVFILE
        if not os.path.exists(the_file):
            _logger.debug(f"Unable to normalize_envfile: does not exist at '{the_file}'")
            return

        config_object = ConfigParser()
        with open(the_file) as stream:
            config_object.read_string("[DEFAULT]\n" + stream.read())

        if self.r_endpoint is None:
            if config_object.has_option("DEFAULT", "REDIS_CLUSTER_ENDPOINT"):
                self.r_endpoint = config_object.get("DEFAULT", "REDIS_CLUSTER_ENDPOINT")
        # if self.r_port is None:
        #     if config_object.has_option("DEFAULT", "pgport"):
        #         self.r_port = int(config_object.get("DEFAULT", "pgport"))
        if self.r_pass is None:
            if config_object.has_option("DEFAULT", "REDIS_AUTH_TOKEN"):
                self.r_pass = config_object.get("DEFAULT", "REDIS_AUTH_TOKEN")

    # ----- Data access methods --------------------------------------------------

    def get_device(self, dev_id: int, org_key: str) -> Dict:
        """
        Get information from redis for a device.

        :param dev_id: device id
        :param org_key: org_key
        :return: device info as dict
        """
        if not isinstance(dev_id, int):
            raise TypeError("`dev_id` must be an int!")
        if not isinstance(org_key, str):
            raise TypeError("`org_key` must be a str!")

        key = "devices" + "-{" + f"{org_key}" + "}" + f"-{dev_id}"
        _logger.debug(f"Redis key: {key}")

        resp = self.conn.get(key)
        answer = json.loads(resp)
        return answer

    def get_group(self, grp_id: int, org_key: str) -> Dict:
        """
        Get information from redis for a device group.

        :param grp_id:  device group id
        :param org_key:  org_key
        :return: device group info as dict
        """
        if not isinstance(grp_id, int):
            raise TypeError("`grp_id` must be an int!")
        if not isinstance(org_key, str):
            raise TypeError("`org_key` must be a str!")

        key = "device_groups" + "-{" + f"{org_key}" + "}" + f"-{grp_id}"
        _logger.debug(f"Redis key: {key}")

        resp = self.conn.get(key)
        answer = json.loads(resp)
        return answer

    def get_org(self, org_key: str) -> Dict:
        """
        Get information from redis for an organization.

        :param org_key: org_key
        :return: organization info as a dict
        """
        if not isinstance(org_key, str):
            raise TypeError("`org_key` must be a str!")

        key = "organizations" + "-{" + f"{org_key}" + "}"
        _logger.debug(f"Redis key: {key}")

        resp = self.conn.get(key)
        answer = json.loads(resp)
        return answer

    def get_env(self, name: str) -> Dict:
        """
        Get information from redis abount an environment.

        :param name: environment name
        :return: environment info as dict
        """
        if not isinstance(name, str):
            raise TypeError("`name` must be a str!")

        key = f"environments-{name}"
        _logger.debug(f"Redis key: {key}")

        resp = self.conn.get(key)
        answer = json.loads(resp)
        return answer
