#  coding: utf-8
#  Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved
################################################################################
import json
import logging
import re
import time
from datetime import datetime, timezone, timedelta
from typing import Type, Dict, Optional, Union

import _pytest.outcomes
import pytest
from psc_orgkey import encode_orgkey

from lib.connect.hint_connect import HintControlConnect
from lib.models.device import Device
from lib.models.device_group import DeviceGroup
from lib.models.environment import Env
from lib.models.organization import Org

_logger = logging.getLogger(__name__)

__all__ = ["TestCommon"]


################################################################################
# Test class base object
################################################################################

class TestCommon(object):
    """
    Base class for all tests, containing common use methods.
    """

    INVOKED = time.strftime("%Y%m%d-%H%M%S", time.localtime(time.time()))
    PYTEST_FAIL = _pytest.outcomes.Failed

    # noinspection PyUnusedFunction
    @classmethod
    def setup_class(cls):
        _logger.info("\n" + "=" * 80 + "\n" + f"  Executing {cls.__name__} ({cls.__module__})\n" + "=" * 80)

        cls.LQ_VERSION = 17  # Must match BASEMESSAGE.LIVE_QUERY_VERSION in "internal/hints/basemessage.go"
        cls.ORGKEY = encode_orgkey("local", cls.uid())  # usable, valid OrgKey -- don't use in deletion tests!
        cls.TOMORROW = TestCommon.stamp(1440)  # 24 hours from now, assured to be LATER than any hinting config
        cls.YESTERDAY = TestCommon.stamp(-1440)  # 24 hours ago, assured to be EARLIER than any hinting config

    # noinspection PyUnusedFunction
    @classmethod
    def teardown_class(cls):
        pass

    class Timer(object):
        """
        Context block object that can be used to time operations.
        """

        # noinspection PyUnusedLocal
        def __init__(self, tag: str = None):
            self._start = time.time()
            self._tag = "" if tag is None else tag + " "

        # noinspection PyUnusedLocal
        def __enter__(self):
            self._start = time.time()
            return self

        # noinspection PyUnusedLocal
        def __exit__(self, exc_type: Type[Exception], exc_val: Exception, exc_tb):
            _logger.info(f"{self._tag}TIME: {time.time() - self._start:.2f}s")

    # ----- Test assistance methods -----------------------------------------------------------------

    def new_orgkey(self, env: str = "local", the_id: int = None) -> str:
        """Generate a new orgkey."""
        if env is not None and not isinstance(env, str):
            raise TypeError("If specified, `env` must be a str!")
        if the_id is not None and not isinstance(the_id, int):
            raise TypeError("If specified, `the_id` must be an str!")

        if the_id is None:
            the_id = self.uid()

        return encode_orgkey(env, the_id)

    class HintPrep(object):
        def __init__(self):
            self.start_time: datetime = Optional[None]
            self.name: str = ""
            self.env: Optional[Env] = None
            self.orgkey: str = ""
            self.org: Optional[Org] = None
            self.group_id: int = 0
            self.group: Optional[DeviceGroup] = None
            self.dev_id: int = 0
            self.dev: Optional[Device] = None
            self.done_time: datetime = Optional[None]

    def prepare_for_hinting(self, hc_connect: HintControlConnect, use_dev: Device = None, use_group: DeviceGroup = None,
                            use_org: Org = None, use_env: Env = None, no_env: bool = False) -> HintPrep:
        """
        Prepare for a new hinting test by updating a set of configuration objects.  Returns a HintPrep() object with the following fields:
          - name: the environment name used
          - env: the environment object data
          - orgkey: the orginization key used
          - org: the organization object data
          - group_id: the group id used
          - group: the device group object data
          - dev_id: the device id used
          - dev: the device object data
          - now: the time this hinting config was generated

        Any Device, DeviceGroup or Org objects not otherwise supplied will be created with .full(no_live_query=True),
        so any live query timestamps will use current time.

        :param hc_connect: connection to service
        :param use_dev: (optional) specific device info to be used
        :param use_group: (optional) specific device group info to be used
        :param use_org: (optional) specific organization info to be used
        :param use_env: (optional) specific environment info to be used
        :param no_env: (optional, used for negative testing) If True don't actually create the Env
        :return: HintPrep object with creation information
        """
        if use_dev is not None and not isinstance(use_dev, Device):
            raise TypeError("If specified, `use_dev` must be a Device() object!")
        if use_group is not None and not isinstance(use_group, DeviceGroup):
            raise TypeError("If specified, `use_group` must be a DeviceGroup() object!")
        if use_org is not None and not isinstance(use_org, Org):
            raise TypeError("If specified, `use_org` must be an Org() object!")
        if use_env is not None and not isinstance(use_env, Env):
            raise TypeError("If specified, `use_env` must be an Env() object!")

        info = self.HintPrep()
        info.start_time = self.stamp()  # time stamp after all config objects (and their timestamps) are generated
        time.sleep(0.01)

        info.name = f"qa{self.uid()}"
        if no_env:
            info.env = None
        else:
            info.env = Env.full() if use_env is None else use_env
            hc_connect.v1_env_update(info.env, env_name=info.name)

        info.orgkey = encode_orgkey(info.name, self.uid())
        info.org = Org.full(no_live_query=True) if use_org is None else use_org
        hc_connect.v1_org_update(info.org, org_key=info.orgkey)

        info.group_id = self.uid()
        info.group = DeviceGroup.full(no_live_query=True) if use_group is None else use_group
        hc_connect.v1_devicegroup_update(info.group, org_key=info.orgkey, group_id=info.group_id)

        info.dev_id = self.uid()
        info.dev = Device.full(no_live_query=True) if use_dev is None else use_dev
        hc_connect.v1_device_update(info.dev, org_key=info.orgkey, device_id=info.dev_id)

        time.sleep(0.01)
        info.done_time = self.stamp()  # time stamp after all config objects (and their timestamps) are generated

        return info

    # ----- Other common use methods -----------------------------------------------------------------

    # noinspection PyUnusedFunction
    @staticmethod
    def from_json(data: str) -> dict:
        """Convert JSON information from a source into a more accessible dict object.

        :param data: string containing JSON text
        :return: dict object
        :exception: JSONDecodeError
        :exception: TypeError
        """
        return json.loads(data)

    # noinspection PyUnusedFunction
    @staticmethod
    def to_json(data: dict) -> str:
        """Convert dict into a JSON format for later use.

        :param data: dict to be converted
        :return: string of JSON text
        :exception: TypeError
        """
        return json.dumps(data, sort_keys=True)

    @staticmethod
    def unique(prefix: str = "") -> str:
        """Generate a unique string based on current time.

        :param prefix: (optional) prefix
        :return: Unique time.
        """
        if not isinstance(prefix, str):
            raise TypeError("If specified, `prefix` must be a string!")

        letters = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

        now = int(time.time() * 10000)  # to the 1/10th millisecond
        time.sleep(0.01)  # added to ensure unique time if called repeatedly
        tag = ""
        for _ in range(3):
            tag = letters[int(now % 26)] + tag
            now = now / 26
        return f"{prefix}{tag}{int(now)}"

    @staticmethod
    def uid() -> int:
        """Create unique id based on current time, to the millisecond."""
        time.sleep(0.01)  # assure unique
        return int(time.time() * 1000.0)

    @staticmethod
    def stamp(delta: int = 0) -> datetime:
        """generate a quick timestamp based on current time, to the nearest millisecond

        :param delta: time variance in minutes (default 0)
        """
        use_datetime = datetime.now(timezone.utc) + timedelta(minutes=delta)
        try:
            ms = round(max(0, min(use_datetime.microsecond, 999999)), -3)
            return use_datetime.replace(microsecond=ms)
        except Exception as err:
            _logger.error(
                f"micro={use_datetime.microsecond} -> {err}")  # Something odd pops up spurriously, log for now to try and figure what
        return use_datetime

    def is_now(self, check: Union[datetime, str], delta: int = 5) -> bool:
        """
        Verify that supplied time is "now" within +/- some wiggle room in seconds.  Designed to take
        a datetime, or a string in iso format (2023-03-28T13:14:26.460021-04:00) or UTC format
        (2023-03-28T15:52:49.619770769Z).

        NOTE: in the case of UTC format, we need to remove nanoseconds, if seen

        :param check: time to be checked (datetime or postgres string format)
        :param delta: wiggle room in seconds (default 5)
        :return: True if close enough
        """
        now = self.stamp()
        use_check = check
        if isinstance(check, str):
            try:
                use_check = datetime.fromisoformat(check)
            except ValueError:
                # NOTE: This is a hack until we can get the new container built and we can use dateutil.parse
                p = re.compile(r"^(\S*)\.(\d+)Z$")
                m = p.match(check)
                if m and len(m.group(2)) > 6:
                    check = m.group(1) + "." + m.group(2)[0:6] + "Z"
                    _logger.info(f"Trimmed nanoseconds in {check} (.{m.group(2)}Z)")
                use_check = datetime.strptime(check, "%Y-%m-%dT%H:%M:%S.%fZ")
        diff = now.timestamp() - use_check.timestamp()
        return abs(diff) <= delta

    @staticmethod
    def epoch() -> datetime:
        """generate an timestamp for the epoch date (0)."""
        return datetime.fromtimestamp(0, tz=timezone.utc)

    def pinned(self, minutes: int = 0):
        """Generate a fixed timestamp of known value (used primarily for internal testing).

        :param minutes: minutes index (default 0)
        """
        return datetime.replace(self.stamp(), year=1999, month=9, day=13, hour=0, minute=0, second=0,
                                microsecond=0) + timedelta(minutes=minutes)

    @staticmethod
    def verify_response(connect: HintControlConnect, return_code: int, message: str = None) -> None:
        """
        Quick validate of last response.  If there has been no activity, nothing is checked.

        :param connect: last connection used
        :param return_code: expected http response code
        :param message: expected message, if code is correct (optional)
        """
        if not isinstance(connect, HintControlConnect):
            raise TypeError("`connect` must be a HintControlConnect() object!")

        # Check return code
        if connect.last_status is None:  # no operation yet
            return
        if connect.last_status != return_code:
            pytest.fail(f"Saw return_code {connect.last_status} instead of expected {return_code}")

        # Return code ok, check message (if specified)
        if message is not None:
            if connect.last_content is None:
                pytest.fail("We expected response content, but none was provided!")
            if isinstance(connect.last_content, Dict):
                if "message" not in connect.last_content:
                    pytest.fail("Response content was dict, but did not have expected 'message' entry!")
                if message not in connect.last_content['message']:
                    pytest.fail(
                        f"Response content message did not match `{message}`\n  {connect.last_content['message']}")
            else:
                pytest.fail(f"Response content was not in dict format:\n  {connect.last_content}")
