#  coding: utf-8
#  Copyright (c) 2023 VMware Carbon Black, all rights reserved
################################################################################

import logging
import os
import re
import shutil
import subprocess
from typing import List, Tuple, Optional

from _pytest.config import Config
from filelock import FileLock

from lib.connect.hint_connect import HintControlConnect
from lib.postgres import PostgresDB
from lib.redis import RedisDB
from .configuration import FunctionalTestConfig

_logger = logging.getLogger(__name__)

__all__ = ["Artifacts", "Setup"]


class Artifacts(object):
    """
    Class dealing with pre-generated sensor and process data, as well as with runtime artifacts.
    """
    # location of the artifacts
    LOCATION = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "artifacts"))

    # lock file used whan altering the artifacts folder
    LOCK = "/tmp/artifacts_initialization.lock"

    # special artifact file names
    RUN_FAILURES = "test_run_failures.txt"
    TEST_TIMING = "test_timing.data"

    # --------------------------------------------------------------------------------

    def __init__(self):
        self._ensure_structure()

    # noinspection OsChmod
    def _ensure_structure(self):
        """
        Make sure the artifacts structure exists.
        """
        with FileLock(self.LOCK):
            if not os.path.exists(self.LOCATION):
                os.makedirs(self.LOCATION)
                os.chmod(self.LOCATION, 0o777)

    def _normalize_permissions(self, path: str) -> None:
        """
        Ensure that any created artifact file or directory has permissions that allow for easy user removal.

        :param path: path used for artifact
        """
        tree = path.split("/")
        if len(tree) == 1:
            return  # simple name in artifacts
        path = self.LOCATION
        for item in tree:
            path = os.path.join(path, item)
            if os.path.isdir(path):
                os.chmod(path, 0o777)  # all directories enterable by anyone
            else:
                os.chmod(path, 0o666)  # all files read/write by anyone

    # --------------------------------------------------------------------------------

    def create(self, target: str = None, text: str = None, replace: bool = False, separate: bool = True) -> None:
        """
        Append text to a specified file within the artifacts folder. If the file already exists
        a seperator line will be added before the new text is appended.

        :param target: filename or relative path within the artifacts folder
        :param text: text to be added; if None and file does not exist, the file will be initialized with no content
        :param replace: if True, replace existing text instead of appending
        :param separate: If True, add seperator between entries when appending
        """
        self._ensure_structure()

        if target is None:
            raise TypeError("`target` must be specified")

        the_path = os.path.join(self.LOCATION, target)
        the_root = os.path.dirname(the_path)
        if not os.path.exists(the_root):
            os.makedirs(the_root)

        with FileLock(f"/tmp/save_text_{target}.lock"):
            try:
                if not os.path.exists(the_path) and text is None:
                    with open(the_path, "w", encoding="utf-8") as fp:
                        fp.write("")
                else:
                    pre = ""
                    mode = "w"
                    if not replace:
                        if os.path.exists(the_path) and separate:
                            pre = "\n" + "-" * 80 + "\n\n"
                        mode = "a"
                    with open(the_path, mode, encoding="utf-8") as fp:
                        fp.write(f"{pre}{text}\n")

                self._normalize_permissions(target)
            except Exception as err:
                _logger.error(f"Failed to save text to {the_path}: {err}")

    def remove(self, filename: str) -> None:
        """
        Remove specific artifact file.

        :param filename: name or relative path of the artifact file to remove
        """
        path = os.path.join(self.LOCATION, filename)
        if os.path.exists(path):
            os.remove(path)

    def remove_all(self) -> None:
        """
        Remove old runtime artifacts folder and regenerate it.
        """
        with FileLock(self.LOCK):
            if os.path.exists(self.LOCATION):
                shutil.rmtree(self.LOCATION)
        self._ensure_structure()

    def retrieve(self, filename: str) -> List[str]:
        """
        Retieve an artifact file and return it as a list of strings.

        :param filename: filename for text to be retrieved
        :return: list of strings (lines of file content, empty list if no such file)
        """
        self._ensure_structure()

        if not filename:
            raise ValueError("`filename` must be specified!")

        messages = os.path.join(self.LOCATION, filename)
        if not os.path.exists(messages):
            return []

        # Otherwise...
        with open(messages, "r", encoding="utf-8") as fp:
            data = fp.readlines()
        return data


class Setup(object):
    """
    Class containing various test run setup utilities.
    """

    @staticmethod
    def environment_banner() -> str:
        """
        Display the current test environment settings.
        """
        ft_config = FunctionalTestConfig()

        def _right(text: str) -> str:
            return " " * max(int(78 - len(text)), 2) + text  # trap long text

        banner = "\n\n" + "%" * 80 + "\n"
        banner += f"  Hinting microservice: {ft_config.SERVICE_IP}:{ft_config.SERVICE_PORT}\n"
        if ft_config.POSTGRES is not None:
            banner += f"  Postgres service:     {ft_config.POSTGRES}\n"
        if ft_config.REDIS is not None:
            banner += f"  Redis service:        {ft_config.REDIS}\n"

        banner += "\n"
        pwd = os.path.abspath(".")
        banner += _right(f"Executing in `{pwd}`") + "\n"
        banner += "%" * 80 + "\n"

        for item, _ in ft_config.OPTIONS_USED.items():
            banner += f" -- {item}\n"
        return banner

    @staticmethod
    def finalize_working_environment(pt_config: Config, ft_config: FunctionalTestConfig) -> None:
        """
        Make any final environment changes before the tests start,

        :param pt_config: pytest configuration object
        :param ft_config: FunctionalTestConfig object
        """
        # FUTURE: stubbed for future needs
        pass

    @staticmethod
    def normalize_logging(with_logger=None, debug: bool = False) -> None:
        """
        Turn on debug mode for all configured loggers.

        :param with_logger: (optional) source logger from the utility
        :param debug: if True, use DEBUG else INFO
        """
        level = logging.DEBUG if debug else logging.INFO

        fmt = '%(asctime)s - [%(process)d] %(levelname)7s [%(name)s:%(lineno)d] %(message)s'
        logging.basicConfig(level=level, format=fmt)

        # Add last_execution.log to artifacts
        root_logger = logging.getLogger()
        fh = logging.FileHandler(os.path.join(Artifacts().LOCATION, 'last_execution.log'))
        fh.setLevel(logging.DEBUG)
        fmt = logging.Formatter("%(asctime)s - [%(process)d] %(levelname)7s [%(name)s:%(lineno)d] %(message)s")
        fh.setFormatter(fmt)
        root_logger.addHandler(fh)

        loggers = [root_logger]  # get the root logger
        if with_logger is not None:  # add optional utility logger
            loggers.append(with_logger)
        for logger in loggers:
            logger.setLevel(level)

        # make sure other stuff is quiet
        logging.getLogger("filelock").setLevel(logging.ERROR)

        # NOTE: comment out the first line and uncomment the second if you need to get request info
        logging.getLogger("requests").setLevel(logging.ERROR)
        # logging.getLogger("requests").setLevel(logging.DEBUG)

    @staticmethod
    def process_automation_options(pt_config: Config, ft_config: FunctionalTestConfig) -> None:  # pragma: no cover
        """
        Update the FunctionalTestConfig() object with any command line options.

        :param pt_config: pytest configuration object
        :param ft_config: FunctionalTestConfig object
        """
        if pt_config is None:
            raise TypeError("`pt_config` must be supplied (command line request object)")
        if not isinstance(ft_config, FunctionalTestConfig):
            raise TypeError("`ut_config` must be a UnitTestConfig configuration object!")

        # test root is defined by pytest.ini location (i.e. hinting/test/functional/suite)
        ft_config.TEST_ROOT = os.path.dirname(pt_config.inipath.absolute())

        # repository root is assumed to be three levels up
        _logger.debug(f"inipath: {pt_config.inipath.absolute()}")
        _logger.debug(f"TEST_ROOT: {ft_config.TEST_ROOT}")
        ft_config.REPO_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(ft_config.TEST_ROOT)))
        _logger.debug(f"REPO_ROOT: {ft_config.REPO_ROOT}")

        ft_config.DEBUG = pt_config.option.dbg

        ft_config.USING_XDIST = False
        if pt_config.option.numprocesses is not None and pt_config.option.numprocesses > 1:
            ft_config.USING_XDIST = True

        look_for = ['-n', '--numprocesses']
        ft_config.USING_XDIST = True in [str(x).lower() in look_for for x in pt_config.invocation_params.args]

        # This section is used to determine the IP and PORT of the service to be tested.
        if pt_config.option.ip is not None:  # external
            # special case "docker" will search local kubernetes for ip
            ft_config.SERVER_TYPE = 'external'
            ft_config.SERVICE_IP = pt_config.option.ip
        else:
            ft_config.SERVER_TYPE = 'local'
            ft_config.SERVICE_IP = "localhost"

        if pt_config.option.port is not None:
            ft_config.SERVICE_PORT = int(pt_config.option.port)
        else:
            ft_config.SERVICE_PORT = 5004

        # specific envfile for data scraping
        if pt_config.option.envfile is not None:
            ft_config.ENVFILE = pt_config.option.envfile

        # create a connection
        ft_config.CONNECTION = HintControlConnect(ft_config.SERVICE_IP, ft_config.SERVICE_PORT)

        # handle any postgres overrides
        if pt_config.option.pghost is not None:
            ft_config.PGHOST = pt_config.option.pghost
        if pt_config.option.pgport is not None:
            ft_config.PGPORT = int(pt_config.option.pgport)
        if pt_config.option.pguser is not None:
            ft_config.PGUSER = pt_config.option.pguser
        if pt_config.option.pgpass is not None:
            ft_config.PGPASS = pt_config.option.pgpass
        if pt_config.option.dbname is not None:
            ft_config.DBNAME = pt_config.option.dbname

        # create postgres
        ft_config.POSTGRES = PostgresDB(ft_config.PGHOST, ft_config.PGPORT, ft_config.PGUSER, ft_config.PGPASS,
                                        ft_config.DBNAME)

        # handle any redis overrides
        if pt_config.option.rhost is not None:
            ft_config.RHOST = pt_config.option.rhost
        if pt_config.option.rpass is not None:
            ft_config.RPASS = pt_config.option.rpass

        # create redis
        ft_config.REDIS = RedisDB(ft_config.RHOST,  ft_config.RPASS)

    @staticmethod
    def _shell(command: str) -> Tuple[List[str], List[str]]:
        com = command.split(r"\s+")
        raw = subprocess.Popen(com, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        output, errors = raw.communicate()
        raw.wait()
        return output.split("\n"), errors.split("\n")

    @staticmethod
    def _find(lines: List[str], key: str, item: int) -> Optional[str]:
        for line in lines:
            if key in line:
                parts = re.split(r"\s+", line)
                return parts[item]
        return None
