#!/usr/bin/env python
#  coding: utf-8
#  Copyright (c) 2023 VMware Carbon Black, all rights reserved
################################################################################

import argparse
import logging
import random

import sys
from psc_orgkey import encode_orgkey, decode_orgkey, InvalidOrgKeyError

logger = logging.getLogger(__name__)


################################################################################
# Begin main
################################################################################


# --------------------------------------------------------------------------------
def build_cli_parser() -> argparse.Namespace:
    """
    Build the set of accepted options.
    :return: parsed arguments
    """
    parser = argparse.ArgumentParser(description="Create or decode an orgkey.  If no arguments are provided,"
                                     " a random orgkey of environment 'enki' is produced.")

    # for each supported output type, add an option
    parser.add_argument("--debug", action="store_true", dest="debug", default=False,
                        help="enable debug logging\n")
    parser.add_argument("-d", "--decode", action="store", dest="decode", default=None,
                        help=f"decode a supplied string orgkey into parts")
    parser.add_argument("-e", "--encode", action="store", dest="encode", default=None,
                        help=f"encode a supplied env/orgid into an orgkey")

    args = parser.parse_args()
    return args


def main():
    """
    Main routine.
    """
    try:
        args = build_cli_parser()
    except Exception as err:
        logger.error(f"Setup Failure: {err}")
        sys.exit(1)

    level = logging.DEBUG if args.debug else logging.INFO

    fmt = '%(asctime)s - [%(process)d] %(levelname)7s [%(name)s:%(lineno)d] %(message)s'
    logging.basicConfig(level=level, format=fmt)

    if args.decode is None and args.encode is None:
        orgid = random.randint(1, 10000)
        orgkey = encode_orgkey("enki", orgid)
        print(f"\nGenerated OrgKey for ENV 'enki', OrgId '{orgid}' ==> {orgkey}\n")
    elif args.decode is not None:
        orgkey = args.decode.strip().upper()
        try:
            eo = decode_orgkey(orgkey)
            print(f"\nOrgKey '{orgkey}' decoded ==> ENV '{eo.env}', OrgId '{eo.org_id}'\n")
        except InvalidOrgKeyError:
            print(f"\nUnable to decode '{orgkey}'")
    else:
        parts = args.encode.strip().split("/")
        env = parts[0]
        orgid = int(parts[1])
        try:
            orgkey = encode_orgkey(env, orgid).upper()
            print(f"\nGenerated OrgKey for ENV '{env}', OrgId '{orgid}' ==> {orgkey}\n")
        except Exception as err:
            print(f"\nUnable to encode '{env}/{orgid}': {err}")


if __name__ == "__main__":
    main()
