#!/usr/bin/env python
#  coding: utf-8
#  Copyright (c) 2023 VMware Carbon Black, all rights reserved
################################################################################

import argparse
import base64
import logging
import os
import subprocess
import sys

import yaml

logger = logging.getLogger(__name__)


################################################################################
# Begin main
################################################################################


def update_envfile(path: str, show: bool):
    """
    Scrape the infrastructure and grab the passwords, which we then use to update our envfile.

    :param path: path to the env file
    :param show: if true, only display to console
    """
    proc = subprocess.Popen('kubectl get secrets --namespace=local-apps -o yaml',
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (out, err) = proc.communicate()
    data = yaml.safe_load(out.decode())

    if len(err) > 0:
        print(f"Unable to get secrets from kubectl: \n   {err.decode()}")
        sys.exit(1)

    logger.debug("envfile:\n" + out.decode() + "-" * 82 + "\n")
    postgres_pass = None
    redis_pass = None
    for item in data['items']:
        if "password" not in item['data']:
            continue
        source = item['data']['password']
        if "postgres" in item['metadata']['name']:
            postgres_pass = source
        elif "redis-ec" in item['metadata']['name']:
            redis_pass = source
    if postgres_pass is not None:
        postgres_pass = base64.b64decode(postgres_pass).decode()
    if redis_pass is not None:
        redis_pass = base64.b64decode(redis_pass).decode()

    if show:
        print("=" * 82)
        if postgres_pass is not None:
            print(f"PGPASSWORD:       {postgres_pass}")
        else:
            print("...No PGPASSWORD seen in extracted secrets")
        if redis_pass is not None:
            print(f"REDIS_AUTH_TOKEN: {redis_pass}")
        else:
            print("...No REDIS_AUTH_TOKEN seen in extracted secrets")
        print("=" * 82)
    else:
        # Read and update actual envfile
        print(f"Reading from '{path}'...")
        with open(path, 'r') as fp:
            lines = fp.readlines()

        # substitute the right lines
        newlines = []
        for item in lines:
            if item.startswith("PGPASSWORD") and postgres_pass is not None:
                new_data = f"PGPASSWORD={postgres_pass}\n"
                newlines.append(new_data)
                if new_data != item:
                    print("... updating PGPASSWORD")
            elif item.startswith("REDIS_AUTH_TOKEN") and redis_pass is not None:
                new_data = f"REDIS_AUTH_TOKEN={redis_pass}\n"
                newlines.append(new_data)
                if new_data != item:
                    print("... updating REDIS_AUTH_TOKEN")
            else:  # no change
                newlines.append(item)

        # write out new envfile
        with open(path, 'w') as fp:
            fp.writelines(newlines)
        print(f"Updated '{path}'")


# --------------------------------------------------------------------------------
def build_cli_parser() -> argparse.Namespace:
    """
    Build the set of accepted options.
    :return: parsed arguments
    """
    parser = argparse.ArgumentParser(description="Prepare a local docker or remote IP for automation.")

    # for each supported output type, add an option
    parser.add_argument("--debug", action="store_true", dest="debug", default=False,
                        help="enable debug logging\n")
    path_default = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "envfile")
    parser.add_argument("--path", action="store", dest="path", default=path_default,
                        help=f"alternate filepath (otherwise assumed to be '{path_default}')")
    parser.add_argument("--show", action="store_true", dest="show", default=False,
                        help="show passwords on console instead of updating envfile")

    args = parser.parse_args()
    return args


def main():
    """
    Main routine.
    """
    try:
        args = build_cli_parser()
    except Exception as err:
        logger.error(f"Setup Failure: {err}")
        sys.exit(1)

    level = logging.DEBUG if args.debug else logging.INFO

    fmt = '%(asctime)s - [%(process)d] %(levelname)7s [%(name)s:%(lineno)d] %(message)s'
    logging.basicConfig(level=level, format=fmt)

    if not args.show and not os.path.exists(args.path):
        print(f"Environment file '{args.path}' does not exist.")
        sys.exit(1)

    update_envfile(args.path, args.show)


if __name__ == "__main__":
    main()
