// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicechangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicegroupchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

const (
	pokeTestEnv       = "POKETEST"
	pokeTestGroupId   = 777
	pokeTestDeviceId  = 888
	pokeTestOrgId     = 111
	userAgentHeader   = "User-Agent"
	userAgentPokeTest = "poketest"
)

var epoch = time.Unix(0, 0) // a REAL old sensor check in, assured to generate hints

// Basic list of routes used
var (
	deviceRouteFormatString      = "%s/hinting/v1/orgs/%s/devices/%d"
	deviceGroupRouteFormatString = "%s/hinting/v1/orgs/%s/devicegroups/%d"
	orgRouteFormatString         = "%s/hinting/v1/orgs/%s"
	envRouteFormatString         = "%s/hinting/v1/envs/%s"
	hintRouteFormatString        = "%s/hinting/v1/hints"
)

var (
	orgKey, _    = orgkeylib.Encode(pokeTestEnv, pokeTestOrgId)
	orgKeyString = orgKey.String()
)

func init() {
}

type HeaderMap map[string]string

type HttpClient struct {
	client     *http.Client
	targetHost string
	headers    HeaderMap
}

// SampleClient -- Used to store client information for a test
type SampleClient struct {
	*testing.T
	orgKey        string
	envName       string
	deviceId      int64
	deviceGroupId int64
	api           *HttpClient
}

// NewHTTPClient -- Generate a new sample client for use in a test
func NewHTTPClient(targetHost string, headers HeaderMap) *HttpClient {
	client := &http.Client{CheckRedirect: func(req *http.Request, via []*http.Request) error {
		req.Header.Set(userAgentHeader, userAgentPokeTest)
		return nil
	}}
	return &HttpClient{client: client, targetHost: targetHost, headers: headers}
}

func NewTestClient(t *testing.T, testClient *HttpClient) *SampleClient {
	return &SampleClient{
		api: testClient, T: t, orgKey: orgKeyString, deviceId: pokeTestDeviceId, deviceGroupId: pokeTestGroupId, envName: pokeTestEnv,
	}
}

func (hc HttpClient) Do(request *http.Request) (*http.Response, error) {
	for key, value := range hc.headers {
		request.Header.Set(key, value)
	}
	request.Header.Set(userAgentHeader, userAgentPokeTest)
	return hc.client.Do(request)
}

// ----- Device methods ----------------------------------------

func (hc HttpClient) putDevice(device *models.APIDevice) (*http.Response, error) {
	marshalledDevice, err := json.Marshal(device)
	if err != nil {
		return nil, err
	}
	req, err := http.NewRequest(http.MethodPut, fmt.Sprintf(deviceRouteFormatString, hc.targetHost, orgKeyString, device.Id), bytes.NewReader(marshalledDevice))
	if err != nil {
		return nil, err
	}
	return hc.Do(req)
}

func (c SampleClient) TestPutDevice() {
	device := models.APIDevice{
		APIDeviceURI: models.APIDeviceURI{Id: c.deviceId, OrgKey: c.orgKey},
		APIDeviceBody: models.APIDeviceBody{
			Changes: []string{devicechangetype.DEV_GROUP_CHANGE.String(), devicechangetype.DEV_REP_REQUEST.String()},
		},
	}
	response, err := c.api.putDevice(&device)
	if err != nil {
		c.Fatalf("Error interacting with devices route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from devices call %d %s", response.StatusCode, readBody)
	}
}

func (c SampleClient) TestPutDeviceLQ() {
	now := time.Now().UnixMilli()
	device := models.APIDevice{
		APIDeviceURI: models.APIDeviceURI{Id: c.deviceId, OrgKey: c.orgKey},
		APIDeviceBody: models.APIDeviceBody{
			Changes:       []string{devicechangetype.DEV_LIVE_QUERY_CHANGE.String(), devicechangetype.DEV_REP_REQUEST.String()},
			LiveQueryTime: &now,
		},
	}
	response, err := c.api.putDevice(&device)
	if err != nil {
		c.Fatalf("Error interacting with devices route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from devices call %d %s", response.StatusCode, readBody)
	}
}

func (c SampleClient) TestPutDeviceLQNil() {
	device := models.APIDevice{
		APIDeviceURI: models.APIDeviceURI{Id: c.deviceId, OrgKey: c.orgKey},
		APIDeviceBody: models.APIDeviceBody{
			Changes: []string{devicechangetype.DEV_LIVE_QUERY_CHANGE.String()},
		},
	}
	response, err := c.api.putDevice(&device)
	if err != nil {
		c.Fatalf("Error interacting with devices route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from devices call %d %s", response.StatusCode, readBody)
	}
}

// ----- Group methods ----------------------------------------

func (hc HttpClient) putGroup(group *models.APIDeviceGroup) (*http.Response, error) {
	marshalledGroup, err := json.Marshal(group)
	if err != nil {
		return nil, err
	}
	req, err := http.NewRequest(http.MethodPut, fmt.Sprintf(deviceGroupRouteFormatString, hc.targetHost, orgKeyString, group.Id), bytes.NewReader(marshalledGroup))
	if err != nil {
		return nil, err
	}
	return hc.Do(req)
}

func (c SampleClient) TestPutGroup() {
	group := models.APIDeviceGroup{
		APIDeviceGroupURI:  models.APIDeviceGroupURI{Id: c.deviceGroupId, OrgKey: c.orgKey},
		APIDeviceGroupBody: models.APIDeviceGroupBody{Changes: []string{devicegroupchangetype.DG_POLICY_CHANGE.String()}},
	}
	response, err := c.api.putGroup(&group)
	if err != nil {
		c.Fatalf("Error interacting with device groups route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		c.Fatalf("Bad status returned from device groups call %d", response.StatusCode)
	}
}

func (c SampleClient) TestPutGroupLQ() {
	now := time.Now().UnixMilli()
	group := models.APIDeviceGroup{
		APIDeviceGroupURI:  models.APIDeviceGroupURI{Id: c.deviceGroupId, OrgKey: c.orgKey},
		APIDeviceGroupBody: models.APIDeviceGroupBody{Changes: []string{devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()}, LiveQueryTime: &now},
	}
	response, err := c.api.putGroup(&group)
	if err != nil {
		c.Fatalf("Error interacting with device groups route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		c.Fatalf("Bad status returned from device groups call %d", response.StatusCode)
	}
}

func (c SampleClient) TestPutGroupLQNil() {
	group := models.APIDeviceGroup{
		APIDeviceGroupURI:  models.APIDeviceGroupURI{Id: c.deviceGroupId, OrgKey: c.orgKey},
		APIDeviceGroupBody: models.APIDeviceGroupBody{Changes: []string{devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String()}},
	}
	response, err := c.api.putGroup(&group)
	if err != nil {
		c.Fatalf("Error interacting with device groups route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		c.Fatalf("Bad status returned from device groups call %d", response.StatusCode)
	}
}

// ----- Org methods ----------------------------------------

func (hc HttpClient) putOrg(org *models.APIOrg) (*http.Response, error) {
	marshalledOrg, err := json.Marshal(org)
	if err != nil {
		return nil, err
	}
	req, err := http.NewRequest(http.MethodPut, fmt.Sprintf(orgRouteFormatString, hc.targetHost, orgKeyString), bytes.NewReader(marshalledOrg))
	if err != nil {
		return nil, err
	}
	return hc.Do(req)
}

func (c SampleClient) TestPutOrg() {
	isKillSwitchOn := true
	org := models.APIOrg{
		APIOrgURI: models.APIOrgURI{OrgKey: c.orgKey},
		APIOrgBody: models.APIOrgBody{
			DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}},
			CanaryPayloadDeviceTypes: []string{string(devicetype.LINUX)},
			IsKillSwitchOn:           &isKillSwitchOn,
			Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String(), orgchangetype.ORG_THREATHUNTER_RULES.String()},
		},
	}
	response, err := c.api.putOrg(&org)
	if err != nil {
		c.Fatalf("Error interacting with org route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from org call %d %s", response.StatusCode, readBody)
	}
}

func (c SampleClient) TestPutOrgLQ() {
	now := time.Now().UnixMilli()
	isKillSwitchOn := true
	org := models.APIOrg{
		APIOrgURI: models.APIOrgURI{OrgKey: c.orgKey},
		APIOrgBody: models.APIOrgBody{
			DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_LIVE_QUERY_CHANGE.String()}},
			CanaryPayloadDeviceTypes: []string{string(devicetype.LINUX)},
			IsKillSwitchOn:           &isKillSwitchOn,
			Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String(), orgchangetype.ORG_LIVE_QUERY_CHANGE.String()},
			LiveQueryTime:            &now,
		},
	}
	response, err := c.api.putOrg(&org)
	if err != nil {
		c.Fatalf("Error interacting with org route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from org call %d %s", response.StatusCode, readBody)
	}
}

func (c SampleClient) TestPutOrgLQNil() {
	isKillSwitchOn := true
	org := models.APIOrg{
		APIOrgURI: models.APIOrgURI{OrgKey: c.orgKey},
		APIOrgBody: models.APIOrgBody{
			DeploymentTypeChanges:    map[string][]string{string(deploymenttype.ENDPOINT): {orgchangetype.ORG_LIVE_QUERY_CHANGE.String()}},
			CanaryPayloadDeviceTypes: []string{string(devicetype.LINUX)},
			IsKillSwitchOn:           &isKillSwitchOn,
			Changes:                  []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String(), orgchangetype.ORG_LIVE_QUERY_CHANGE.String()},
		},
	}
	response, err := c.api.putOrg(&org)
	if err != nil {
		c.Fatalf("Error interacting with org route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from org call %d %s", response.StatusCode, readBody)
	}
}

// ----- Env methods ----------------------------------------

func (hc HttpClient) putEnv(env *models.APIEnv) (*http.Response, error) {
	marshalledEnv, err := json.Marshal(env.APIEnvBody)
	if err != nil {
		return nil, err
	}
	envPUT, err := http.NewRequest(http.MethodPut, fmt.Sprintf(envRouteFormatString, hc.targetHost, env.Name), bytes.NewReader(marshalledEnv))
	if err != nil {
		return nil, err
	}
	return hc.Do(envPUT)
}

func (c SampleClient) TestPutEnv() {
	env := &models.APIEnv{APIEnvURI: models.APIEnvURI{Name: c.envName}, APIEnvBody: models.APIEnvBody{CertServiceTimeChanged: true, PolicyServiceTimeChanged: true, CanaryPayloadDeviceTypes: []string{string(devicetype.WINDOWS)}}}
	response, err := c.api.putEnv(env)
	if err != nil {
		c.Fatalf("Error interacting with env route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		c.Fatalf("Bad status returned from env call %d", response.StatusCode)
	}
}

// ----- Hint methods ----------------------------------------

func (hc HttpClient) postHintRequest(request *hints.HintRequestMessage) (*http.Response, error) {
	marshalledHintRequest, err := json.Marshal(request)
	if err != nil {
		return nil, err
	}
	req, err := http.NewRequest(http.MethodPost, fmt.Sprintf(hintRouteFormatString, hc.targetHost), bytes.NewReader(marshalledHintRequest))
	if err != nil {
		return nil, err
	}
	return hc.Do(req)
}

func (c SampleClient) TestHints() {
	hintRequest := hints.NewHintRequestMessage(c.orgKey, c.deviceGroupId, c.deviceId,
		hints.BASEMESSAGE.LIVE_QUERY_VERSION, &epoch, &epoch, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	response, err := c.api.postHintRequest(hintRequest)
	if err != nil {
		c.Fatalf("Error interacting with hint route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from hint call %d %s", response.StatusCode, readBody)
	}

	var parsedResponse hints.HintResult
	if err = json.NewDecoder(response.Body).Decode(&parsedResponse); err != nil {
		c.Fatalf("Failed to decode response %v", err)
	}

	if !doHintsMatch(&parsedResponse, hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA, hinttype.HASH_LIST_REFRESH, hinttype.ZIP_CONTAINER, hinttype.LIVE_QUERY, hinttype.REP_REQUEST) {
		c.Fatalf("Unexpected hints returned %v", parsedResponse.Hints.ToStringArray())
	}
}

func (c SampleClient) TestHintsNoGroupOrOrg() {
	hintRequest := hints.NewHintRequestMessage("NOTFOUND", 0, c.deviceId, hints.BASEMESSAGE.LIVE_QUERY_VERSION, &epoch, &epoch, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	response, err := c.api.postHintRequest(hintRequest)
	if err != nil {
		c.Fatalf("Error interacting with hint route %v", err)
	}

	if response.StatusCode != http.StatusBadRequest {
		c.Fatalf("Unexpected HTTP response code, expected %d but got %d", http.StatusBadRequest, response.StatusCode)
	}
}

func (c SampleClient) TestHintsEmpty() {
	now := time.Now()
	hintRequest := hints.NewHintRequestMessage(c.orgKey, c.deviceGroupId, c.deviceId, hints.BASEMESSAGE.LIVE_QUERY_VERSION,
		&now, &now, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	response, err := c.api.postHintRequest(hintRequest)
	if err != nil {
		c.Fatalf("Error interacting with hint route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from hint call %d %s", response.StatusCode, readBody)
	}

	var parsedResponse hints.HintResult
	if err = json.NewDecoder(response.Body).Decode(&parsedResponse); err != nil {
		c.Fatalf("Failed to decode response %v", err)
	}

	if !doHintsMatch(&parsedResponse) {
		c.Fatalf("Unexpected hints returned %v", parsedResponse.Hints)
	}
}

func (c SampleClient) TestHintsAudit() {
	yp := epoch.UnixMilli()
	// NOTE: auditHints defined so as to FAIL the audit and produce log messages; hints should still be properly returned
	auditHints := []string{"NEW_CONFIG_DATA", "NEW_POLICY", "ONETIME_HASH_DELETE_LIST"}
	ahfc := hints.AuditHintsFromCBD{
		AuditHints: auditHints, AuditLastUpdateTimestamp: &yp, AuditLastLiveQueryTimestamp: &yp,
	}
	hintRequest := hints.NewHintRequestMessageWithAuditHints(c.orgKey, c.deviceGroupId, c.deviceId, hints.BASEMESSAGE.LIVE_QUERY_VERSION, &epoch, &epoch,
		false, devicetype.WINDOWS, deploymenttype.ENDPOINT, ahfc)
	response, err := c.api.postHintRequest(hintRequest)
	if err != nil {
		c.Fatalf("Error interacting with hint route %v", err)
	}
	if response.StatusCode != http.StatusOK {
		readBody, _ := io.ReadAll(response.Body)
		c.Fatalf("Bad status returned from hint call %d %s", response.StatusCode, readBody)
	}

	var parsedResponse hints.HintResult
	if err = json.NewDecoder(response.Body).Decode(&parsedResponse); err != nil {
		c.Fatalf("Failed to decode response %v", err)
	}

	if !doHintsMatch(&parsedResponse, hinttype.NEW_POLICY, hinttype.NEW_CONFIG_DATA, hinttype.HASH_LIST_REFRESH, hinttype.ZIP_CONTAINER, hinttype.LIVE_QUERY, hinttype.REP_REQUEST) {
		c.Fatalf("Unexpected hints returned %v", parsedResponse.Hints.ToStringArray())
	}
}

func doHintsMatch(hr *hints.HintResult, hints ...hinttype.HintType) bool {
	return *(hinttype.NewHintSet(hints...)) == *hr.Hints
}
